#include <algorithm>
#include <string>

#include "USBLookupTables.h"
#include "USBTypes.h"

const char* GetDescriptorName(U8 descriptor)
{
	switch (descriptor)
	{
	case DT_DEVICE:						return "DEVICE";
	case DT_CONFIGURATION:				return "CONFIGURATION";
	case DT_STRING:						return "STRING";
	case DT_INTERFACE:					return "INTERFACE";
	case DT_ENDPOINT:					return "ENDPOINT";
	case DT_DEVICE_QUALIFIER:			return "DEVICE_QUALIFIER";
	case DT_OTHER_SPEED_CONFIGURATION:	return "OTHER_SPEED_CONFIGURATION";
	case DT_INTERFACE_POWER:			return "INTERFACE_POWER";

	// HID Class specific descriptors
	case DT_HID:				return "HID";
	case DT_HID_REPORT:			return "HID Report Descriptor";
	case DT_HID_PHYS:			return "HID Physical Descriptor";

	// CDC Class specific descriptors
	case DT_CDC_CS_INTERFACE:	return "CS_INTERFACE";
	case DT_CDC_CS_ENDPOINT:	return "CS_ENDPOINT";
	}

	return "<unknown>";
}

const char* GetRequestName(U8 request)
{
	switch (request)
	{
	case GET_STATUS:		return "GET_STATUS";
	case CLEAR_FEATURE:		return "CLEAR_FEATURE";
	case SET_FEATURE:		return "SET_FEATURE";
	case SET_ADDRESS:		return "SET_ADDRESS";
	case GET_DESCRIPTOR:	return "GET_DESCRIPTOR";
	case SET_DESCRIPTOR:	return "SET_DESCRIPTOR";
	case GET_CONFIGURATION:	return "GET_CONFIGURATION";
	case SET_CONFIGURATION:	return "SET_CONFIGURATION";
	case GET_INTERFACE:		return "GET_INTERFACE";
	case SET_INTERFACE:		return "SET_INTERFACE";
	case SYNCH_FRAME:		return "SYNCH_FRAME";
	}

	return "<unknown>";
}

const char* GetHIDRequestName(U8 request)
{
	switch (request)
	{
	case GET_REPORT:	return "GET_REPORT";
	case GET_IDLE:		return "GET_IDLE";
	case GET_PROTOCOL:	return "GET_PROTOCOL";
	case SET_REPORT:	return "SET_REPORT";
	case SET_IDLE:		return "SET_IDLE";
	case SET_PROTOCOL:	return "SET_PROTOCOL";
	}

	return "<unknown>";
}

const char* GetCDCRequestName(U8 request)
{
	switch (request)
	{
	case SEND_ENCAPSULATED_COMMAND:		return "SEND_ENCAPSULATED_COMMAND";
	case GET_ENCAPSULATED_RESPONSE:		return "GET_ENCAPSULATED_RESPONSE";
	case SET_COMM_FEATURE:				return "SET_COMM_FEATURE";
	case GET_COMM_FEATURE:				return "GET_COMM_FEATURE";
	case CLEAR_COMM_FEATURE:			return "CLEAR_COMM_FEATURE";
	case SET_AUX_LINE_STATE:			return "SET_AUX_LINE_STATE";
	case SET_HOOK_STATE:				return "SET_HOOK_STATE";
	case PULSE_SETUP:					return "PULSE_SETUP";
	case SEND_PULSE:					return "SEND_PULSE";
	case SET_PULSE_TIME:				return "SET_PULSE_TIME";
	case RING_AUX_JACK:					return "RING_AUX_JACK";
	case SET_LINE_CODING:				return "SET_LINE_CODING";
	case GET_LINE_CODING:				return "GET_LINE_CODING";
	case SET_CONTROL_LINE_STATE:		return "SET_CONTROL_LINE_STATE";
	case SEND_BREAK:					return "SEND_BREAK";
	case SET_RINGER_PARMS:				return "SET_RINGER_PARMS";
	case GET_RINGER_PARMS:				return "GET_RINGER_PARMS";
	case SET_OPERATION_PARMS:			return "SET_OPERATION_PARMS";
	case GET_OPERATION_PARMS:			return "GET_OPERATION_PARMS";
	case SET_LINE_PARMS:				return "SET_LINE_PARMS";
	case GET_LINE_PARMS:				return "GET_LINE_PARMS";
	case DIAL_DIGITS:					return "DIAL_DIGITS";
	case SET_UNIT_PARAMETER:			return "SET_UNIT_PARAMETER";
	case GET_UNIT_PARAMETER:			return "GET_UNIT_PARAMETER";
	case CLEAR_UNIT_PARAMETER:			return "CLEAR_UNIT_PARAMETER";
	case GET_PROFILE:					return "GET_PROFILE";
	case SET_ETHERNET_MULTICAST_FILTERS:				return "SET_ETHERNET_MULTICAST_FILTERS";
	case SET_ETHERNET_POWER_MANAGEMENT_PATTERN_FILTER:	return "SET_ETHERNET_POWER_MANAGEMENT_PATTERN_FILTER";
	case GET_ETHERNET_POWER_MANAGEMENT_PATTERN_FILTER:	return "GET_ETHERNET_POWER_MANAGEMENT_PATTERN_FILTER";
	case SET_ETHERNET_PACKET_FILTER:	return "SET_ETHERNET_PACKET_FILTER";
	case GET_ETHERNET_STATISTIC:		return "GET_ETHERNET_STATISTIC";
	case SET_ATM_DATA_FORMAT:			return "SET_ATM_DATA_FORMAT";
	case GET_ATM_DEVICE_STATISTICS:		return "GET_ATM_DEVICE_STATISTICS";
	case SET_ATM_DEFAULT_VC:			return "SET_ATM_DEFAULT_VC";
	case GET_ATM_VC_STATISTICS:			return "GET_ATM_VC_STATISTICS";
	}

	return "<unknown>";
}

const char* GetHIDCountryName(U8 countryCode)
{
	const char* CountryNames[] = {
		"Not Supported",
		"Arabic",
		"Belgian",
		"Canadian-Bilingual",
		"Canadian-French",
		"Czech Republic",
		"Danish",
		"Finnish",
		"French",
		"German",
		"Greek",
		"Hebrew",
		"Hungary",
		"International (ISO)",
		"Italian",
		"Japan (Katakana)",
		"Korean",
		"Latin American",
		"Netherlands/Dutch",
		"Norwegian",
		"Persian (Farsi)",
		"Poland",
		"Portuguese",
		"Russia",
		"Slovakia",
		"Spanish",
		"Swedish",
		"Swiss/French",
		"Swiss/German",
		"Switzerland",
		"Taiwan",
		"Turkish-Q",
		"UK",
		"US",
		"Yugoslavia",
		"Turkish-F"
	};

	if (countryCode < sizeof(CountryNames) / sizeof(*CountryNames))
		return CountryNames[countryCode];

	return "Reserved";
}

const char* GetCDCEthFeatureSelectorName(U8 feature)
{
	const char* Features[] = {
		"RESERVED",
		"XMIT_OK",
		"RCV_OK",
		"XMIT_ERROR",
		"RCV_ERROR",
		"RCV_NO_BUFFER",
		"DIRECTED_BYTES_XMIT",
		"DIRECTED_FRAMES_XMIT",
		"MULTICAST_BYTES_XMIT",
		"MULTICAST_FRAMES_XMIT",
		"BROADCAST_BYTES_XMIT",
		"BROADCAST_FRAMES_XMIT",
		"DIRECTED_BYTES_RCV",
		"DIRECTED_FRAMES_RCV",
		"MULTICAST_BYTES_RCV",
		"MULTICAST_FRAMES_RCV",
		"BROADCAST_BYTES_RCV",
		"BROADCAST_FRAMES_RCV",
		"RCV_CRC_ERROR",
		"TRANSMIT_QUEUE_LENGTH",
		"RCV_ERROR_ALIGNMENT",
		"XMIT_ONE_COLLISION",
		"XMIT_MORE_COLLISIONS",
		"XMIT_DEFERRED",
		"XMIT_MAX_COLLISIONS",
		"RCV_OVERRUN",
		"XMIT_UNDERRUN",
		"XMIT_HEARTBEAT_FAILURE",
		"XMIT_TIMES_CRS_LOST",
		"XMIT_LATE_COLLISIONS"
	};

	if (feature < sizeof(Features) / sizeof(*Features))
		return Features[feature];

	return Features[0];
}

const char* GetCDCATMFeatureSelectorName(U8 feature)
{
	const char* Features[] = {
		"RESERVED",
		"US_CELLS_SENT",
		"DS_CELLS_RECEIVED",
		"DS_CELLS_USB_CONGESTION",
		"DS_CELLS_AAL5_CRC_ERROR",
		"DS_CELLS_HEC_ERROR",
		"DS_CELLS_HEC_ERROR_CORRECTED",
	};

	if (feature < sizeof(Features) / sizeof(*Features))
		return Features[feature];

	return Features[0];
}

const char* GetCDCDescriptorSubtypeName(U8 subtype)
{
	const char* Names[] = {
		"Header Functional Descriptor",
		"Call Management Functional Descriptor",
		"Abstract Control Management Functional Descriptor",
		"Direct Line Management Functional Descriptor",
		"Telephone Ringer Functional Descriptor",
		"Telephone Call and Line State Reporting Capabilities Functional Descriptor",
		"Union Functional descriptor",
		"Country Selection Functional Descriptor",
		"Telephone Operational Modes Functional Descriptor",
		"USB Terminal Functional Descriptor",
		"Network Channel Terminal Descriptor",
		"Protocol Unit Functional Descriptor",
		"Extension Unit Functional Descriptor",
		"Multi-Channel Management Functional Descriptor",
		"CAPI Control Management Functional Descriptor",
		"Ethernet Networking Functional Descriptor",
		"ATM Networking Functional Descriptor",
	};

	if (subtype < sizeof(Names) / sizeof(*Names))
		return Names[subtype];

	return "<unknown>";
}

struct IdNamePair
{
	U16			id;
	const char*	name;

	bool operator < (const IdNamePair& rhs) const
	{
		return id < rhs.id;
	}
};

const IdNamePair USBLangIDs[] =
{
{0x0401, "Arabic (Saudi Arabia)"},
{0x0402, "Bulgarian"},
{0x0403, "Catalan"},
{0x0404, "Chinese (Taiwan)"},
{0x0405, "Czech"},
{0x0406, "Danish"},
{0x0407, "German (Standard)"},
{0x0408, "Greek"},
{0x0409, "English (United States)"},
{0x040a, "Spanish (Traditional Sort)"},
{0x040b, "Finnish"},
{0x040c, "French (Standard)"},
{0x040d, "Hebrew"},
{0x040e, "Hungarian"},
{0x040f, "Icelandic"},
{0x0410, "Italian (Standard)"},
{0x0411, "Japanese"},
{0x0412, "Korean"},
{0x0413, "Dutch (Netherlands)"},
{0x0414, "Norwegian (Bokmal)"},
{0x0415, "Polish"},
{0x0416, "Portuguese (Brazil)"},
{0x0418, "Romanian"},
{0x0419, "Russian"},
{0x041a, "Croatian"},
{0x041b, "Slovak"},
{0x041c, "Albanian"},
{0x041d, "Swedish"},
{0x041e, "Thai"},
{0x041f, "Turkish"},
{0x0420, "Urdu (Pakistan)"},
{0x0421, "Indonesian"},
{0x0422, "Ukrainian"},
{0x0423, "Belarussian"},
{0x0424, "Slovenian"},
{0x0425, "Estonian"},
{0x0426, "Latvian"},
{0x0427, "Lithuanian"},
{0x0429, "Farsi"},
{0x042a, "Vietnamese"},
{0x042b, "Armenian."},
{0x042c, "Azeri (Latin)"},
{0x042d, "Basque"},
{0x042f, "Macedonian"},
{0x0430, "Sutu"},
{0x0436, "Afrikaans"},
{0x0437, "Georgian."},
{0x0438, "Faeroese"},
{0x0439, "Hindi."},
{0x043e, "Malay (Malaysian)"},
{0x043f, "Kazakh"},
{0x0441, "Swahili (Kenya)"},
{0x0443, "Uzbek (Latin)"},
{0x0444, "Tatar (Tatarstan)"},
{0x0445, "Bengali."},
{0x0446, "Punjabi."},
{0x0447, "Gujarati."},
{0x0448, "Oriya."},
{0x0449, "Tamil."},
{0x044a, "Telugu."},
{0x044b, "Kannada."},
{0x044c, "Malayalam."},
{0x044d, "Assamese."},
{0x044e, "Marathi."},
{0x044f, "Sanskrit."},
{0x0455, "Burmese"},
{0x0457, "Konkani."},
{0x0458, "Manipuri"},
{0x0459, "Sindhi"},
{0x04ff, "HID (Usage Data Descriptor)"},
{0x0801, "Arabic (Iraq)"},
{0x0804, "Chinese (PRC)"},
{0x0807, "German (Switzerland)"},
{0x0809, "English (United Kingdom)"},
{0x080a, "Spanish (Mexican)"},
{0x080c, "French (Belgian)"},
{0x0810, "Italian (Switzerland)"},
{0x0812, "Korean (Johab)"},
{0x0813, "Dutch (Belgium)"},
{0x0814, "Norwegian (Nynorsk)"},
{0x0816, "Portuguese (Standard)"},
{0x081a, "Serbian (Latin)"},
{0x081d, "Swedish (Finland)"},
{0x0820, "Urdu (India)"},
{0x0827, "Lithuanian (Classic)"},
{0x082c, "Azeri (Cyrillic)"},
{0x083e, "Malay (Brunei Darussalam)"},
{0x0843, "Uzbek (Cyrillic)"},
{0x0860, "Kashmiri (India)"},
{0x0861, "Nepali (India)."},
{0x0c01, "Arabic (Egypt)"},
{0x0c04, "Chinese (Hong Kong SAR, PRC)"},
{0x0c07, "German (Austria)"},
{0x0c09, "English (Australian)"},
{0x0c0a, "Spanish (Modern Sort)"},
{0x0c0c, "French (Canadian)"},
{0x0c1a, "Serbian (Cyrillic)"},
{0x1001, "Arabic (Libya)"},
{0x1004, "Chinese (Singapore)"},
{0x1007, "German (Luxembourg)"},
{0x1009, "English (Canadian)"},
{0x100a, "Spanish (Guatemala)"},
{0x100c, "French (Switzerland)"},
{0x1401, "Arabic (Algeria)"},
{0x1404, "Chinese (Macau SAR)"},
{0x1407, "German (Liechtenstein)"},
{0x1409, "English (New Zealand)"},
{0x140a, "Spanish (Costa Rica)"},
{0x140c, "French (Luxembourg)"},
{0x1801, "Arabic (Morocco)"},
{0x1809, "English (Ireland)"},
{0x180a, "Spanish (Panama)"},
{0x180c, "French (Monaco)"},
{0x1c01, "Arabic (Tunisia)"},
{0x1c09, "English (South Africa)"},
{0x1c0a, "Spanish (Dominican Republic)"},
{0x2001, "Arabic (Oman)"},
{0x2009, "English (Jamaica)"},
{0x200a, "Spanish (Venezuela)"},
{0x2401, "Arabic (Yemen)"},
{0x2409, "English (Caribbean)"},
{0x240a, "Spanish (Colombia)"},
{0x2801, "Arabic (Syria)"},
{0x2809, "English (Belize)"},
{0x280a, "Spanish (Peru)"},
{0x2c01, "Arabic (Jordan)"},
{0x2c09, "English (Trinidad)"},
{0x2c0a, "Spanish (Argentina)"},
{0x3001, "Arabic (Lebanon)"},
{0x3009, "English (Zimbabwe)"},
{0x300a, "Spanish (Ecuador)"},
{0x3401, "Arabic (Kuwait)"},
{0x3409, "English (Philippines)"},
{0x340a, "Spanish (Chile)"},
{0x3801, "Arabic (U.A.E.)"},
{0x380a, "Spanish (Uruguay)"},
{0x3c01, "Arabic (Bahrain)"},
{0x3c0a, "Spanish (Paraguay)"},
{0x4001, "Arabic (Qatar)"},
{0x400a, "Spanish (Bolivia)"},
{0x440a, "Spanish (El Salvador)"},
{0x480a, "Spanish (Honduras)"},
{0x4c0a, "Spanish (Nicaragua)"},
{0x500a, "Spanish (Puerto Rico)"},
{0xf0ff, "HID (Vendor Defined 1)"},
{0xf4ff, "HID (Vendor Defined 2)"},
{0xf8ff, "HID (Vendor Defined 3)"},
{0xfcff, "HID (Vendor Defined 4)"},

{0xffff, NULL},
};


const IdNamePair VendorIDs[] =
{
{0x0001, "Fry's Electronics"},
{0x0002, "Ingram"},
{0x0003, "Club Mac"},
{0x0004, "Nebraska Furniture Mart"},
{0x0053, "Planex"},
{0x0079, "DragonRise Inc."},
{0x0105, "Trust International B.V."},
{0x0145, "Unknown"},
{0x017c, "MLK"},
{0x0200, "TP-Link"},
{0x0204, "Chipsbank Microelectronics Co., Ltd"},
{0x0218, "Hangzhou Worlde"},
{0x02ad, "HUMAX Co., Ltd."},
{0x0300, "MM300 eBook Reader"},
{0x0324, "OCZ Technology Inc"},
{0x0325, "OCZ Technology Inc"},
{0x0386, "LTS"},
{0x03d9, "Shenzhen Sinote Tech-Electron Co., Ltd"},
{0x03da, "Bernd Walter Computer Technology"},
{0x03e8, "EndPoints, Inc."},
{0x03e9, "Thesys Microelectronics"},
{0x03ea, "Data Broadcasting Corp."},
{0x03eb, "Atmel Corp."},
{0x03ec, "Iwatsu America, Inc."},
{0x03ed, "Mitel Corp."},
{0x03ee, "Mitsumi"},
{0x03f0, "Hewlett-Packard"},
{0x03f1, "Genoa Technology"},
{0x03f2, "Oak Technology, Inc."},
{0x03f3, "Adaptec, Inc."},
{0x03f4, "Diebold, Inc."},
{0x03f5, "Siemens Electromechanical"},
{0x03f8, "Epson Imaging Technology Center"},
{0x03f9, "KeyTronic Corp."},
{0x03fb, "OPTi, Inc."},
{0x03fc, "Elitegroup Computer Systems"},
{0x03fd, "Xilinx, Inc."},
{0x03fe, "Farallon Comunications"},
{0x0400, "National Semiconductor Corp."},
{0x0401, "National Registry, Inc."},
{0x0402, "ALi Corp."},
{0x0403, "Future Technology Devices International, Ltd"},
{0x0404, "NCR Corp."},
{0x0405, "Synopsys, Inc."},
{0x0406, "Fujitsu-ICL Computers"},
{0x0407, "Fujitsu Personal Systems, Inc."},
{0x0408, "Quanta Computer, Inc."},
{0x0409, "NEC Corp."},
{0x040a, "Kodak Co."},
{0x040b, "Weltrend Semiconductor"},
{0x040c, "VTech Computers, Ltd"},
{0x040d, "VIA Technologies, Inc."},
{0x040e, "MCCI"},
{0x040f, "Echo Speech Corp."},
{0x0411, "BUFFALO INC. (formerly MelCo., Inc.)"},
{0x0412, "Award Software International"},
{0x0413, "Leadtek Research, Inc."},
{0x0414, "Giga-Byte Technology Co., Ltd"},
{0x0416, "Winbond Electronics Corp."},
{0x0417, "Symbios Logic"},
{0x0418, "AST Research"},
{0x0419, "Samsung Info. Systems America, Inc."},
{0x041a, "Phoenix Technologies, Ltd"},
{0x041b, "d'TV"},
{0x041d, "S3, Inc."},
{0x041e, "Creative Technology, Ltd"},
{0x041f, "LCS Telegraphics"},
{0x0420, "Chips and Technologies"},
{0x0421, "Nokia Mobile Phones"},
{0x0422, "ADI Systems, Inc."},
{0x0423, "Computer Access Technology Corp."},
{0x0424, "Standard Microsystems Corp."},
{0x0425, "Motorola Semiconductors HK, Ltd"},
{0x0426, "Integrated Device Technology, Inc."},
{0x0427, "Motorola Electronics Taiwan, Ltd"},
{0x0428, "Advanced Gravis Computer Tech, Ltd"},
{0x0429, "Cirrus Logic"},
{0x042a, "Ericsson Austrian, AG"},
{0x042b, "Intel Corp."},
{0x042c, "Innovative Semiconductors, Inc."},
{0x042d, "Micronics"},
{0x042e, "Acer, Inc."},
{0x042f, "Molex, Inc."},
{0x0430, "Sun Microsystems, Inc."},
{0x0431, "Itac Systems, Inc."},
{0x0432, "Unisys Corp."},
{0x0433, "Alps Electric, Inc."},
{0x0434, "Samsung Info. Systems America, Inc."},
{0x0435, "Hyundai Electronics America"},
{0x0436, "Taugagreining HF"},
{0x0437, "Framatome Connectors USA"},
{0x0438, "Advanced Micro Devices, Inc."},
{0x0439, "Voice Technologies Group"},
{0x043d, "Lexmark International, Inc."},
{0x043e, "LG Electronics USA, Inc."},
{0x043f, "RadiSys Corp."},
{0x0440, "Eizo Nanao Corp."},
{0x0441, "Winbond Systems Lab."},
{0x0442, "Ericsson, Inc."},
{0x0443, "Gateway, Inc."},
{0x0445, "Lucent Technologies, Inc."},
{0x0446, "NMB Technologies Corp."},
{0x0447, "Momentum Microsystems"},
{0x044a, "Shamrock Tech. Co., Ltd"},
{0x044b, "WSI"},
{0x044c, "CCL/ITRI"},
{0x044d, "Siemens Nixdorf AG"},
{0x044e, "Alps Electric Co., Ltd"},
{0x044f, "ThrustMaster, Inc."},
{0x0450, "DFI, Inc."},
{0x0451, "Texas Instruments, Inc."},
{0x0452, "Mitsubishi Electronics America, Inc."},
{0x0453, "CMD Technology"},
{0x0454, "Vobis Microcomputer AG"},
{0x0455, "Telematics International, Inc."},
{0x0456, "Analog Devices, Inc."},
{0x0457, "Silicon Integrated Systems Corp."},
{0x0458, "KYE Systems Corp. (Mouse Systems)"},
{0x0459, "Adobe Systems, Inc."},
{0x045a, "SONICblue, Inc."},
{0x045b, "Hitachi, Ltd"},
{0x045d, "Nortel Networks, Ltd"},
{0x045e, "Microsoft Corp."},
{0x0460, "Ace Cad Enterprise Co., Ltd"},
{0x0461, "Primax Electronics, Ltd"},
{0x0463, "MGE UPS Systems"},
{0x0464, "AMP/Tycoelectronics Corp."},
{0x0467, "AT&T Paradyne"},
{0x0468, "Wieson Technologies Co., Ltd"},
{0x046a, "Cherry GmbH"},
{0x046b, "American Megatrends, Inc."},
{0x046c, "Toshiba Corp., Digital Media Equipment"},
{0x046d, "Logitech, Inc."},
{0x046e, "Behavior Tech. Computer Corp."},
{0x046f, "Crystal Semiconductor"},
{0x0471, "Philips (or NXP)"},
{0x0472, "Chicony Electronics Co., Ltd"},
{0x0473, "Sanyo Information Business Co., Ltd"},
{0x0474, "Sanyo Electric Co., Ltd"},
{0x0475, "Relisys/Teco Information System"},
{0x0476, "AESP"},
{0x0477, "Seagate Technology, Inc."},
{0x0478, "Connectix Corp."},
{0x0479, "Advanced Peripheral Laboratories"},
{0x047a, "Semtech Corp."},
{0x047b, "Silitek Corp."},
{0x047c, "Dell Computer Corp."},
{0x047d, "Kensington"},
{0x047e, "Agere Systems, Inc. (Lucent)"},
{0x047f, "Plantronics, Inc."},
{0x0480, "Toshiba America Info. Systems, Inc."},
{0x0481, "Zenith Data Systems"},
{0x0482, "Kyocera Corp."},
{0x0483, "STMicroelectronics"},
{0x0484, "Specialix"},
{0x0485, "Nokia Monitors"},
{0x0486, "ASUS Computers, Inc."},
{0x0487, "Stewart Connector"},
{0x0488, "Cirque Corp."},
{0x0489, "Foxconn / Hon Hai"},
{0x048a, "S-MOS Systems, Inc."},
{0x048c, "Alps Electric Ireland, Ltd"},
{0x048d, "Integrated Technology Express, Inc."},
{0x048f, "Eicon Tech."},
{0x0490, "United Microelectronics Corp."},
{0x0491, "Capetronic"},
{0x0492, "Samsung SemiConductor, Inc."},
{0x0493, "MAG Technology Co., Ltd"},
{0x0495, "ESS Technology, Inc."},
{0x0496, "Micron Electronics"},
{0x0497, "Smile International"},
{0x0498, "Capetronic (Kaohsiung) Corp."},
{0x0499, "Yamaha Corp."},
{0x049a, "Gandalf Technologies, Ltd"},
{0x049b, "Curtis Computer Products"},
{0x049c, "Acer Advanced Labs, Inc."},
{0x049d, "VLSI Technology"},
{0x049f, "Compaq Computer Corp."},
{0x04a0, "Digital Equipment Corp."},
{0x04a1, "SystemSoft Corp."},
{0x04a2, "FirePower Systems"},
{0x04a3, "Trident Microsystems, Inc."},
{0x04a4, "Hitachi, Ltd"},
{0x04a5, "Acer Peripherals Inc. (now BenQ Corp.)"},
{0x04a6, "Nokia Display Products"},
{0x04a7, "Visioneer"},
{0x04a8, "Multivideo Labs, Inc."},
{0x04a9, "Canon, Inc."},
{0x04aa, "DaeWoo Telecom, Ltd"},
{0x04ab, "Chromatic Research"},
{0x04ac, "Micro Audiometrics Corp."},
{0x04ad, "Dooin Electronics"},
{0x04af, "Winnov L.P."},
{0x04b0, "Nikon Corp."},
{0x04b1, "Pan International"},
{0x04b3, "IBM Corp."},
{0x04b4, "Cypress Semiconductor Corp."},
{0x04b5, "ROHM LSI Systems USA, LLC"},
{0x04b6, "Hint Corp."},
{0x04b7, "Compal Electronics, Inc."},
{0x04b8, "Seiko Epson Corp."},
{0x04b9, "Rainbow Technologies, Inc."},
{0x04ba, "Toucan Systems, Ltd"},
{0x04bb, "I-O Data Device, Inc."},
{0x04bd, "Toshiba Electronics Taiwan Corp."},
{0x04be, "Telia Research AB"},
{0x04bf, "TDK Corp."},
{0x04c1, "U.S. Robotics (3Com)"},
{0x04c2, "Methode Electronics Far East PTE, Ltd"},
{0x04c3, "Maxi Switch, Inc."},
{0x04c4, "Lockheed Martin Energy Research"},
{0x04c5, "Fujitsu, Ltd"},
{0x04c6, "Toshiba America Electronic Components"},
{0x04c7, "Micro Macro Technologies"},
{0x04c8, "Konica Corp."},
{0x04ca, "Lite-On Technology Corp."},
{0x04cb, "Fuji Photo Film Co., Ltd"},
{0x04cc, "ST-Ericsson"},
{0x04cd, "Tatung Co. Of America"},
{0x04ce, "ScanLogic Corp."},
{0x04cf, "Myson Century, Inc."},
{0x04d0, "Digi International"},
{0x04d1, "ITT Canon"},
{0x04d2, "Altec Lansing Technologies"},
{0x04d3, "VidUS, Inc."},
{0x04d4, "LSI Logic, Inc."},
{0x04d5, "Forte Technologies, Inc."},
{0x04d6, "Mentor Graphics"},
{0x04d7, "Oki Semiconductor"},
{0x04d8, "Microchip Technology, Inc."},
{0x04d9, "Holtek Semiconductor, Inc."},
{0x04da, "Panasonic (Matsushita)"},
{0x04db, "Hypertec Pty, Ltd"},
{0x04dc, "Huan Hsin Holdings, Ltd"},
{0x04dd, "Sharp Corp."},
{0x04de, "MindShare, Inc."},
{0x04df, "Interlink Electronics"},
{0x04e1, "Iiyama North America, Inc."},
{0x04e2, "Exar Corp."},
{0x04e3, "Zilog, Inc."},
{0x04e4, "ACC Microelectronics"},
{0x04e5, "Promise Technology"},
{0x04e6, "SCM Microsystems, Inc."},
{0x04e7, "Elo TouchSystems"},
{0x04e8, "Samsung Electronics Co., Ltd"},
{0x04e9, "PC-Tel, Inc."},
{0x04ea, "Brooktree Corp."},
{0x04eb, "Northstar Systems, Inc."},
{0x04ec, "Tokyo Electron Device, Ltd"},
{0x04ed, "Annabooks"},
{0x04ef, "Pacific Electronic International, Inc."},
{0x04f0, "Daewoo Electronics Co., Ltd"},
{0x04f1, "Victor Company of Japan, Ltd"},
{0x04f2, "Chicony Electronics Co., Ltd"},
{0x04f3, "Elan Microelectronics Corp."},
{0x04f4, "Harting Elektronik, Inc."},
{0x04f5, "Fujitsu-ICL Systems, Inc."},
{0x04f6, "Norand Corp."},
{0x04f7, "Newnex Technology Corp."},
{0x04f8, "FuturePlus Systems"},
{0x04f9, "Brother Industries, Ltd"},
{0x04fa, "Dallas Semiconductor"},
{0x04fb, "Biostar Microtech International Corp."},
{0x04fc, "Sunplus Technology Co., Ltd"},
{0x04fd, "Soliton Systems, K.K."},
{0x04fe, "PFU, Ltd"},
{0x04ff, "E-CMOS Corp."},
{0x0500, "Siam United Hi-Tech"},
{0x0501, "Fujikura DDK, Ltd"},
{0x0502, "Acer, Inc."},
{0x0503, "Hitachi America, Ltd"},
{0x0504, "Hayes Microcomputer Products"},
{0x0506, "3Com Corp."},
{0x0507, "Hosiden Corp."},
{0x0508, "Clarion Co., Ltd"},
{0x0509, "Aztech Systems, Ltd"},
{0x050a, "Cinch Connectors"},
{0x050b, "Cable System International"},
{0x050c, "InnoMedia, Inc."},
{0x050d, "Belkin Components"},
{0x050e, "Neon Technology, Inc."},
{0x050f, "KC Technology, Inc."},
{0x0510, "Sejin Electron, Inc."},
{0x0511, "N'Able (DataBook) Technologies, Inc."},
{0x0512, "Hualon Microelectronics Corp."},
{0x0513, "digital-X, Inc."},
{0x0514, "FCI Electronics"},
{0x0515, "ACTC"},
{0x0516, "Longwell Electronics"},
{0x0517, "Butterfly Communications"},
{0x0518, "EzKEY Corp."},
{0x0519, "Star Micronics Co., Ltd"},
{0x051a, "WYSE Technology"},
{0x051b, "Silicon Graphics"},
{0x051c, "Shuttle, Inc."},
{0x051d, "American Power Conversion"},
{0x051e, "Scientific Atlanta, Inc."},
{0x051f, "IO Systems (Elite Electronics), Inc."},
{0x0520, "Taiwan Semiconductor Manufacturing Co."},
{0x0521, "Airborn Connectors"},
{0x0522, "Advanced Connectek, Inc."},
{0x0523, "ATEN GmbH"},
{0x0524, "Sola Electronics"},
{0x0525, "Netchip Technology, Inc."},
{0x0526, "Temic MHS S.A."},
{0x0527, "ALTRA"},
{0x0528, "ATI Technologies, Inc."},
{0x0529, "Aladdin Knowledge Systems"},
{0x052a, "Crescent Heart Software"},
{0x052b, "Tekom Technologies, Inc."},
{0x052c, "Canon Information Systems, Inc."},
{0x052d, "Avid Electronics Corp."},
{0x052e, "Standard Microsystems Corp."},
{0x052f, "Unicore Software, Inc."},
{0x0530, "American Microsystems, Inc."},
{0x0531, "Wacom Technology Corp."},
{0x0532, "Systech Corp."},
{0x0533, "Alcatel Mobile Phones"},
{0x0534, "Motorola, Inc."},
{0x0535, "LIH TZU Electric Co., Ltd"},
{0x0536, "Hand Held Products (Welch Allyn, Inc.)"},
{0x0537, "Inventec Corp."},
{0x0538, "Caldera International, Inc. (SCO)"},
{0x0539, "Shyh Shiun Terminals Co., Ltd"},
{0x053a, "PrehKeyTec GmbH"},
{0x053b, "Global Village Communication"},
{0x053c, "Institut of Microelectronic & Mechatronic Systems"},
{0x053d, "Silicon Architect"},
{0x053e, "Mobility Electronics"},
{0x053f, "Synopsys, Inc."},
{0x0540, "UniAccess AB"},
{0x0541, "Sirf Technology, Inc."},
{0x0543, "ViewSonic Corp."},
{0x0544, "Cristie Electronics, Ltd"},
{0x0545, "Xirlink, Inc."},
{0x0546, "Polaroid Corp."},
{0x0547, "Anchor Chips, Inc."},
{0x0548, "Tyan Computer Corp."},
{0x0549, "Pixera Corp."},
{0x054a, "Fujitsu Microelectronics, Inc."},
{0x054b, "New Media Corp."},
{0x054c, "Sony Corp."},
{0x054d, "Try Corp."},
{0x054e, "Proside Corp."},
{0x054f, "WYSE Technology Taiwan"},
{0x0550, "Fuji Xerox Co., Ltd"},
{0x0551, "CompuTrend Systems, Inc."},
{0x0552, "Philips Monitors"},
{0x0553, "STMicroelectronics Imaging Division (VLSI Vision)"},
{0x0554, "Dictaphone Corp."},
{0x0555, "ANAM S&T Co., Ltd"},
{0x0556, "Asahi Kasei Microsystems Co., Ltd"},
{0x0557, "ATEN International Co., Ltd"},
{0x0558, "Truevision, Inc."},
{0x0559, "Cadence Design Systems, Inc."},
{0x055a, "Kenwood USA"},
{0x055b, "KnowledgeTek, Inc."},
{0x055c, "Proton Electronic Ind."},
{0x055d, "Samsung Electro-Mechanics Co."},
{0x055e, "CTX Opto-Electronics Corp."},
{0x055f, "Mustek Systems, Inc."},
{0x0560, "Interface Corp."},
{0x0561, "Oasis Design, Inc."},
{0x0562, "Telex Communications, Inc."},
{0x0563, "Immersion Corp."},
{0x0564, "Kodak Digital Product Center, Japan Ltd. (formerly Chinon Industries Inc.)"},
{0x0565, "Peracom Networks, Inc."},
{0x0566, "Monterey International Corp."},
{0x0567, "Xyratex International, Ltd"},
{0x0568, "Quartz Ingenierie"},
{0x0569, "SegaSoft"},
{0x056a, "Wacom Co., Ltd"},
{0x056b, "Decicon, Inc."},
{0x056c, "eTEK Labs"},
{0x056d, "EIZO Corp."},
{0x056e, "Elecom Co., Ltd"},
{0x056f, "Korea Data Systems Co., Ltd"},
{0x0570, "Epson America"},
{0x0571, "Interex, Inc."},
{0x0572, "Conexant Systems (Rockwell), Inc."},
{0x0573, "Zoran Co. Personal Media Division (Nogatech)"},
{0x0574, "City University of Hong Kong"},
{0x0575, "Philips Creative Display Solutions"},
{0x0576, "BAFO/Quality Computer Accessories"},
{0x0577, "ELSA"},
{0x0578, "Intrinsix Corp."},
{0x0579, "GVC Corp."},
{0x057a, "Samsung Electronics America"},
{0x057b, "Y-E Data, Inc."},
{0x057c, "AVM GmbH"},
{0x057d, "Shark Multimedia, Inc."},
{0x057e, "Nintendo Co., Ltd"},
{0x057f, "QuickShot, Ltd"},
{0x0580, "Denron, Inc."},
{0x0581, "Racal Data Group"},
{0x0582, "Roland Corp."},
{0x0583, "Padix Co., Ltd (Rockfire)"},
{0x0584, "RATOC System, Inc."},
{0x0585, "FlashPoint Technology, Inc."},
{0x0586, "ZyXEL Communications Corp."},
{0x0587, "America Kotobuki Electronics Industries, Inc."},
{0x0588, "Sapien Design"},
{0x0589, "Victron"},
{0x058a, "Nohau Corp."},
{0x058b, "Infineon Technologies"},
{0x058c, "In Focus Systems"},
{0x058d, "Micrel Semiconductor"},
{0x058e, "Tripath Technology, Inc."},
{0x058f, "Alcor Micro Corp."},
{0x0590, "Omron Corp."},
{0x0591, "Questra Consulting"},
{0x0592, "Powerware Corp."},
{0x0593, "Incite"},
{0x0594, "Princeton Graphic Systems"},
{0x0595, "Zoran Microelectronics, Ltd"},
{0x0596, "MicroTouch Systems, Inc."},
{0x0597, "Trisignal Communications"},
{0x0598, "Niigata Canotec Co., Inc."},
{0x0599, "Brilliance Semiconductor, Inc."},
{0x059a, "Spectrum Signal Processing, Inc."},
{0x059b, "Iomega Corp."},
{0x059c, "A-Trend Technology Co., Ltd"},
{0x059d, "Advanced Input Devices"},
{0x059e, "Intelligent Instrumentation"},
{0x059f, "LaCie, Ltd"},
{0x05a0, "Vetronix Corp."},
{0x05a1, "USC Corp."},
{0x05a2, "Fuji Film Microdevices Co., Ltd"},
{0x05a3, "ARC International"},
{0x05a4, "Ortek Technology, Inc."},
{0x05a5, "Sampo Technology Corp."},
{0x05a6, "Cisco Systems, Inc."},
{0x05a7, "Bose Corp."},
{0x05a8, "Spacetec IMC Corp."},
{0x05a9, "OmniVision Technologies, Inc."},
{0x05aa, "Utilux South China, Ltd"},
{0x05ab, "In-System Design"},
{0x05ac, "Apple, Inc."},
{0x05ad, "Y.C. Cable U.S.A., Inc."},
{0x05ae, "Synopsys, Inc."},
{0x05af, "Jing-Mold Enterprise Co., Ltd"},
{0x05b0, "Fountain Technologies, Inc."},
{0x05b1, "First International Computer, Inc."},
{0x05b4, "LG Semicon Co., Ltd"},
{0x05b5, "Dialogic Corp."},
{0x05b6, "Proxima Corp."},
{0x05b7, "Medianix Semiconductor, Inc."},
{0x05b8, "Agiler, Inc."},
{0x05b9, "Philips Research Laboratories"},
{0x05ba, "DigitalPersona, Inc."},
{0x05bb, "Grey Cell Systems"},
{0x05bc, "3G Green Green Globe Co., Ltd"},
{0x05bd, "RAFI GmbH & Co. KG"},
{0x05be, "Tyco Electronics (Raychem)"},
{0x05bf, "S & S Research"},
{0x05c0, "Keil Software"},
{0x05c1, "Kawasaki Microelectronics, Inc."},
{0x05c2, "Media Phonics (Suisse) S.A."},
{0x05c5, "Digi International, Inc."},
{0x05c6, "Qualcomm, Inc."},
{0x05c7, "Qtronix Corp."},
{0x05c8, "Cheng Uei Precision Industry Co., Ltd (Foxlink)"},
{0x05c9, "Semtech Corp."},
{0x05ca, "Ricoh Co., Ltd"},
{0x05cb, "PowerVision Technologies, Inc."},
{0x05cc, "ELSA AG"},
{0x05cd, "Silicom, Ltd"},
{0x05ce, "sci-worx GmbH"},
{0x05cf, "Sung Forn Co., Ltd"},
{0x05d0, "GE Medical Systems Lunar"},
{0x05d1, "Brainboxes, Ltd"},
{0x05d2, "Wave Systems Corp."},
{0x05d3, "Tohoku Ricoh Co., Ltd"},
{0x05d5, "Super Gate Technology Co., Ltd"},
{0x05d6, "Philips Semiconductors, CICT"},
{0x05d7, "Thomas & Betts Corp."},
{0x05d8, "Ultima Electronics Corp."},
{0x05d9, "Axiohm Transaction Solutions"},
{0x05da, "Microtek International, Inc."},
{0x05db, "Sun Corp. (Suntac?)"},
{0x05dc, "Lexar Media, Inc."},
{0x05dd, "Delta Electronics, Inc."},
{0x05df, "Silicon Vision, Inc."},
{0x05e0, "Symbol Technologies"},
{0x05e1, "Syntek Semiconductor Co., Ltd"},
{0x05e2, "ElecVision, Inc."},
{0x05e3, "Genesys Logic, Inc."},
{0x05e4, "Red Wing Corp."},
{0x05e5, "Fuji Electric Co., Ltd"},
{0x05e6, "Keithley Instruments"},
{0x05e8, "ICC, Inc."},
{0x05e9, "Kawasaki LSI"},
{0x05eb, "FFC, Ltd"},
{0x05ec, "COM21, Inc."},
{0x05ee, "Cytechinfo Inc."},
{0x05ef, "AVB, Inc. [anko?]"},
{0x05f0, "Canopus Co., Ltd"},
{0x05f1, "Compass Communications"},
{0x05f2, "Dexin Corp., Ltd"},
{0x05f3, "PI Engineering, Inc."},
{0x05f5, "Unixtar Technology, Inc."},
{0x05f6, "AOC International"},
{0x05f7, "RFC Distribution(s) PTE, Ltd"},
{0x05f9, "PSC Scanning, Inc."},
{0x05fa, "Siemens Telecommunications Systems, Ltd"},
{0x05fc, "Harman Multimedia"},
{0x05fd, "InterAct, Inc."},
{0x05fe, "Chic Technology Corp."},
{0x05ff, "LeCroy Corp."},
{0x0600, "Barco Display Systems"},
{0x0601, "Jazz Hipster Corp."},
{0x0602, "Vista Imaging, Inc."},
{0x0603, "Novatek Microelectronics Corp."},
{0x0604, "Jean Co., Ltd"},
{0x0605, "Anchor C&C Co., Ltd"},
{0x0606, "Royal Information Electronics Co., Ltd"},
{0x0607, "Bridge Information Co., Ltd"},
{0x0608, "Genrad Ads"},
{0x0609, "SMK Manufacturing, Inc."},
{0x060a, "Worthington Data Solutions, Inc."},
{0x060b, "Solid Year"},
{0x060c, "EEH Datalink GmbH"},
{0x060d, "Auctor Corp."},
{0x060e, "Transmonde Technologies, Inc."},
{0x060f, "Joinsoon Electronics Mfg. Co., Ltd"},
{0x0610, "Costar Electronics, Inc."},
{0x0611, "Totoku Electric Co., Ltd"},
{0x0613, "TransAct Technologies, Inc."},
{0x0614, "Bio-Rad Laboratories"},
{0x0615, "Quabbin Wire & Cable Co., Inc."},
{0x0616, "Future Techno Designs PVT, Ltd"},
{0x0617, "Swiss Federal Insitute of Technology"},
{0x0618, "MacAlly"},
{0x0619, "Seiko Instruments, Inc."},
{0x061a, "Veridicom International, Inc."},
{0x061b, "Promptus Communications, Inc."},
{0x061c, "Act Labs, Ltd"},
{0x061d, "Quatech, Inc."},
{0x061e, "Nissei Electric Co."},
{0x0620, "Alaris, Inc."},
{0x0621, "ODU-Steckverbindungssysteme GmbH & Co. KG"},
{0x0622, "Iotech, Inc."},
{0x0623, "Littelfuse, Inc."},
{0x0624, "Avocent Corp."},
{0x0625, "TiMedia Technology Co., Ltd"},
{0x0626, "Nippon Systems Development Co., Ltd"},
{0x0627, "Adomax Technology Co., Ltd"},
{0x0628, "Tasking Software, Inc."},
{0x0629, "Zida Technologies, Ltd"},
{0x062a, "Creative Labs"},
{0x062b, "Greatlink Electronics Taiwan, Ltd"},
{0x062c, "Institute for Information Industry"},
{0x062d, "Taiwan Tai-Hao Enterprises Co., Ltd"},
{0x062e, "Mainsuper Enterprises Co., Ltd"},
{0x062f, "Sin Sheng Terminal & Machine, Inc."},
{0x0631, "JUJO Electronics Corp."},
{0x0633, "Cyrix Corp."},
{0x0634, "Micron Technology, Inc."},
{0x0635, "Methode Electronics, Inc."},
{0x0636, "Sierra Imaging, Inc."},
{0x0638, "Avision, Inc."},
{0x0639, "Chrontel, Inc."},
{0x063a, "Techwin Corp."},
{0x063b, "Taugagreining HF"},
{0x063c, "Yamaichi Electronics Co., Ltd (Sakura)"},
{0x063d, "Fong Kai Industrial Co., Ltd"},
{0x063e, "RealMedia Technology, Inc."},
{0x063f, "New Technology Cable, Ltd"},
{0x0640, "Hitex Development Tools"},
{0x0641, "Woods Industries, Inc."},
{0x0642, "VIA Medical Corp."},
{0x0644, "TEAC Corp."},
{0x0645, "Who? Vision Systems, Inc."},
{0x0646, "UMAX"},
{0x0647, "Acton Research Corp."},
{0x0648, "Inside Out Networks"},
{0x0649, "Weli Science Co., Ltd"},
{0x064b, "Analog Devices, Inc. (White Mountain DSP)"},
{0x064c, "Ji-Haw Industrial Co., Ltd"},
{0x064d, "TriTech Microelectronics, Ltd"},
{0x064e, "Suyin Corp."},
{0x064f, "WIBU-Systems AG"},
{0x0650, "Dynapro Systems"},
{0x0651, "Likom Technology Sdn. Bhd."},
{0x0652, "Stargate Solutions, Inc."},
{0x0653, "CNF, Inc."},
{0x0654, "Granite Microsystems, Inc."},
{0x0655, "Space Shuttle Hi-Tech Co., Ltd"},
{0x0656, "Glory Mark Electronic, Ltd"},
{0x0657, "Tekcon Electronics Corp."},
{0x0658, "Sigma Designs, Inc."},
{0x0659, "Aethra"},
{0x065a, "Optoelectronics Co., Ltd"},
{0x065b, "Tracewell Systems"},
{0x065e, "Silicon Graphics"},
{0x065f, "Good Way Technology Co., Ltd & GWC technology Inc."},
{0x0660, "TSAY-E (BVI) International, Inc."},
{0x0661, "Hamamatsu Photonics K.K."},
{0x0662, "Kansai Electric Co., Ltd"},
{0x0663, "Topmax Electronic Co., Ltd"},
{0x0664, "ET&T Technology Co., Ltd."},
{0x0665, "Cypress Semiconductor"},
{0x0667, "Aiwa Co., Ltd"},
{0x0668, "WordWand"},
{0x0669, "Oce' Printing Systems GmbH"},
{0x066a, "Total Technologies, Ltd"},
{0x066b, "Linksys, Inc."},
{0x066d, "Entrega, Inc."},
{0x066e, "Acer Semiconductor America, Inc."},
{0x066f, "SigmaTel, Inc."},
{0x0670, "Sequel Imaging"},
{0x0672, "Labtec, Inc."},
{0x0673, "HCL"},
{0x0674, "Key Mouse Electronic Enterprise Co., Ltd"},
{0x0675, "DrayTek Corp."},
{0x0676, "Teles AG"},
{0x0677, "Aiwa Co., Ltd"},
{0x0678, "ACard Technology Corp."},
{0x067b, "Prolific Technology, Inc."},
{0x067c, "Efficient Networks, Inc."},
{0x067d, "Hohner Corp."},
{0x067e, "Intermec Technologies Corp."},
{0x067f, "Virata, Ltd"},
{0x0680, "Realtek Semiconductor Corp., CPP Div. (Avance Logic)"},
{0x0681, "Siemens Information and Communication Products"},
{0x0682, "Victor Company of Japan, Ltd"},
{0x0684, "Actiontec Electronics, Inc."},
{0x0685, "ZD Incorporated"},
{0x0686, "Minolta Co., Ltd"},
{0x068a, "Pertech, Inc."},
{0x068b, "Potrans International, Inc."},
{0x068e, "CH Products, Inc."},
{0x0690, "Golden Bridge Electech, Inc."},
{0x0693, "Hagiwara Sys-Com Co., Ltd"},
{0x0694, "Lego Group"},
{0x0698, "Chuntex (CTX)"},
{0x0699, "Tektronix, Inc."},
{0x069a, "Askey Computer Corp."},
{0x069b, "Thomson, Inc."},
{0x069d, "Hughes Network Systems (HNS)"},
{0x069e, "Welcat Inc."},
{0x069f, "Allied Data Technologies BV"},
{0x06a2, "Topro Technology, Inc."},
{0x06a3, "Saitek PLC"},
{0x06a4, "Xiamen Doowell Electron Co., Ltd"},
{0x06a5, "Divio"},
{0x06a7, "MicroStore, Inc."},
{0x06a8, "Topaz Systems, Inc."},
{0x06a9, "Westell"},
{0x06aa, "Sysgration, Ltd"},
{0x06ac, "Fujitsu Laboratories of America, Inc."},
{0x06ad, "Greatland Electronics Taiwan, Ltd"},
{0x06ae, "Professional Multimedia Testing Centre"},
{0x06af, "Harting, Inc. of North America"},
{0x06b8, "Pixela Corp."},
{0x06b9, "Alcatel Telecom"},
{0x06ba, "Smooth Cord & Connector Co., Ltd"},
{0x06bb, "EDA, Inc."},
{0x06bc, "Oki Data Corp."},
{0x06bd, "AGFA-Gevaert NV"},
{0x06be, "AME Optimedia Technology Co., Ltd"},
{0x06bf, "Leoco Corp."},
{0x06c2, "Phidgets Inc. (formerly GLAB)"},
{0x06c4, "Bizlink International Corp."},
{0x06c5, "Hagenuk, GmbH"},
{0x06c6, "Infowave Software, Inc."},
{0x06c8, "SIIG, Inc."},
{0x06c9, "Taxan (Europe), Ltd"},
{0x06ca, "Newer Technology, Inc."},
{0x06cb, "Synaptics, Inc."},
{0x06cc, "Terayon Communication Systems"},
{0x06cd, "Keyspan"},
{0x06ce, "Contec"},
{0x06cf, "SpheronVR AG"},
{0x06d0, "LapLink, Inc."},
{0x06d1, "Daewoo Electronics Co., Ltd"},
{0x06d3, "Mitsubishi Electric Corp."},
{0x06d4, "Cisco Systems"},
{0x06d5, "Toshiba"},
{0x06d6, "Aashima Technology B.V."},
{0x06d7, "Network Computing Devices (NCD)"},
{0x06d8, "Technical Marketing Research, Inc."},
{0x06da, "Phoenixtec Power Co., Ltd"},
{0x06db, "Paradyne"},
{0x06dc, "Foxlink Image Technology Co., Ltd"},
{0x06de, "Heisei Electronics Co., Ltd"},
{0x06e0, "Multi-Tech Systems, Inc."},
{0x06e1, "ADS Technologies, Inc."},
{0x06e4, "Alcatel Microelectronics"},
{0x06e6, "Tiger Jet Network, Inc."},
{0x06ea, "Sirius Technologies"},
{0x06eb, "PC Expert Tech. Co., Ltd"},
{0x06ef, "I.A.C. Geometrische Ingenieurs B.V."},
{0x06f0, "T.N.C Industrial Co., Ltd"},
{0x06f1, "Opcode Systems, Inc."},
{0x06f2, "Emine Technology Co."},
{0x06f6, "Wintrend Technology Co., Ltd"},
{0x06f7, "Wailly Technology Ltd"},
{0x06f8, "Guillemot Corp."},
{0x06f9, "ASYST electronic d.o.o."},
{0x06fa, "HSD S.r.L"},
{0x06fc, "Motorola Semiconductor Products Sector"},
{0x06fd, "Boston Acoustics"},
{0x06fe, "Gallant Computer, Inc."},
{0x0701, "Supercomal Wire & Cable SDN. BHD."},
{0x0703, "Bvtech Industry, Inc."},
{0x0705, "NKK Corp."},
{0x0706, "Ariel Corp."},
{0x0707, "Standard Microsystems Corp."},
{0x0708, "Putercom Co., Ltd"},
{0x0709, "Silicon Systems, Ltd (SSL)"},
{0x070a, "Oki Electric Industry Co., Ltd"},
{0x070d, "Comoss Electronic Co., Ltd"},
{0x070e, "Excel Cell Electronic Co., Ltd"},
{0x0710, "Connect Tech, Inc."},
{0x0711, "Magic Control Technology Corp."},
{0x0713, "Interval Research Corp."},
{0x0714, "NewMotion, Inc."},
{0x0717, "ZNK Corp."},
{0x0718, "Imation Corp."},
{0x0719, "Tremon Enterprises Co., Ltd"},
{0x071b, "Domain Technologies, Inc."},
{0x071c, "Xionics Document Technologies, Inc."},
{0x071d, "Eicon Networks Corp."},
{0x071e, "Ariston Technologies"},
{0x0723, "Centillium Communications Corp."},
{0x0726, "Vanguard International Semiconductor-America"},
{0x0729, "Amitm"},
{0x072e, "Sunix Co., Ltd"},
{0x072f, "Advanced Card Systems, Ltd"},
{0x0731, "Susteen, Inc."},
{0x0732, "Goldfull Electronics & Telecommunications Corp."},
{0x0733, "ViewQuest Technologies, Inc."},
{0x0734, "Lasat Communications A/S"},
{0x0735, "Asuscom Network"},
{0x0736, "Lorom Industrial Co., Ltd"},
{0x0738, "Mad Catz, Inc."},
{0x073a, "Chaplet Systems, Inc."},
{0x073b, "Suncom Technologies"},
{0x073c, "Industrial Electronic Engineers, Inc."},
{0x073d, "Eutron S.p.a."},
{0x073e, "NEC, Inc."},
{0x0745, "Syntech Information Co., Ltd"},
{0x0746, "Onkyo Corp."},
{0x0747, "Labway Corp."},
{0x0748, "Strong Man Enterprise Co., Ltd"},
{0x0749, "EVer Electronics Corp."},
{0x074a, "Ming Fortune Industry Co., Ltd"},
{0x074b, "Polestar Tech. Corp."},
{0x074c, "C-C-C Group PLC"},
{0x074d, "Micronas GmbH"},
{0x074e, "Digital Stream Corp."},
{0x0755, "Aureal Semiconductor"},
{0x0757, "Network Technologies, Inc."},
{0x075b, "Sophisticated Circuits, Inc."},
{0x0763, "Midiman"},
{0x0764, "Cyber Power System, Inc."},
{0x0765, "X-Rite, Inc."},
{0x0766, "Jess-Link Products Co., Ltd"},
{0x0767, "Tokheim Corp."},
{0x0768, "Camtel Technology Corp."},
{0x0769, "Surecom Technology Corp."},
{0x076a, "Smart Technology Enablers, Inc."},
{0x076b, "OmniKey AG"},
{0x076c, "Partner Tech"},
{0x076d, "Denso Corp."},
{0x076e, "Kuan Tech Enterprise Co., Ltd"},
{0x076f, "Jhen Vei Electronic Co., Ltd"},
{0x0770, "Welch Allyn, Inc - Medical Division"},
{0x0771, "Observator Instruments BV"},
{0x0772, "Your data Our Care"},
{0x0774, "AmTRAN Technology Co., Ltd"},
{0x0775, "Longshine Electronics Corp."},
{0x0776, "Inalways Corp."},
{0x0777, "Comda Enterprise Corp."},
{0x0778, "Volex, Inc."},
{0x0779, "Fairchild Semiconductor"},
{0x077a, "Sankyo Seiki Mfg. Co., Ltd"},
{0x077b, "Linksys"},
{0x077c, "Forward Electronics Co., Ltd"},
{0x077d, "Griffin Technology"},
{0x077f, "Well Excellent & Most Corp."},
{0x0780, "Sagem Monetel GmbH"},
{0x0781, "SanDisk Corp."},
{0x0782, "Trackerball"},
{0x0783, "C3PO"},
{0x0784, "Vivitar, Inc."},
{0x0785, "NTT-ME"},
{0x0789, "Logitec Corp."},
{0x078b, "Happ Controls, Inc."},
{0x078c, "GTCO/CalComp"},
{0x078e, "Brincom, Inc."},
{0x0790, "Pro-Image Manufacturing Co., Ltd"},
{0x0791, "Copartner Wire and Cable Mfg. Corp."},
{0x0792, "Axis Communications AB"},
{0x0793, "Wha Yu Industrial Co., Ltd"},
{0x0794, "ABL Electronics Corp."},
{0x0795, "RealChip, Inc."},
{0x0796, "Certicom Corp."},
{0x0797, "Grandtech Semiconductor Corp."},
{0x0798, "Optelec"},
{0x0799, "Altera"},
{0x079b, "Sagem"},
{0x079d, "Alfadata Computer Corp."},
{0x07a1, "Digicom S.p.A."},
{0x07a2, "National Technical Systems"},
{0x07a3, "Onnto Corp."},
{0x07a4, "Be, Inc."},
{0x07a6, "ADMtek, Inc."},
{0x07aa, "Corega K.K."},
{0x07ab, "Freecom Technologies"},
{0x07af, "Microtech"},
{0x07b0, "Trust Technologies"},
{0x07b1, "IMP, Inc."},
{0x07b2, "Motorola BCS, Inc."},
{0x07b3, "Plustek, Inc."},
{0x07b4, "Olympus Optical Co., Ltd"},
{0x07b5, "Mega World International, Ltd"},
{0x07b6, "Marubun Corp."},
{0x07b7, "TIME Interconnect, Ltd"},
{0x07b8, "AboCom Systems Inc"},
{0x07bc, "Canon Computer Systems, Inc."},
{0x07bd, "Webgear, Inc."},
{0x07be, "Veridicom"},
{0x07c0, "Code Mercenaries Hard- und Software GmbH"},
{0x07c1, "Keisokugiken"},
{0x07c4, "Datafab Systems, Inc."},
{0x07c5, "APG Cash Drawer"},
{0x07c6, "ShareWave, Inc."},
{0x07c7, "Powertech Industrial Co., Ltd"},
{0x07c8, "B.U.G., Inc."},
{0x07c9, "Allied Telesyn International"},
{0x07ca, "AVerMedia Technologies, Inc."},
{0x07cb, "Kingmax Technology, Inc."},
{0x07cc, "Carry Computer Eng., Co., Ltd"},
{0x07cd, "Elektor"},
{0x07cf, "Casio Computer Co., Ltd"},
{0x07d0, "Dazzle"},
{0x07d1, "D-Link System"},
{0x07d2, "Aptio Products, Inc."},
{0x07d3, "Cyberdata Corp."},
{0x07d5, "Radiant Systems"},
{0x07d7, "GCC Technologies, Inc."},
{0x07da, "Arasan Chip Systems"},
{0x07de, "Diamond Multimedia"},
{0x07df, "David Electronics Co., Ltd"},
{0x07e1, "Ambient Technologies, Inc."},
{0x07e2, "Elmeg GmbH & Co., Ltd"},
{0x07e3, "Planex Communications, Inc."},
{0x07e4, "Movado Enterprise Co., Ltd"},
{0x07e5, "QPS, Inc."},
{0x07e6, "Allied Cable Corp."},
{0x07e7, "Mirvo Toys, Inc."},
{0x07e8, "Labsystems"},
{0x07ea, "Iwatsu Electric Co., Ltd"},
{0x07eb, "Double-H Technology Co., Ltd"},
{0x07ec, "Taiyo Electric Wire & Cable Co., Ltd"},
{0x07ee, "Torex Retail (formerly Logware)"},
{0x07ef, "STSN"},
{0x07f2, "Microcomputer Applications, Inc."},
{0x07f6, "Circuit Assembly Corp."},
{0x07f7, "Century Corp."},
{0x07f9, "Dotop Technology, Inc."},
{0x07fa, "DrayTek Corp."},
{0x07fd, "Mark of the Unicorn"},
{0x07ff, "Unknown"},
{0x0801, "MagTek"},
{0x0802, "Mako Technologies, LLC"},
{0x0803, "Zoom Telephonics, Inc."},
{0x0809, "Genicom Technology, Inc."},
{0x080a, "Evermuch Technology Co., Ltd"},
{0x080b, "Cross Match Technologies"},
{0x080c, "Datalogic S.p.A."},
{0x080d, "Teco Image Systems Co., Ltd"},
{0x0810, "Personal Communication Systems, Inc."},
{0x0813, "Mattel, Inc."},
{0x0819, "eLicenser"},
{0x081a, "MG Logic"},
{0x081b, "Indigita Corp."},
{0x081c, "Mipsys"},
{0x081e, "AlphaSmart, Inc."},
{0x0822, "Reudo Corp."},
{0x0825, "GC Protronics"},
{0x0826, "Data Transit"},
{0x0827, "BroadLogic, Inc."},
{0x0828, "Sato Corp."},
{0x0829, "DirecTV Broadband, Inc. (Telocity)"},
{0x082d, "Handspring"},
{0x0830, "Palm, Inc."},
{0x0832, "Kouwell Electronics Corp."},
{0x0833, "Sourcenext Corp."},
{0x0835, "Action Star Enterprise Co., Ltd"},
{0x0836, "TrekStor"},
{0x0839, "Samsung Techwin Co., Ltd"},
{0x083a, "Accton Technology Corp."},
{0x083f, "Global Village"},
{0x0840, "Argosy Research, Inc."},
{0x0841, "Rioport.com, Inc."},
{0x0844, "Welland Industrial Co., Ltd"},
{0x0846, "NetGear, Inc."},
{0x084d, "Minton Optic Industry Co., Inc."},
{0x084e, "KB Gear"},
{0x084f, "Empeg"},
{0x0850, "Fast Point Technologies, Inc."},
{0x0851, "Macronix International Co., Ltd"},
{0x0852, "CSEM"},
{0x0853, "Topre Corporation"},
{0x0854, "ActiveWire, Inc."},
{0x0856, "B&B Electronics"},
{0x0858, "Hitachi Maxell, Ltd"},
{0x0859, "Minolta Systems Laboratory, Inc."},
{0x085a, "Xircom"},
{0x085c, "ColorVision, Inc."},
{0x0862, "Teletrol Systems, Inc."},
{0x0863, "Filanet Corp."},
{0x0864, "NetGear, Inc."},
{0x0867, "Data Translation, Inc."},
{0x086a, "Emagic Soft- und Hardware GmbH"},
{0x086c, "DeTeWe - Deutsche Telephonwerke AG & Co."},
{0x086e, "System TALKS, Inc."},
{0x086f, "MEC IMEX, Inc."},
{0x0870, "Metricom"},
{0x0871, "SanDisk, Inc."},
{0x0873, "Xpeed, Inc."},
{0x0874, "A-Tec Subsystem, Inc."},
{0x0879, "Comtrol Corp."},
{0x087c, "Adesso/Kbtek America, Inc."},
{0x087d, "Jaton Corp."},
{0x087e, "Fujitsu Computer Products of America"},
{0x087f, "QualCore Logic Inc."},
{0x0880, "APT Technologies, Inc."},
{0x0883, "Recording Industry Association of America (RIAA)"},
{0x0885, "Boca Research, Inc."},
{0x0886, "XAC Automation Corp."},
{0x0887, "Hannstar Electronics Corp."},
{0x088a, "TechTools"},
{0x088b, "MassWorks, Inc."},
{0x088c, "Swecoin AB"},
{0x088e, "iLok"},
{0x0892, "DioGraphy, Inc."},
{0x0897, "Lauterbach"},
{0x089c, "United Technologies Research Cntr."},
{0x089d, "Icron Technologies Corp."},
{0x089e, "NST Co., Ltd"},
{0x089f, "Primex Aerospace Co."},
{0x08a5, "e9, Inc."},
{0x08a6, "Toshiba TEC"},
{0x08a8, "Andrea Electronics"},
{0x08a9, "CWAV Inc."},
{0x08ae, "Macally (Mace Group, Inc.)"},
{0x08b4, "Sorenson Vision, Inc."},
{0x08b7, "NATSU"},
{0x08b8, "J. Gordon Electronic Design, Inc."},
{0x08b9, "RadioShack Corp. (Tandy)"},
{0x08bb, "Texas Instruments"},
{0x08bd, "Citizen Watch Co., Ltd"},
{0x08c3, "Precise Biometrics"},
{0x08c4, "Proxim, Inc."},
{0x08c7, "Key Nice Enterprise Co., Ltd"},
{0x08c8, "2Wire, Inc."},
{0x08c9, "Nippon Telegraph and Telephone Corp."},
{0x08ca, "Aiptek International, Inc."},
{0x08cd, "Jue Hsun Ind. Corp."},
{0x08ce, "Long Well Electronics Corp."},
{0x08cf, "Productivity Enhancement Products"},
{0x08d1, "smartBridges, Inc."},
{0x08d3, "Virtual Ink"},
{0x08d4, "Fujitsu Siemens Computers"},
{0x08d8, "IXXAT Automation GmbH"},
{0x08d9, "Increment P Corp."},
{0x08dd, "Billionton Systems, Inc."},
{0x08de, "???"},
{0x08df, "Spyrus, Inc."},
{0x08e3, "Olitec, Inc."},
{0x08e4, "Pioneer Corp."},
{0x08e5, "Litronic"},
{0x08e6, "Gemalto (was Gemplus)"},
{0x08e7, "Pan-International Wire & Cable"},
{0x08e8, "Integrated Memory Logic"},
{0x08e9, "Extended Systems, Inc."},
{0x08ea, "Ericsson, Inc., Blue Ridge Labs"},
{0x08ec, "M-Systems Flash Disk Pioneers"},
{0x08ed, "MediaTek Inc."},
{0x08ee, "CCSI/Hesso"},
{0x08f0, "Corex Technologies"},
{0x08f1, "CTI Electronics Corp."},
{0x08f2, "Gotop Information Inc."},
{0x08f5, "SysTec Co., Ltd"},
{0x08f6, "Logic 3 International, Ltd"},
{0x08f7, "Vernier"},
{0x08f8, "Keen Top International Enterprise Co., Ltd"},
{0x08f9, "Wipro Technologies"},
{0x08fa, "Caere"},
{0x08fb, "Socket Communications"},
{0x08fc, "Sicon Cable Technology Co., Ltd"},
{0x08fd, "Digianswer A/S"},
{0x08ff, "AuthenTec, Inc."},
{0x0900, "Pinnacle Systems, Inc."},
{0x0901, "VST Technologies"},
{0x0906, "Faraday Technology Corp."},
{0x0908, "Siemens AG"},
{0x0909, "Audio-Technica Corp."},
{0x090a, "Trumpion Microelectronics, Inc."},
{0x090b, "Neurosmith"},
{0x090c, "Silicon Motion, Inc. - Taiwan (formerly Feiya Technology Corp.)"},
{0x090d, "Multiport Computer Vertriebs GmbH"},
{0x090e, "Shining Technology, Inc."},
{0x090f, "Fujitsu Devices, Inc."},
{0x0910, "Alation Systems, Inc."},
{0x0911, "Philips Speech Processing"},
{0x0912, "Voquette, Inc."},
{0x0915, "GlobeSpan, Inc."},
{0x0917, "SmartDisk Corp."},
{0x0919, "Tiger Electronics"},
{0x091e, "Garmin International"},
{0x0920, "Echelon Co."},
{0x0921, "GoHubs, Inc."},
{0x0922, "Dymo-CoStar Corp."},
{0x0923, "IC Media Corp."},
{0x0924, "Xerox"},
{0x0925, "Lakeview Research"},
{0x0927, "Summus, Ltd"},
{0x0928, "PLX Technology, Inc. (formerly Oxford Semiconductor, Ltd)"},
{0x0929, "American Biometric Co."},
{0x092a, "Toshiba Information & Industrial Sys. And Services"},
{0x092b, "Sena Technologies, Inc."},
{0x092f, "Northern Embedded Science/CAVNEX"},
{0x0930, "Toshiba Corp."},
{0x0931, "Harmonic Data Systems, Ltd"},
{0x0932, "Crescentec Corp."},
{0x0933, "Quantum Corp."},
{0x0934, "Spirent Communications"},
{0x0936, "NuTesla"},
{0x0939, "Lumberg, Inc."},
{0x093a, "Pixart Imaging, Inc."},
{0x093b, "Plextor Corp."},
{0x093c, "Intrepid Control Systems, Inc."},
{0x093d, "InnoSync, Inc."},
{0x093e, "J.S.T. Mfg. Co., Ltd"},
{0x093f, "Olympia Telecom Vertriebs GmbH"},
{0x0940, "Japan Storage Battery Co., Ltd"},
{0x0941, "Photobit Corp."},
{0x0942, "i2Go.com, LLC"},
{0x0943, "HCL Technologies India Private, Ltd"},
{0x0944, "KORG, Inc."},
{0x0945, "Pasco Scientific"},
{0x0948, "Kronauer music in digital"},
{0x094b, "Linkup Systems Corp."},
{0x094d, "Cable Television Laboratories"},
{0x094f, "Yano"},
{0x0951, "Kingston Technology"},
{0x0954, "RPM Systems Corp."},
{0x0955, "NVidia Corp."},
{0x0956, "BSquare Corp."},
{0x0957, "Agilent Technologies, Inc."},
{0x0958, "CompuLink Research, Inc."},
{0x0959, "Cologne Chip AG"},
{0x095a, "Portsmith"},
{0x095b, "Medialogic Corp."},
{0x095c, "K-Tec Electronics"},
{0x095d, "Polycom, Inc."},
{0x0967, "Acer (??)"},
{0x0968, "Catalyst Enterprises, Inc."},
{0x096e, "Feitian Technologies, Inc."},
{0x0971, "Gretag-Macbeth AG"},
{0x0973, "Schlumberger"},
{0x0974, "Datagraphix, a business unit of Anacomp"},
{0x0975, "OL'E Communications, Inc."},
{0x0976, "Adirondack Wire & Cable"},
{0x0977, "Lightsurf Technologies"},
{0x0978, "Beckhoff GmbH"},
{0x0979, "Jeilin Technology Corp., Ltd"},
{0x097a, "Minds At Work LLC"},
{0x097b, "Knudsen Engineering, Ltd"},
{0x097c, "Marunix Co., Ltd"},
{0x097d, "Rosun Technologies, Inc."},
{0x097e, "Biopac Systems Inc."},
{0x097f, "Barun Electronics Co., Ltd"},
{0x0981, "Oak Technology, Ltd"},
{0x0984, "Apricorn"},
{0x0985, "cab Produkttechnik GmbH & Co KG"},
{0x0986, "Matsushita Electric Works, Ltd."},
{0x098c, "Vitana Corp."},
{0x098d, "INDesign"},
{0x098e, "Integrated Intellectual Property, Inc."},
{0x098f, "Kenwood TMI Corp."},
{0x0993, "Gemstar eBook Group, Ltd"},
{0x0996, "Integrated Telecom Express, Inc."},
{0x099a, "Zippy Technology Corp."},
{0x09a3, "PairGain Technologies"},
{0x09a4, "Contech Research, Inc."},
{0x09a5, "VCON Telecommunications"},
{0x09a6, "Poinchips"},
{0x09a7, "Data Transmission Network Corp."},
{0x09a8, "Lin Shiung Enterprise Co., Ltd"},
{0x09a9, "Smart Card Technologies Co., Ltd"},
{0x09aa, "Intersil Corp."},
{0x09ab, "Japan Cash Machine Co., Ltd."},
{0x09ae, "Tripp Lite"},
{0x09b2, "Franklin Electronic Publishers, Inc."},
{0x09b3, "Altius Solutions, Inc."},
{0x09b4, "MDS Telephone Systems"},
{0x09b5, "Celltrix Technology Co., Ltd"},
{0x09bc, "Grundig"},
{0x09be, "MySmart.Com"},
{0x09bf, "Auerswald GmbH & Co. KG"},
{0x09c0, "Genpix Electronics, LLC"},
{0x09c1, "Arris Interactive LLC"},
{0x09c2, "Nisca Corp."},
{0x09c3, "ActivCard, Inc."},
{0x09c4, "ACTiSYS Corp."},
{0x09c5, "Memory Corp."},
{0x09cc, "Workbit Corp."},
{0x09cd, "Psion Dacom Home Networks, Ltd"},
{0x09ce, "City Electronics, Ltd"},
{0x09cf, "Electronics Testing Center, Taiwan"},
{0x09d1, "NeoMagic, Inc."},
{0x09d2, "Vreelin Engineering, Inc."},
{0x09d3, "Com One"},
{0x09d7, "Novatel Wireless"},
{0x09d9, "KRF Tech, Ltd"},
{0x09da, "A4 Tech Co., Ltd"},
{0x09db, "Measurement Computing Corp."},
{0x09dc, "Aimex Corp."},
{0x09dd, "Fellowes, Inc."},
{0x09df, "Addonics Technologies Corp."},
{0x09e1, "Intellon Corp."},
{0x09e5, "Jo-Dan International, Inc."},
{0x09e6, "Silutia, Inc."},
{0x09e7, "Real 3D, Inc."},
{0x09e8, "AKAI  Professional M.I. Corp."},
{0x09e9, "Chen-Source, Inc."},
{0x09eb, "IM Networks, Inc."},
{0x09ef, "Xitel"},
{0x09f3, "GoFlight, Inc."},
{0x09f5, "AresCom"},
{0x09f6, "RocketChips, Inc."},
{0x09f7, "Edu-Science (H.K.), Ltd"},
{0x09f8, "SoftConnex Technologies, Inc."},
{0x09f9, "Bay Associates"},
{0x09fa, "Mtek Vision"},
{0x09fb, "Altera"},
{0x09ff, "Gain Technology Corp."},
{0x0a00, "Liquid Audio"},
{0x0a01, "ViA, Inc."},
{0x0a05, "Unknown Manufacturer"},
{0x0a07, "Ontrak Control Systems Inc."},
{0x0a0b, "Cybex Computer Products Co."},
{0x0a11, "Xentec, Inc."},
{0x0a12, "Cambridge Silicon Radio, Ltd"},
{0x0a13, "Telebyte, Inc."},
{0x0a14, "Spacelabs Medical, Inc."},
{0x0a15, "Scalar Corp."},
{0x0a16, "Trek Technology (S) PTE, Ltd"},
{0x0a17, "Pentax Corp."},
{0x0a18, "Heidelberger Druckmaschinen AG"},
{0x0a19, "Hua Geng Technologies, Inc."},
{0x0a21, "Medtronic Physio Control Corp."},
{0x0a22, "Century Semiconductor USA, Inc."},
{0x0a27, "Datacard Group"},
{0x0a2c, "AK-Modul-Bus Computer GmbH"},
{0x0a34, "TG3 Electronics, Inc."},
{0x0a35, "Radikal Technologies"},
{0x0a39, "Gilat Satellite Networks, Ltd"},
{0x0a3a, "PentaMedia Co., Ltd"},
{0x0a3c, "NTT DoCoMo, Inc."},
{0x0a3d, "Varo Vision"},
{0x0a3f, "Swissonic AG"},
{0x0a43, "Boca Systems, Inc."},
{0x0a46, "Davicom Semiconductor, Inc."},
{0x0a47, "Hirose Electric"},
{0x0a48, "I/O Interconnect"},
{0x0a4b, "Fujitsu Media Devices, Ltd"},
{0x0a4c, "Computex Co., Ltd"},
{0x0a4d, "Evolution Electronics, Ltd"},
{0x0a4e, "Steinberg Soft-und Hardware GmbH"},
{0x0a4f, "Litton Systems, Inc."},
{0x0a50, "Mimaki Engineering Co., Ltd"},
{0x0a51, "Sony Electronics, Inc."},
{0x0a52, "Jebsee Electronics Co., Ltd"},
{0x0a53, "Portable Peripheral Co., Ltd"},
{0x0a5a, "Electronics For Imaging, Inc."},
{0x0a5b, "EAsics NV"},
{0x0a5c, "Broadcom Corp."},
{0x0a5d, "Diatrend Corp."},
{0x0a5f, "Zebra"},
{0x0a62, "MPMan"},
{0x0a66, "ClearCube Technology"},
{0x0a67, "Medeli Electronics Co., Ltd"},
{0x0a68, "Comaide Corp."},
{0x0a69, "Chroma ate, Inc."},
{0x0a6b, "Green House Co., Ltd"},
{0x0a6c, "Integrated Circuit Systems, Inc."},
{0x0a6d, "UPS Manufacturing"},
{0x0a6e, "Benwin"},
{0x0a6f, "Core Technology, Inc."},
{0x0a70, "International Game Technology"},
{0x0a71, "VIPColor Technologies USA, Inc."},
{0x0a72, "Sanwa Denshi"},
{0x0a73, "Mackie Designs"},
{0x0a7d, "NSTL, Inc."},
{0x0a7e, "Octagon Systems Corp."},
{0x0a80, "Rexon Technology Corp., Ltd"},
{0x0a81, "Chesen Electronics Corp."},
{0x0a82, "Syscan"},
{0x0a83, "NextComm, Inc."},
{0x0a84, "Maui Innovative Peripherals"},
{0x0a85, "Idexx Labs"},
{0x0a86, "NITGen Co., Ltd"},
{0x0a8d, "Picturetel"},
{0x0a8e, "Japan Aviation Electronics Industry, Ltd"},
{0x0a90, "Candy Technology Co., Ltd"},
{0x0a91, "Globlink Technology, Inc."},
{0x0a92, "EGO SYStems, Inc."},
{0x0a93, "C Technologies AB"},
{0x0a94, "Intersense"},
{0x0aa3, "Lava Computer Mfg., Inc."},
{0x0aa4, "Develco Elektronik"},
{0x0aa5, "First International Digital"},
{0x0aa6, "Perception Digital, Ltd"},
{0x0aa7, "Wincor Nixdorf International GmbH"},
{0x0aa8, "TriGem Computer, Inc."},
{0x0aa9, "Baromtec Co."},
{0x0aaa, "Japan CBM Corp."},
{0x0aab, "Vision Shape Europe SA"},
{0x0aac, "iCompression, Inc."},
{0x0aad, "Rohde & Schwarz GmbH & Co. KG"},
{0x0aae, "NEC infrontia Corp. (Nitsuko)"},
{0x0aaf, "Digitalway Co., Ltd"},
{0x0ab0, "Arrow Strong Electronics Co., Ltd"},
{0x0ab1, "FEIG ELECTRONIC GmbH"},
{0x0aba, "Ellisys"},
{0x0abe, "Stereo-Link"},
{0x0abf, "Diolan"},
{0x0ac3, "Sanyo Semiconductor Company Micro"},
{0x0ac4, "Leco Corp."},
{0x0ac5, "I & C Corp."},
{0x0ac6, "Singing Electrons, Inc."},
{0x0ac7, "Panwest Corp."},
{0x0ac8, "Z-Star Microelectronics Corp."},
{0x0ac9, "Micro Solutions, Inc."},
{0x0aca, "OPEN Networks Ltd"},
{0x0acc, "Koga Electronics Co."},
{0x0acd, "ID Tech"},
{0x0ace, "ZyDAS"},
{0x0acf, "Intoto, Inc."},
{0x0ad0, "Intellix Corp."},
{0x0ad1, "Remotec Technology, Ltd"},
{0x0ad2, "Service & Quality Technology Co., Ltd"},
{0x0ada, "Data Encryption Systems Ltd."},
{0x0ae3, "Allion Computer, Inc."},
{0x0ae4, "Taito Corp."},
{0x0ae7, "Neodym Systems, Inc."},
{0x0ae8, "System Support Co., Ltd"},
{0x0ae9, "North Shore Circuit Design L.L.P."},
{0x0aea, "SciEssence, LLC"},
{0x0aeb, "TTP Communications, Ltd"},
{0x0aec, "Neodio Technologies Corp."},
{0x0af0, "Option"},
{0x0af6, "Silver I Co., Ltd"},
{0x0af7, "B2C2, Inc."},
{0x0af9, "Hama, Inc."},
{0x0afc, "Zaptronix Ltd"},
{0x0afd, "Tateno Dennou, Inc."},
{0x0afe, "Cummins Engine Co."},
{0x0aff, "Jump Zone Network Products, Inc."},
{0x0b00, "INGENICO"},
{0x0b05, "ASUSTek Computer, Inc."},
{0x0b0b, "Datamax-O'Neil"},
{0x0b0c, "Todos AB"},
{0x0b0d, "ProjectLab"},
{0x0b0e, "GN Netcom"},
{0x0b0f, "AVID Technology"},
{0x0b10, "Pcally"},
{0x0b11, "I Tech Solutions Co., Ltd"},
{0x0b1e, "Electronic Warfare Assoc., Inc. (EWA)"},
{0x0b1f, "Insyde Software Corp."},
{0x0b20, "TransDimension, Inc."},
{0x0b21, "Yokogawa Electric Corp."},
{0x0b22, "Japan System Development Co., Ltd"},
{0x0b23, "Pan-Asia Electronics Co., Ltd"},
{0x0b24, "Link Evolution Corp."},
{0x0b27, "Ritek Corp."},
{0x0b28, "Kenwood Corp."},
{0x0b2c, "Village Center, Inc."},
{0x0b30, "PNY Technologies, Inc."},
{0x0b33, "Contour Design, Inc."},
{0x0b37, "Hitachi ULSI Systems Co., Ltd"},
{0x0b38, "Gear Head"},
{0x0b39, "Omnidirectional Control Technology, Inc."},
{0x0b3a, "IPaxess"},
{0x0b3b, "Tekram Technology Co., Ltd"},
{0x0b3c, "Olivetti Techcenter"},
{0x0b3e, "Kikusui Electronics Corp."},
{0x0b41, "Hal Corp."},
{0x0b43, "Play.com, Inc."},
{0x0b47, "Sportbug.com, Inc."},
{0x0b48, "TechnoTrend AG"},
{0x0b49, "ASCII Corp."},
{0x0b4b, "Pine Corp. Ltd."},
{0x0b4d, "Graphtec America, Inc."},
{0x0b4e, "Musical Electronics, Ltd"},
{0x0b50, "Dumpries Co., Ltd"},
{0x0b51, "Comfort Keyboard Co."},
{0x0b52, "Colorado MicroDisplay, Inc."},
{0x0b54, "Sinbon Electronics Co., Ltd"},
{0x0b56, "TYI Systems, Ltd"},
{0x0b57, "Beijing HanwangTechnology Co., Ltd"},
{0x0b59, "Lake Communications, Ltd"},
{0x0b5a, "Corel Corp."},
{0x0b5f, "Green Electronics Co., Ltd"},
{0x0b60, "Nsine, Ltd"},
{0x0b61, "NEC Viewtechnology, Ltd"},
{0x0b62, "Orange Micro, Inc."},
{0x0b63, "ADLink Technology, Inc."},
{0x0b64, "Wonderful Wire Cable Co., Ltd"},
{0x0b65, "Expert Magnetics Corp."},
{0x0b66, "Cybiko Inc."},
{0x0b67, "Fairbanks Scales"},
{0x0b69, "CacheVision"},
{0x0b6a, "Maxim Integrated Products"},
{0x0b6f, "Nagano Japan Radio Co., Ltd"},
{0x0b70, "PortalPlayer, Inc."},
{0x0b71, "SHIN-EI Sangyo Co., Ltd"},
{0x0b72, "Embedded Wireless Technology Co., Ltd"},
{0x0b73, "Computone Corp."},
{0x0b75, "Roland DG Corp."},
{0x0b79, "Sunrise Telecom, Inc."},
{0x0b7a, "Zeevo, Inc."},
{0x0b7b, "Taiko Denki Co., Ltd"},
{0x0b7c, "ITRAN Communications, Ltd"},
{0x0b7d, "Astrodesign, Inc."},
{0x0b81, "id3 Semiconductors"},
{0x0b84, "Rextron Technology, Inc."},
{0x0b85, "Elkat Electronics, Sdn., Bhd."},
{0x0b86, "Exputer Systems, Inc."},
{0x0b87, "Plus-One I & T, Inc."},
{0x0b88, "Sigma Koki Co., Ltd, Technology Center"},
{0x0b89, "Advanced Digital Broadcast, Ltd"},
{0x0b8c, "SMART Technologies Inc."},
{0x0b95, "ASIX Electronics Corp."},
{0x0b96, "Sewon Telecom"},
{0x0b97, "O2 Micro, Inc."},
{0x0b98, "Playmates Toys, Inc."},
{0x0b99, "Audio International, Inc."},
{0x0b9b, "Dipl.-Ing. Stefan Kunde"},
{0x0b9d, "Softprotec Co."},
{0x0b9f, "Chippo Technologies"},
{0x0baf, "U.S. Robotics"},
{0x0bb0, "Concord Camera Corp."},
{0x0bb1, "Infinilink Corp."},
{0x0bb2, "Ambit Microsystems Corp."},
{0x0bb3, "Ofuji Technology"},
{0x0bb4, "HTC (High Tech Computer Corp.)"},
{0x0bb5, "Murata Manufacturing Co., Ltd"},
{0x0bb6, "Network Alchemy"},
{0x0bb7, "Joytech Computer Co., Ltd"},
{0x0bb8, "Hitachi Semiconductor and Devices Sales Co., Ltd"},
{0x0bb9, "Eiger M&C Co., Ltd"},
{0x0bba, "ZAccess Systems"},
{0x0bbb, "General Meters Corp."},
{0x0bbc, "Assistive Technology, Inc."},
{0x0bbd, "System Connection, Inc."},
{0x0bc0, "Knilink Technology, Inc."},
{0x0bc1, "Fuw Yng Electronics Co., Ltd"},
{0x0bc2, "Seagate RSS LLC"},
{0x0bc3, "IPWireless, Inc."},
{0x0bc4, "Microcube Corp."},
{0x0bc5, "JCN Co., Ltd"},
{0x0bc6, "ExWAY, Inc."},
{0x0bc7, "X10 Wireless Technology, Inc."},
{0x0bc8, "Telmax Communications"},
{0x0bc9, "ECI Telecom, Ltd"},
{0x0bca, "Startek Engineering, Inc."},
{0x0bcb, "Perfect Technic Enterprise Co., Ltd"},
{0x0bd7, "Andrew Pargeter & Associates"},
{0x0bda, "Realtek Semiconductor Corp."},
{0x0bdb, "Ericsson Business Mobile Networks BV"},
{0x0bdc, "Y Media Corp."},
{0x0bdd, "Orange PCS"},
{0x0be2, "Kanda Tsushin Kogyo Co., Ltd"},
{0x0be3, "TOYO Corp."},
{0x0be4, "Elka International, Ltd"},
{0x0be5, "DOME imaging systems, Inc."},
{0x0be6, "Dong Guan Humen Wonderful Wire Cable Factory"},
{0x0bed, "MEI"},
{0x0bee, "LTK Industries, Ltd"},
{0x0bef, "Way2Call Communications"},
{0x0bf0, "Pace Micro Technology PLC"},
{0x0bf1, "Intracom S.A."},
{0x0bf2, "Konexx"},
{0x0bf6, "Addonics Technologies, Inc."},
{0x0bf7, "Sunny Giken, Inc."},
{0x0bf8, "Fujitsu Siemens Computers"},
{0x0bfd, "Kvaser AB"},
{0x0c04, "MOTO Development Group, Inc."},
{0x0c05, "Appian Graphics"},
{0x0c06, "Hasbro Games, Inc."},
{0x0c07, "Infinite Data Storage, Ltd"},
{0x0c08, "Agate"},
{0x0c09, "Comjet Information System"},
{0x0c0a, "Highpoint Technologies, Inc."},
{0x0c0b, "Dura Micro, Inc. (Acomdata)"},
{0x0c12, "Zeroplus"},
{0x0c15, "Iris Graphics"},
{0x0c16, "Gyration, Inc."},
{0x0c17, "Cyberboard A/S"},
{0x0c18, "SynerTek Korea, Inc."},
{0x0c19, "cyberPIXIE, Inc."},
{0x0c1a, "Silicon Motion, Inc."},
{0x0c1b, "MIPS Technologies"},
{0x0c1c, "Hang Zhou Silan Electronics Co., Ltd"},
{0x0c22, "Tally Printer Corp."},
{0x0c23, "Lernout + Hauspie"},
{0x0c24, "Taiyo Yuden"},
{0x0c25, "Sampo Corp."},
{0x0c26, "Prolific Technology Inc."},
{0x0c27, "RFIDeas, Inc"},
{0x0c2e, "Metrologic Instruments"},
{0x0c35, "Eagletron, Inc."},
{0x0c36, "E Ink Corp."},
{0x0c37, "e.Digital"},
{0x0c38, "Der An Electric Wire & Cable Co., Ltd"},
{0x0c39, "IFR"},
{0x0c3a, "Furui Precise Component (Kunshan) Co., Ltd"},
{0x0c3b, "Komatsu, Ltd"},
{0x0c3c, "Radius Co., Ltd"},
{0x0c3d, "Innocom, Inc."},
{0x0c3e, "Nextcell, Inc."},
{0x0c44, "Motorola iDEN"},
{0x0c45, "Microdia"},
{0x0c46, "WaveRider Communications, Inc."},
{0x0c4a, "ALGE-TIMING GmbH"},
{0x0c4b, "Reiner SCT Kartensysteme GmbH"},
{0x0c4c, "Needham's Electronics"},
{0x0c52, "Sealevel Systems, Inc."},
{0x0c53, "ViewPLUS, Inc."},
{0x0c54, "Glory, Ltd"},
{0x0c55, "Spectrum Digital, Inc."},
{0x0c56, "Billion Bright, Ltd"},
{0x0c57, "Imaginative Design Operation Co., Ltd"},
{0x0c58, "Vidar Systems Corp."},
{0x0c59, "Dong Guan Shinko Wire Co., Ltd"},
{0x0c5a, "TRS International Mfg., Inc."},
{0x0c5e, "Xytronix Research & Design"},
{0x0c60, "Apogee Electronics Corp."},
{0x0c62, "Chant Sincere Co., Ltd"},
{0x0c63, "Toko, Inc."},
{0x0c64, "Signality System Engineering Co., Ltd"},
{0x0c65, "Eminence Enterprise Co., Ltd"},
{0x0c66, "Rexon Electronics Corp."},
{0x0c67, "Concept Telecom, Ltd"},
{0x0c6a, "ACS"},
{0x0c6c, "JETI Technische Instrumente GmbH"},
{0x0c70, "MCT Elektronikladen"},
{0x0c72, "PEAK System"},
{0x0c74, "Optronic Laboratories Inc."},
{0x0c76, "JMTek, LLC."},
{0x0c77, "Sipix Group, Ltd"},
{0x0c78, "Detto Corp."},
{0x0c79, "NuConnex Technologies Pte., Ltd"},
{0x0c7a, "Wing-Span Enterprise Co., Ltd"},
{0x0c86, "NDA Technologies, Inc."},
{0x0c88, "Kyocera Wireless Corp."},
{0x0c89, "Honda Tsushin Kogyo Co., Ltd"},
{0x0c8a, "Pathway Connectivity, Inc."},
{0x0c8b, "Wavefly Corp."},
{0x0c8c, "Coactive Networks"},
{0x0c8d, "Tempo"},
{0x0c8e, "Cesscom Co., Ltd"},
{0x0c8f, "Applied Microsystems"},
{0x0c94, "Cryptera"},
{0x0c98, "Berkshire Products, Inc."},
{0x0c99, "Innochips Co., Ltd"},
{0x0c9a, "Hanwool Robotics Corp."},
{0x0c9b, "Jobin Yvon, Inc."},
{0x0c9d, "SemTek"},
{0x0ca2, "Zyfer"},
{0x0ca3, "Sega Corp."},
{0x0ca4, "ST&T Instrument Corp."},
{0x0ca5, "BAE Systems Canada, Inc."},
{0x0ca6, "Castles Technology Co., Ltd"},
{0x0ca7, "Information Systems Laboratories"},
{0x0cad, "Motorola CGISS"},
{0x0cae, "Ascom Business Systems, Ltd"},
{0x0caf, "Buslink"},
{0x0cb0, "Flying Pig Systems"},
{0x0cb1, "Innovonics, Inc."},
{0x0cb6, "Celestix Networks, Pte., Ltd"},
{0x0cb7, "Singatron Enterprise Co., Ltd"},
{0x0cb8, "Opticis Co., Ltd"},
{0x0cba, "Trust Electronic (Shanghai) Co., Ltd"},
{0x0cbb, "Shanghai Darong Electronics Co., Ltd"},
{0x0cbc, "Palmax Technology Co., Ltd"},
{0x0cbd, "Pentel Co., Ltd (Electronics Equipment Div.)"},
{0x0cbe, "Keryx Technologies, Inc."},
{0x0cbf, "Union Genius Computer Co., Ltd"},
{0x0cc0, "Kuon Yi Industrial Corp."},
{0x0cc1, "Given Imaging, Ltd"},
{0x0cc2, "Timex Corp."},
{0x0cc3, "Rimage Corp."},
{0x0cc4, "emsys GmbH"},
{0x0cc5, "Sendo"},
{0x0cc6, "Intermagic Corp."},
{0x0cc7, "Kontron Medical AG"},
{0x0cc8, "Technotools Corp."},
{0x0cc9, "BroadMAX Technologies, Inc."},
{0x0cca, "Amphenol"},
{0x0ccb, "SKNet Co., Ltd"},
{0x0ccc, "Domex Technology Corp."},
{0x0ccd, "TerraTec Electronic GmbH"},
{0x0cd4, "Bang Olufsen"},
{0x0cd5, "LabJack Corporation"},
{0x0cd7, "NewChip S.r.l."},
{0x0cd8, "JS Digitech, Inc."},
{0x0cd9, "Hitachi Shin Din Cable, Ltd"},
{0x0cde, "Z-Com"},
{0x0ce5, "Validation Technologies International"},
{0x0ce9, "pico Technology"},
{0x0cf1, "e-Conn Electronic Co., Ltd"},
{0x0cf2, "ENE Technology, Inc."},
{0x0cf3, "Atheros Communications, Inc."},
{0x0cf4, "Fomtex Corp."},
{0x0cf5, "Cellink Co., Ltd"},
{0x0cf6, "Compucable Corp."},
{0x0cf7, "ishoni Networks"},
{0x0cf8, "Clarisys, Inc."},
{0x0cf9, "Central System Research Co., Ltd"},
{0x0cfa, "Inviso, Inc."},
{0x0cfc, "Minolta-QMS, Inc."},
{0x0cff, "SAFA MEDIA Co., Ltd."},
{0x0d06, "telos EDV Systementwicklung GmbH"},
{0x0d08, "UTStarcom"},
{0x0d0b, "Contemporary Controls"},
{0x0d0c, "Astron Electronics Co., Ltd"},
{0x0d0d, "MKNet Corp."},
{0x0d0e, "Hybrid Networks, Inc."},
{0x0d0f, "Feng Shin Cable Co., Ltd"},
{0x0d10, "Elastic Networks"},
{0x0d11, "Maspro Denkoh Corp."},
{0x0d12, "Hansol Electronics, Inc."},
{0x0d13, "BMF Corp."},
{0x0d14, "Array Comm, Inc."},
{0x0d15, "OnStream b.v."},
{0x0d16, "Hi-Touch Imaging Technologies Co., Ltd"},
{0x0d17, "NALTEC, Inc."},
{0x0d18, "coaXmedia"},
{0x0d19, "Hank Connection Industrial Co., Ltd"},
{0x0d28, "NXP"},
{0x0d32, "Leo Hui Electric Wire & Cable Co., Ltd"},
{0x0d33, "AirSpeak, Inc."},
{0x0d34, "Rearden Steel Technologies"},
{0x0d35, "Dah Kun Co., Ltd"},
{0x0d3a, "Posiflex Technologies, Inc."},
{0x0d3c, "Sri Cable Technology, Ltd"},
{0x0d3d, "Tangtop Technology Co., Ltd"},
{0x0d3e, "Fitcom, inc."},
{0x0d3f, "MTS Systems Corp."},
{0x0d40, "Ascor, Inc."},
{0x0d41, "Ta Yun Terminals Industrial Co., Ltd"},
{0x0d42, "Full Der Co., Ltd"},
{0x0d46, "Kobil Systems GmbH"},
{0x0d48, "Promethean Limited"},
{0x0d49, "Maxtor"},
{0x0d4a, "NF Corp."},
{0x0d4b, "Grape Systems, Inc."},
{0x0d4c, "Tedas AG"},
{0x0d4d, "Coherent, Inc."},
{0x0d4e, "Agere Systems Netherland BV"},
{0x0d4f, "EADS Airbus France"},
{0x0d50, "Cleware GmbH"},
{0x0d51, "Volex (Asia) Pte., Ltd"},
{0x0d53, "HMI Co., Ltd"},
{0x0d54, "Holon Corp."},
{0x0d55, "ASKA Technologies, Inc."},
{0x0d56, "AVLAB Technology, Inc."},
{0x0d57, "Solomon Microtech, Ltd"},
{0x0d5c, "SMC Networks, Inc."},
{0x0d5e, "Myacom, Ltd"},
{0x0d5f, "CSI, Inc."},
{0x0d60, "IVL Technologies, Ltd"},
{0x0d61, "Meilu Electronics (Shenzhen) Co., Ltd"},
{0x0d62, "Darfon Electronics Corp."},
{0x0d63, "Fritz Gegauf AG"},
{0x0d64, "DXG Technology Corp."},
{0x0d65, "KMJP Co., Ltd"},
{0x0d66, "TMT"},
{0x0d67, "Advanet, Inc."},
{0x0d68, "Super Link Electronics Co., Ltd"},
{0x0d69, "NSI"},
{0x0d6a, "Megapower International Corp."},
{0x0d6b, "And-Or Logic"},
{0x0d70, "Try Computer Co., Ltd"},
{0x0d71, "Hirakawa Hewtech Corp."},
{0x0d72, "Winmate Communication, Inc."},
{0x0d73, "Hit's Communications, Inc."},
{0x0d76, "MFP Korea, Inc."},
{0x0d77, "Power Sentry/Newpoint"},
{0x0d78, "Japan Distributor Corp."},
{0x0d7a, "MARX Datentechnik GmbH"},
{0x0d7b, "Wellco Technology Co., Ltd"},
{0x0d7c, "Taiwan Line Tek Electronic Co., Ltd"},
{0x0d7d, "Phison Electronics Corp."},
{0x0d7e, "American Computer & Digital Components"},
{0x0d7f, "Essential Reality LLC"},
{0x0d80, "H.R. Silvine Electronics, Inc."},
{0x0d81, "TechnoVision"},
{0x0d83, "Think Outside, Inc."},
{0x0d87, "Dolby Laboratories Inc."},
{0x0d89, "Oz Software"},
{0x0d8a, "King Jim Co., Ltd"},
{0x0d8b, "Ascom Telecommunications, Ltd"},
{0x0d8c, "C-Media Electronics, Inc."},
{0x0d8d, "Promotion & Display Technology, Ltd"},
{0x0d8e, "Global Sun Technology, Inc."},
{0x0d8f, "Pitney Bowes"},
{0x0d90, "Sure-Fire Electrical Corp."},
{0x0d96, "Skanhex Technology, Inc."},
{0x0d97, "Santa Barbara Instrument Group"},
{0x0d98, "Mars Semiconductor Corp."},
{0x0d99, "Trazer Technologies, Inc."},
{0x0d9a, "RTX Telecom AS"},
{0x0d9b, "Tat Shing Electrical Co."},
{0x0d9c, "Chee Chen Hi-Technology Co., Ltd"},
{0x0d9d, "Sanwa Supply, Inc."},
{0x0d9e, "Avaya"},
{0x0d9f, "Powercom Co., Ltd"},
{0x0da0, "Danger Research"},
{0x0da1, "Suzhou Peter's Precise Industrial Co., Ltd"},
{0x0da2, "Land Instruments International, Ltd"},
{0x0da3, "Nippon Electro-Sensory Devices Corp."},
{0x0da4, "Polar Electro OY"},
{0x0da7, "IOGear, Inc."},
{0x0da8, "softDSP Co., Ltd"},
{0x0dab, "Cubig Group"},
{0x0dad, "Westover Scientific"},
{0x0db0, "Micro Star International"},
{0x0db1, "Wen Te Electronics Co., Ltd"},
{0x0db2, "Shian Hwi Plug Parts, Plastic Factory"},
{0x0db3, "Tekram Technology Co., Ltd"},
{0x0db4, "Chung Fu Chen Yeh Enterprise Corp."},
{0x0db7, "ELCON Systemtechnik"},
{0x0dbc, "A&D Medical"},
{0x0dbe, "Jiuh Shiuh Precision Industry Co., Ltd"},
{0x0dbf, "Jess-Link International"},
{0x0dc0, "G7 Solutions (formerly Great Notions)"},
{0x0dc1, "Tamagawa Seiki Co., Ltd"},
{0x0dc3, "Athena Smartcard Solutions, Inc."},
{0x0dc4, "Macpower Peripherals, Ltd"},
{0x0dc5, "SDK Co., Ltd"},
{0x0dc6, "Precision Squared Technology Corp."},
{0x0dc7, "First Cable Line, Inc."},
{0x0dcd, "NetworkFab Corp."},
{0x0dd0, "Access Solutions"},
{0x0dd1, "Contek Electronics Co., Ltd"},
{0x0dd2, "Power Quotient International Co., Ltd"},
{0x0dd3, "MediaQ"},
{0x0dd4, "Custom Engineering SPA"},
{0x0dd5, "California Micro Devices"},
{0x0dd7, "Kocom Co., Ltd"},
{0x0dd8, "Netac Technology Co., Ltd"},
{0x0dd9, "HighSpeed Surfing"},
{0x0dda, "Integrated Circuit Solution, Inc."},
{0x0ddb, "Tamarack, Inc."},
{0x0ddd, "Datelink Technology Co., Ltd"},
{0x0dde, "Ubicom, Inc."},
{0x0de0, "BD Consumer Healthcare"},
{0x0de7, "USBmicro"},
{0x0dea, "UTECH Electronic (D.G.) Co., Ltd."},
{0x0ded, "Novasonics"},
{0x0dee, "Lifetime Memory Products"},
{0x0def, "Full Rise Electronic Co., Ltd"},
{0x0df4, "NET&SYS"},
{0x0df6, "Sitecom Europe B.V."},
{0x0df7, "Mobile Action Technology, Inc."},
{0x0dfa, "Toyo Communication Equipment Co., Ltd"},
{0x0dfc, "GeneralTouch Technology Co., Ltd"},
{0x0e03, "Nippon Systemware Co., Ltd"},
{0x0e08, "Winbest Technology Co., Ltd"},
{0x0e0b, "Amigo Technology Inc."},
{0x0e0c, "Gesytec"},
{0x0e0f, "VMware, Inc."},
{0x0e16, "JMTek, LLC"},
{0x0e17, "Walex Electronic, Ltd"},
{0x0e1a, "Unisys"},
{0x0e1b, "Crewave"},
{0x0e20, "Pegasus Technologies Ltd."},
{0x0e21, "Cowon Systems, Inc."},
{0x0e22, "Symbian Ltd."},
{0x0e23, "Liou Yuane Enterprise Co., Ltd"},
{0x0e25, "VinChip Systems, Inc."},
{0x0e26, "J-Phone East Co., Ltd"},
{0x0e30, "HeartMath LLC"},
{0x0e34, "Micro Computer Control Corp."},
{0x0e35, "3Pea Technologies, Inc."},
{0x0e36, "TiePie engineering"},
{0x0e38, "Stratitec, Inc."},
{0x0e39, "Smart Modular Technologies, Inc."},
{0x0e3a, "Neostar Technology Co., Ltd"},
{0x0e3b, "Mansella, Ltd"},
{0x0e41, "Line6, Inc."},
{0x0e44, "Sun-Riseful Technology Co., Ltd."},
{0x0e48, "Julia Corp., Ltd"},
{0x0e4a, "Shenzhen Bao Hing Electric Wire & Cable Mfr. Co."},
{0x0e4c, "Radica Games, Ltd"},
{0x0e50, "TechnoData Interware"},
{0x0e55, "Speed Dragon Multimedia, Ltd"},
{0x0e56, "Kingston Technology Company, Inc."},
{0x0e5a, "Active Co., Ltd"},
{0x0e5b, "Union Power Information Industrial Co., Ltd"},
{0x0e5c, "Bitland Information Technology Co., Ltd"},
{0x0e5d, "Neltron Industrial Co., Ltd"},
{0x0e5e, "Conwise Technology Co., Ltd."},
{0x0e66, "Hawking Technologies"},
{0x0e67, "Fossil, Inc."},
{0x0e6a, "Megawin Technology Co., Ltd"},
{0x0e6f, "Logic3"},
{0x0e70, "Tokyo Electronic Industry Co., Ltd"},
{0x0e72, "Hsi-Chin Electronics Co., Ltd"},
{0x0e75, "TVS Electronics, Ltd"},
{0x0e79, "Archos, Inc."},
{0x0e7b, "On-Tech Industry Co., Ltd"},
{0x0e7e, "Gmate, Inc."},
{0x0e82, "Ching Tai Electric Wire & Cable Co., Ltd"},
{0x0e83, "Shin An Wire & Cable Co."},
{0x0e8c, "Well Force Electronic Co., Ltd"},
{0x0e8d, "MediaTek Inc."},
{0x0e8f, "GreenAsia Inc."},
{0x0e90, "WiebeTech, LLC"},
{0x0e91, "VTech Engineering Canada, Ltd"},
{0x0e92, "C's Glory Enterprise Co., Ltd"},
{0x0e93, "eM Technics Co., Ltd"},
{0x0e95, "Future Technology Co., Ltd"},
{0x0e96, "Aplux Communications, Ltd"},
{0x0e97, "Fingerworks, Inc."},
{0x0e98, "Advanced Analogic Technologies, Inc."},
{0x0e99, "Parallel Dice Co., Ltd"},
{0x0e9a, "TA HSING Industries, Ltd"},
{0x0e9b, "ADTEC Corp."},
{0x0e9c, "Streamzap, Inc."},
{0x0e9f, "Tamura Corp."},
{0x0ea0, "Ours Technology, Inc."},
{0x0ea6, "Nihon Computer Co., Ltd"},
{0x0ea7, "MSL Enterprises Corp."},
{0x0ea8, "CenDyne, Inc."},
{0x0ead, "Humax Co., Ltd"},
{0x0eb0, "NovaTech"},
{0x0eb1, "WIS Technologies, Inc."},
{0x0eb2, "Y-S Electronic Co., Ltd"},
{0x0eb3, "Saint Technology Corp."},
{0x0eb7, "Endor AG"},
{0x0eb8, "Mettler Toledo"},
{0x0ebe, "VWeb Corp."},
{0x0ebf, "Omega Technology of Taiwan, Inc."},
{0x0ec0, "LHI Technology (China) Co., Ltd"},
{0x0ec1, "Abit Computer Corp."},
{0x0ec2, "Sweetray Industrial, Ltd"},
{0x0ec3, "Axell Co., Ltd"},
{0x0ec4, "Ballracing Developments, Ltd"},
{0x0ec5, "GT Information System Co., Ltd"},
{0x0ec6, "InnoVISION Multimedia, Ltd"},
{0x0ec7, "Theta Link Corp."},
{0x0ecd, "Lite-On IT Corp."},
{0x0ece, "TaiSol Electronics Co., Ltd"},
{0x0ecf, "Phogenix Imaging, LLC"},
{0x0ed1, "WinMaxGroup"},
{0x0ed2, "Kyoto Micro Computer Co., Ltd"},
{0x0ed3, "Wing-Tech Enterprise Co., Ltd"},
{0x0ed5, "Fiberbyte"},
{0x0eda, "Noriake Itron Corp."},
{0x0edf, "e-MDT Co., Ltd"},
{0x0ee0, "Shima Seiki Mfg., Ltd"},
{0x0ee1, "Sarotech Co., Ltd"},
{0x0ee2, "AMI Semiconductor, Inc."},
{0x0ee3, "ComTrue Technology Corp."},
{0x0ee4, "Sunrich Technology, Ltd"},
{0x0eee, "Digital Stream Technology, Inc."},
{0x0eef, "D-WAV Scientific Co., Ltd"},
{0x0ef0, "Hitachi Cable, Ltd"},
{0x0ef1, "Aichi Micro Intelligent Corp."},
{0x0ef2, "I/O Magic Corp."},
{0x0ef3, "Lynn Products, Inc."},
{0x0ef4, "DSI Datotech"},
{0x0ef5, "PointChips"},
{0x0ef6, "Yield Microelectronics Corp."},
{0x0ef7, "SM Tech Co., Ltd (Tulip)"},
{0x0efd, "Oasis Semiconductor"},
{0x0efe, "Wem Technology, Inc."},
{0x0f03, "Unitek UPS Systems"},
{0x0f06, "Visual Frontier Enterprise Co., Ltd"},
{0x0f08, "CSL Wire & Plug (Shen Zhen) Co."},
{0x0f0c, "CAS Corp."},
{0x0f0d, "Hori Co., Ltd"},
{0x0f0e, "Energy Full Corp."},
{0x0f11, "LD Didactic GmbH"},
{0x0f12, "Mars Engineering Corp."},
{0x0f13, "Acetek Technology Co., Ltd"},
{0x0f18, "Finger Lakes Instrumentation"},
{0x0f19, "Oracom Co., Ltd"},
{0x0f1b, "Onset Computer Corp."},
{0x0f1c, "Funai Electric Co., Ltd"},
{0x0f1d, "Iwill Corp."},
{0x0f21, "IOI Technology Corp."},
{0x0f22, "Senior Industries, Inc."},
{0x0f23, "Leader Tech Manufacturer Co., Ltd"},
{0x0f24, "Flex-P Industries, Snd., Bhd."},
{0x0f2d, "ViPower, Inc."},
{0x0f2e, "Geniality Maple Technology Co., Ltd"},
{0x0f2f, "Priva Design Services"},
{0x0f30, "Jess Technology Co., Ltd"},
{0x0f31, "Chrysalis Development"},
{0x0f32, "YFC-BonEagle Electric Co., Ltd"},
{0x0f37, "Kokuyo Co., Ltd"},
{0x0f38, "Nien-Yi Industrial Corp."},
{0x0f3d, "Airprime, Incorporated"},
{0x0f41, "RDC Semiconductor Co., Ltd"},
{0x0f42, "Nital Consulting Services, Inc."},
{0x0f44, "Polhemus"},
{0x0f4b, "St. John Technology Co., Ltd"},
{0x0f4c, "WorldWide Cable Opto Corp."},
{0x0f4d, "Microtune, Inc."},
{0x0f4e, "Freedom Scientific"},
{0x0f52, "Wing Key Electrical Co., Ltd"},
{0x0f53, "Dongguan White Horse Cable Factory, Ltd"},
{0x0f54, "Kawai Musical Instruments Mfg. Co., Ltd"},
{0x0f55, "AmbiCom, Inc."},
{0x0f5c, "Prairiecomm, Inc."},
{0x0f5d, "NewAge International, LLC"},
{0x0f5f, "Key Technology Corp."},
{0x0f60, "NTK, Ltd"},
{0x0f61, "Varian, Inc."},
{0x0f62, "Acrox Technologies Co., Ltd"},
{0x0f63, "LeapFrog Enterprises"},
{0x0f68, "Kobe Steel, Ltd"},
{0x0f69, "Dionex Corp."},
{0x0f6a, "Vibren Technologies, Inc."},
{0x0f6e, "INTELLIGENT SYSTEMS"},
{0x0f73, "DFI"},
{0x0f7c, "DQ Technology, Inc."},
{0x0f7d, "NetBotz, Inc."},
{0x0f7e, "Fluke Corp."},
{0x0f88, "VTech Holdings, Ltd"},
{0x0f8b, "Yazaki Corp."},
{0x0f8c, "Young Generation International Corp."},
{0x0f8d, "Uniwill Computer Corp."},
{0x0f8e, "Kingnet Technology Co., Ltd"},
{0x0f8f, "Soma Networks"},
{0x0f97, "CviLux Corp."},
{0x0f98, "CyberBank Corp."},
{0x0f9c, "Hyun Won, Inc."},
{0x0f9e, "Lucent Technologies"},
{0x0fa3, "Starconn Electronic Co., Ltd"},
{0x0fa4, "ATL Technology"},
{0x0fa5, "Sotec Co., Ltd"},
{0x0fa7, "Epox Computer Co., Ltd"},
{0x0fa8, "Logic Controls, Inc."},
{0x0faf, "Winpoint Electronic Corp."},
{0x0fb0, "Haurtian Wire & Cable Co., Ltd"},
{0x0fb1, "Inclose Design, Inc."},
{0x0fb2, "Juan-Chern Industrial Co., Ltd"},
{0x0fb6, "Heber Ltd"},
{0x0fb8, "Wistron Corp."},
{0x0fb9, "AACom Corp."},
{0x0fba, "San Shing Electronics Co., Ltd"},
{0x0fbb, "Bitwise Systems, Inc."},
{0x0fc1, "Mitac Internatinal Corp."},
{0x0fc2, "Plug and Jack Industrial, Inc."},
{0x0fc5, "Delcom Engineering"},
{0x0fc6, "Dataplus Supplies, Inc."},
{0x0fca, "Research In Motion, Ltd."},
{0x0fce, "Sony Ericsson Mobile Communications AB"},
{0x0fcf, "Dynastream Innovations, Inc."},
{0x0fd0, "Tulip Computers B.V."},
{0x0fd1, "Giant Electronics Ltd."},
{0x0fd2, "Seac Banche"},
{0x0fd4, "Tenovis GmbH & Co., KG"},
{0x0fd5, "Direct Access Technology, Inc."},
{0x0fd9, "Elgato Systems GmbH"},
{0x0fda, "Quantec Networks GmbH"},
{0x0fdc, "Micro Plus"},
{0x0fe0, "Osterhout Design Group"},
{0x0fe4, "IN-Tech Electronics, Ltd"},
{0x0fe5, "Greenconn (U.S.A.), Inc."},
{0x0fe6, "Kontron (Industrial Computer Source / ICS Advent)"},
{0x0fe9, "DVICO"},
{0x0fea, "United Computer Accessories"},
{0x0feb, "CRS Electronic Co., Ltd"},
{0x0fec, "UMC Electronics Co., Ltd"},
{0x0fed, "Access Co., Ltd"},
{0x0fee, "Xsido Corp."},
{0x0fef, "MJ Research, Inc."},
{0x0ff6, "Core Valley Co., Ltd"},
{0x0ff7, "CHI SHING Computer Accessories Co., Ltd"},
{0x0ffc, "Clavia DMI AB"},
{0x0fff, "Aopen, Inc."},
{0x1000, "Speed Tech Corp."},
{0x1001, "Ritronics Components (S) Pte., Ltd"},
{0x1003, "Sigma Corp."},
{0x1004, "LG Electronics, Inc."},
{0x1005, "Apacer Technology, Inc."},
{0x1006, "iRiver, Ltd."},
{0x1009, "Emuzed, Inc."},
{0x100a, "AV Chaseway, Ltd"},
{0x100b, "Chou Chin Industrial Co., Ltd"},
{0x100d, "Netopia, Inc."},
{0x1010, "Fukuda Denshi Co., Ltd"},
{0x1011, "Mobile Media Tech."},
{0x1012, "SDKM Fibres, Wires & Cables Berhad"},
{0x1013, "TST-Touchless Sensor Technology AG"},
{0x1014, "Densitron Technologies PLC"},
{0x1015, "Softronics Pty., Ltd"},
{0x1016, "Xiamen Hung's Enterprise Co., Ltd"},
{0x1017, "Speedy Industrial Supplies, Pte., Ltd"},
{0x1019, "Elitegroup Computer Systems (ECS)"},
{0x1020, "Labtec"},
{0x1022, "Shinko Shoji Co., Ltd"},
{0x1025, "Hyper-Paltek"},
{0x1026, "Newly Corp."},
{0x1027, "Time Domain"},
{0x1028, "Inovys Corp."},
{0x1029, "Atlantic Coast Telesys"},
{0x102a, "Ramos Technology Co., Ltd"},
{0x102b, "Infotronic America, Inc."},
{0x102c, "Etoms Electronics Corp."},
{0x102d, "Winic Corp."},
{0x1031, "Comax Technology, Inc."},
{0x1032, "C-One Technology Corp."},
{0x1033, "Nucam Corp."},
{0x1038, "Ideazon, Inc."},
{0x1039, "devolo AG"},
{0x103d, "Stanton"},
{0x1043, "iCreate Technologies Corp."},
{0x1044, "Chu Yuen Enterprise Co., Ltd"},
{0x1046, "Winbond Electronics Corp. [hex]"},
{0x1048, "Targus Group International"},
{0x104b, "Mylex / Buslogic"},
{0x104c, "AMCO TEC International, Inc."},
{0x104d, "Newport Corporation"},
{0x104f, "WB Electronics"},
{0x1050, "Yubico.com"},
{0x1053, "Immanuel Electronics Co., Ltd"},
{0x1054, "BMS International Beheer N.V."},
{0x1055, "Complex Micro Interconnection Co., Ltd"},
{0x1056, "Hsin Chen Ent Co., Ltd"},
{0x1057, "ON Semiconductor"},
{0x1058, "Western Digital Technologies, Inc."},
{0x1059, "Giesecke & Devrient GmbH"},
{0x105c, "Hong Ji Electric Wire & Cable (Dongguan) Co., Ltd"},
{0x105d, "Delkin Devices, Inc."},
{0x105e, "Valence Semiconductor Design, Ltd"},
{0x105f, "Chin Shong Enterprise Co., Ltd"},
{0x1060, "Easthome Industrial Co., Ltd"},
{0x1063, "Motorola Electronics Taiwan, Ltd [hex]"},
{0x1065, "CCYU Technology"},
{0x106a, "Loyal Legend, Ltd"},
{0x106c, "Curitel Communications, Inc."},
{0x106d, "San Chieh Manufacturing, Ltd"},
{0x106e, "ConectL"},
{0x106f, "Money Controls"},
{0x1076, "GCT Semiconductor, Inc."},
{0x107b, "Gateway, Inc."},
{0x107d, "Arlec Australia, Ltd"},
{0x107e, "Midoriya Electric Co., Ltd"},
{0x107f, "KidzMouse, Inc."},
{0x1082, "Shin-Etsukaken Co., Ltd"},
{0x1083, "Canon Electronics, Inc."},
{0x1084, "Pantech Co., Ltd"},
{0x108a, "Chloride Power Protection"},
{0x108b, "Grand-tek Technology Co., Ltd"},
{0x108c, "Robert Bosch GmbH"},
{0x108e, "Lotes Co., Ltd."},
{0x1099, "Surface Optics Corp."},
{0x109a, "DATASOFT Systems GmbH"},
{0x109f, "eSOL Co., Ltd"},
{0x10a0, "Hirotech, Inc."},
{0x10a3, "Mitsubishi Materials Corp."},
{0x10a9, "SK Teletech Co., Ltd"},
{0x10aa, "Cables To Go"},
{0x10ab, "USI Co., Ltd"},
{0x10ac, "Honeywell, Inc."},
{0x10ae, "Princeton Technology Corp."},
{0x10af, "Liebert Corp."},
{0x10b5, "Comodo (PLX?)"},
{0x10b8, "DiBcom"},
{0x10bb, "TM Technology, Inc."},
{0x10bc, "Dinging Technology Co., Ltd"},
{0x10bd, "TMT Technology, Inc."},
{0x10bf, "SmartHome"},
{0x10c4, "Cygnal Integrated Products, Inc."},
{0x10c5, "Sanei Electric, Inc."},
{0x10c6, "Intec, Inc."},
{0x10cb, "Eratech"},
{0x10cc, "GBM Connector Co., Ltd"},
{0x10cd, "Kycon, Inc."},
{0x10ce, "Silicon Labs"},
{0x10cf, "Velleman Components, Inc."},
{0x10d1, "Hottinger Baldwin Measurement"},
{0x10d4, "Man Boon Manufactory, Ltd"},
{0x10d5, "Uni Class Technology Co., Ltd"},
{0x10d6, "Actions Semiconductor Co., Ltd"},
{0x10de, "Authenex, Inc."},
{0x10df, "In-Win Development, Inc."},
{0x10e0, "Post-Op Video, Inc."},
{0x10e1, "CablePlus, Ltd"},
{0x10e2, "Nada Electronics, Ltd"},
{0x10ec, "Vast Technologies, Inc."},
{0x10f0, "Nexio Co., Ltd"},
{0x10f1, "Importek"},
{0x10f5, "Turtle Beach"},
{0x10fb, "Pictos Technologies, Inc."},
{0x10fd, "Anubis Electronics, Ltd"},
{0x10fe, "Thrane & Thrane"},
{0x1100, "VirTouch, Ltd"},
{0x1101, "EasyPass Industrial Co., Ltd"},
{0x1108, "Brightcom Technologies, Ltd"},
{0x110a, "Moxa Technologies Co., Ltd."},
{0x1110, "Analog Devices Canada, Ltd (Allied Telesyn)"},
{0x1111, "Pandora International Ltd."},
{0x1112, "YM ELECTRIC CO., Ltd"},
{0x1113, "Medion AG"},
{0x111e, "VSO Electric Co., Ltd"},
{0x112a, "RedRat"},
{0x112e, "Master Hill Electric Wire and Cable Co., Ltd"},
{0x112f, "Cellon International, Inc."},
{0x1130, "Tenx Technology, Inc."},
{0x1131, "Integrated System Solution Corp."},
{0x1132, "Toshiba Corp., Digital Media Equipment [hex]"},
{0x1136, "CTS Electronincs"},
{0x113c, "Arin Tech Co., Ltd"},
{0x113d, "Mapower Electronics Co., Ltd"},
{0x1141, "V One Multimedia, Pte., Ltd"},
{0x1142, "CyberScan Technologies, Inc."},
{0x1145, "Japan Radio Company"},
{0x1146, "Shimane SANYO Electric Co., Ltd."},
{0x1147, "Ever Great Electric Wire and Cable Co., Ltd"},
{0x114b, "Sphairon Access Systems GmbH"},
{0x114c, "Tinius Olsen Testing Machine Co., Inc."},
{0x114d, "Alpha Imaging Technology Corp."},
{0x114f, "Wavecom"},
{0x115b, "Salix Technology Co., Ltd."},
{0x1162, "Secugen Corp."},
{0x1163, "DeLorme Publishing, Inc."},
{0x1164, "YUAN High-Tech Development Co., Ltd"},
{0x1165, "Telson Electronics Co., Ltd"},
{0x1166, "Bantam Interactive Technologies"},
{0x1167, "Salient Systems Corp."},
{0x1168, "BizConn International Corp."},
{0x116e, "Gigastorage Corp."},
{0x116f, "Silicon 10 Technology Corp."},
{0x1175, "Shengyih Steel Mold Co., Ltd"},
{0x117d, "Santa Electronic, Inc."},
{0x117e, "JNC, Inc."},
{0x1182, "Venture Corp., Ltd"},
{0x1183, "Compaq Computer Corp. [hex] (Digital Dream ??)"},
{0x1184, "Kyocera Elco Corp."},
{0x1188, "Bloomberg L.P."},
{0x1189, "Acer Communications & Multimedia"},
{0x118f, "You Yang Technology Co., Ltd"},
{0x1190, "Tripace"},
{0x1191, "Loyalty Founder Enterprise Co., Ltd"},
{0x1196, "Yankee Robotics, LLC"},
{0x1197, "Technoimagia Co., Ltd"},
{0x1198, "StarShine Technology Corp."},
{0x1199, "Sierra Wireless, Inc."},
{0x119a, "ZHAN QI Technology Co., Ltd"},
{0x119b, "ruwido austria GmbH"},
{0x11a0, "Chipcon AS"},
{0x11a3, "Technovas Co., Ltd"},
{0x11aa, "GlobalMedia Group, LLC"},
{0x11ab, "Exito Electronics Co., Ltd"},
{0x11ac, "Nike"},
{0x11b0, "ATECH FLASH TECHNOLOGY"},
{0x11c5, "Inmax"},
{0x11db, "Topfield Co., Ltd."},
{0x11e6, "K.I. Technology Co. Ltd."},
{0x11f5, "Siemens AG"},
{0x11f6, "Prolific"},
{0x11f7, "Alcatel (?)"},
{0x1203, "TSC Auto ID Technology Co., Ltd"},
{0x1209, "InterBiometrics"},
{0x120e, "Hudson Soft Co., Ltd"},
{0x120f, "Magellan"},
{0x1210, "DigiTech"},
{0x121e, "Jungsoft Co., Ltd"},
{0x1223, "SKYCABLE ENTERPRISE. CO., LTD."},
{0x1228, "Datapaq Limited"},
{0x1230, "Chipidea-Microelectronica, S.A."},
{0x1233, "Denver Electronics"},
{0x1234, "Brain Actuated Technologies"},
{0x1235, "Novation EMS"},
{0x1241, "Belkin"},
{0x124a, "AirVast"},
{0x124b, "Nyko (Honey Bee)"},
{0x124c, "MXI - Memory Experts International, Inc."},
{0x125c, "Apogee Inc."},
{0x125f, "A-DATA Technology Co., Ltd."},
{0x1260, "Standard Microsystems Corp."},
{0x1264, "Covidien Energy-based Devices"},
{0x1266, "Pirelli Broadband Solutions"},
{0x1267, "Logic3 / SpectraVideo plc"},
{0x126c, "Aristocrat Technologies"},
{0x126d, "Bel Stewart"},
{0x126e, "Strobe Data, Inc."},
{0x126f, "TwinMOS"},
{0x1274, "Ensoniq"},
{0x1275, "Xaxero Marine Software Engineering, Ltd."},
{0x1278, "Starlight Xpress"},
{0x1283, "zebris Medical GmbH"},
{0x1286, "Marvell Semiconductor, Inc."},
{0x1291, "Qualcomm Flarion Technologies, Inc. / Leadtek Research, Inc."},
{0x1292, "Innomedia"},
{0x1293, "Belkin Components [hex]"},
{0x1294, "RISO KAGAKU CORP."},
{0x129b, "CyberTAN Technology"},
{0x12a7, "Trendchip Technologies Corp."},
{0x12ab, "Honey Bee Electronic International Ltd."},
{0x12b8, "Zhejiang Xinya Electronic Technology Co., Ltd."},
{0x12b9, "E28"},
{0x12ba, "Licensed by Sony Computer Entertainment America"},
{0x12c4, "Autocue Group Ltd"},
{0x12d1, "Huawei Technologies Co., Ltd."},
{0x12d2, "LINE TECH INDUSTRIAL CO., LTD."},
{0x12d6, "EMS Dr. Thomas Wuensche"},
{0x12d7, "BETTER WIRE FACTORY CO., LTD."},
{0x12e6, "Waldorf Music GmbH"},
{0x12ef, "Tapwave, Inc."},
{0x12f5, "Dynamic System Electronics Corp."},
{0x12f7, "Memorex Products, Inc."},
{0x12fd, "AIN Comm. Technology Co., Ltd"},
{0x12ff, "Fascinating Electronics, Inc."},
{0x1307, "Transcend Information, Inc."},
{0x1308, "Shuttle, Inc."},
{0x1310, "Roper"},
{0x1312, "ICS Electronics"},
{0x1313, "ThorLabs"},
{0x131d, "Natural Point"},
{0x132a, "Envara Inc."},
{0x132b, "Konica Minolta"},
{0x1342, "Mobility"},
{0x1345, "Sino Lite Technology Corp."},
{0x1347, "Moravian Instruments"},
{0x1348, "Katsuragawa Electric Co., Ltd."},
{0x134c, "PanJit International Inc."},
{0x134e, "Digby's Bitpile, Inc. DBA D Bit"},
{0x1357, "P&E Microcomputer Systems"},
{0x1366, "SEGGER"},
{0x136b, "STEC"},
{0x1370, "Swissbit"},
{0x1371, "CNet Technology Inc."},
{0x1376, "Vimtron Electronics Co., Ltd."},
{0x137b, "SCAPS GmbH"},
{0x1385, "Netgear, Inc"},
{0x138a, "Validity Sensors, Inc."},
{0x138e, "Jungo LTD"},
{0x1390, "TOMTOM B.V."},
{0x1391, "IdealTEK, Inc."},
{0x1395, "Sennheiser Communications"},
{0x1397, "BEHRINGER International GmbH"},
{0x1398, "Q-tec"},
{0x13ad, "Baltech"},
{0x13b0, "PerkinElmer Optoelectronics"},
{0x13b1, "Linksys"},
{0x13b2, "Alesis"},
{0x13b3, "Nippon Dics Co., Ltd."},
{0x13ba, "PCPlay"},
{0x13be, "Ricoh Printing Systems, Ltd."},
{0x13ca, "JyeTai Precision Industrial Co., Ltd."},
{0x13cf, "Wisair Ltd."},
{0x13d0, "Techsan Electronics Co., Ltd."},
{0x13d1, "A-Max Technology Macao Commercial Offshore Co. Ltd."},
{0x13d2, "Shark Multimedia"},
{0x13d3, "IMC Networks"},
{0x13dc, "ALEREON, INC."},
{0x13dd, "i.Tech Dynamic Limited"},
{0x13e1, "Kaibo Wire & Cable (Shenzhen) Co., Ltd."},
{0x13e5, "Rane"},
{0x13e6, "TechnoScope Co., Ltd."},
{0x13ea, "Hengstler"},
{0x13ec, "Zydacron"},
{0x13ee, "MosArt"},
{0x13fd, "Initio Corporation"},
{0x13fe, "Kingston Technology Company Inc."},
{0x1400, "Axxion Group Corp."},
{0x1402, "Bowe Bell & Howell"},
{0x1403, "Sitronix"},
{0x1409, "IDS Imaging Development Systems GmbH"},
{0x140e, "Telechips, Inc."},
{0x1410, "Novatel Wireless"},
{0x1415, "Nam Tai E&E Products Ltd. or OmniVision Technologies, Inc."},
{0x1419, "ABILITY ENTERPRISE CO., LTD."},
{0x1429, "Vega Technologies Industrial (Austria) Co."},
{0x142a, "Thales E-Transactions"},
{0x142b, "Arbiter Systems, Inc."},
{0x1430, "RedOctane"},
{0x1431, "Pertech Resources, Inc."},
{0x1435, "Wistron NeWeb"},
{0x1436, "Denali Software, Inc."},
{0x143c, "Altek Corporation"},
{0x1443, "Digilent"},
{0x1446, "X.J.GROUP"},
{0x1453, "Radio Shack"},
{0x1456, "Extending Wire & Cable Co., Ltd."},
{0x1457, "First International Computer, Inc."},
{0x145f, "Trust"},
{0x1460, "Tatung Co."},
{0x1461, "Staccato Communications"},
{0x1462, "Micro Star International"},
{0x1472, "Huawei-3Com"},
{0x147a, "Formosa Industrial Computing, Inc."},
{0x147e, "Upek"},
{0x147f, "Hama GmbH & Co., KG"},
{0x1482, "Vaillant"},
{0x1484, "Elsa AG [hex]"},
{0x1485, "Silicom"},
{0x1487, "DSP Group, Ltd."},
{0x148e, "EVATRONIX SA"},
{0x148f, "Ralink Technology, Corp."},
{0x1491, "Futronic Technology Co. Ltd."},
{0x1493, "Suunto"},
{0x1497, "Panstrong Company Ltd."},
{0x1498, "Microtek International Inc."},
{0x149a, "Imagination Technologies"},
{0x14aa, "WideView Technology Inc."},
{0x14ad, "CTK Corporation"},
{0x14ae, "Printronix Inc."},
{0x14af, "ATP Electronics Inc."},
{0x14b0, "StarTech.com Ltd."},
{0x14b2, "Ralink Technology, Corp."},
{0x14c0, "Rockwell Automation, Inc."},
{0x14c2, "Gemlight Computer, Ltd"},
{0x14c8, "Zytronic"},
{0x14cd, "Super Top"},
{0x14d8, "JAMER INDUSTRIES CO., LTD."},
{0x14dd, "Raritan Computer, Inc."},
{0x14e0, "WiNRADiO Communications"},
{0x14e1, "Dialogue Technology Corp."},
{0x14e5, "SAIN Information & Communications Co., Ltd."},
{0x14ea, "Planex Communications"},
{0x14ed, "Shure Inc."},
{0x14f7, "TechniSat Digital GmbH"},
{0x1500, "Ellisys"},
{0x1501, "Pine-Tum Enterprise Co., Ltd."},
{0x1509, "First International Computer, Inc."},
{0x1513, "medMobile"},
{0x1514, "Actel"},
{0x1516, "CompUSA"},
{0x1518, "Cheshire Engineering Corp."},
{0x1520, "Bitwire Corp."},
{0x1524, "ENE Technology Inc"},
{0x1527, "Silicon Portals"},
{0x1529, "UBIQUAM Co., Ltd."},
{0x152a, "Thesycon Systemsoftware & Consulting GmbH"},
{0x152d, "JMicron Technology Corp. / JMicron USA Technology Corp."},
{0x152e, "LG (HLDS)"},
{0x1532, "Razer USA, Ltd"},
{0x1546, "U-Blox AG"},
{0x1547, "SG Intec Ltd & Co KG"},
{0x154a, "Celectronic GmbH"},
{0x154b, "PNY"},
{0x154d, "ConnectCounty Holdings Berhad"},
{0x154e, "D&M Holdings, Inc. (Denon/Marantz)"},
{0x154f, "SNBC CO., Ltd"},
{0x1554, "Prolink Microsystems Corp."},
{0x1557, "OQO"},
{0x1568, "Sunf Pu Technology Co., Ltd"},
{0x156f, "Quantum Corporation"},
{0x1570, "ALLTOP TECHNOLOGY CO., LTD."},
{0x157b, "Ketron SRL"},
{0x157e, "TRENDnet"},
{0x1582, "Fiberline"},
{0x1587, "SMA Technologie AG"},
{0x158d, "Oakley Inc."},
{0x158e, "JDS Uniphase Corporation (JDSU)"},
{0x1598, "Kunshan Guoji Electronics Co., Ltd."},
{0x15a2, "Freescale Semiconductor, Inc."},
{0x15a4, "Afatech Technologies, Inc."},
{0x15a8, "Teams Power Limited"},
{0x15a9, "Gemtek"},
{0x15aa, "Gearway Electronics (Dong Guan) Co., Ltd."},
{0x15ad, "VMware Inc."},
{0x15ba, "Olimex Ltd."},
{0x15c0, "XL Imaging"},
{0x15c2, "SoundGraph Inc."},
{0x15c5, "Advance Multimedia Internet Technology Inc. (AMIT)"},
{0x15c6, "Laboratoires MXM"},
{0x15c8, "KTF Technologies"},
{0x15c9, "D-Box Technologies"},
{0x15ca, "Textech International Ltd."},
{0x15d5, "Coulomb Electronics Ltd."},
{0x15d9, "Trust International B.V."},
{0x15dc, "Hynix Semiconductor Inc."},
{0x15e0, "Seong Ji Industrial Co., Ltd."},
{0x15e1, "RSA"},
{0x15e4, "Numark"},
{0x15e8, "SohoWare"},
{0x15e9, "Pacific Digital Corp."},
{0x15ec, "Belcarra Technologies Corp."},
{0x15f4, "HanfTek"},
{0x1604, "Tascam"},
{0x1606, "Umax"},
{0x1608, "Inside Out Networks [hex]"},
{0x160a, "VIA Technologies, Inc."},
{0x160e, "INRO"},
{0x1614, "Amoi Electronics"},
{0x1619, "L & K Precision Technology Co., Ltd."},
{0x1621, "Wionics Research"},
{0x1628, "Stonestreet One, Inc."},
{0x162a, "Airgo Networks Inc."},
{0x162f, "WiQuest Communications, Inc."},
{0x1630, "2Wire, Inc."},
{0x1631, "Good Way Technology"},
{0x1645, "Cross Match Technologies GmbH"},
{0x1649, "SofTec Microsystems"},
{0x164a, "ChipX"},
{0x164c, "Matrix Vision GmbH"},
{0x1657, "Struck Innovative Systeme GmbH"},
{0x165b, "Frontier Design Group"},
{0x165c, "Kondo Kagaku"},
{0x1660, "Creatix Polymedia GmbH"},
{0x1668, "Actiontec Electronics, Inc. [hex]"},
{0x1669, "PiKRON Ltd. [hex]"},
{0x166a, "Clipsal"},
{0x1677, "China Huada Integrated Circuit Design (Group) Co., Ltd. (CIDC Group)"},
{0x1679, "Total Phase"},
{0x1680, "Golden Bridge Electech Inc."},
{0x1681, "Prevo Technologies, Inc."},
{0x1682, "Maxwise Production Enterprise Ltd."},
{0x1684, "Godspeed Computer Corp."},
{0x1685, "Delock"},
{0x1686, "ZOOM Corporation"},
{0x1687, "Kingmax Digital Inc."},
{0x1688, "Saab AB"},
{0x1689, "Razer USA, Ltd"},
{0x168c, "Atheros Communications"},
{0x1690, "Askey Computer Corp. [hex]"},
{0x1696, "Hitachi Video and Information System, Inc."},
{0x1697, "VTec Test, Inc."},
{0x16a5, "Shenzhen Zhengerya Cable Co., Ltd."},
{0x16a6, "Unigraf"},
{0x16ab, "Global Sun Technology"},
{0x16ac, "Dongguan ChingLung Wire & Cable Co., Ltd."},
{0x16b4, "iStation"},
{0x16b5, "Persentec, Inc."},
{0x16c0, "Van Ooijen Technische Informatica"},
{0x16ca, "Wireless Cables, Inc."},
{0x16cc, "silex technology, Inc."},
{0x16d0, "MCS"},
{0x16d3, "Frontline Test Equipment, Inc."},
{0x16d5, "AnyDATA Corporation"},
{0x16d6, "JABLOCOM s.r.o."},
{0x16d8, "CMOTECH Co., Ltd."},
{0x16dc, "Wiener, Plein & Baus"},
{0x16df, "King Billion Electronics Co., Ltd."},
{0x16f0, "GN ReSound A/S"},
{0x16f5, "Futurelogic Inc."},
{0x1706, "BlueView Technologies, Inc."},
{0x1707, "ARTIMI"},
{0x170b, "Swissonic"},
{0x170d, "Avnera"},
{0x1725, "Vitesse Semiconductor"},
{0x1726, "Axesstel, Inc."},
{0x172f, "Waltop International Corp."},
{0x1733, "Cellink Technology Co., Ltd"},
{0x1736, "CANON IMAGING SYSTEM TECHNOLOGIES INC."},
{0x1737, "Linksys"},
{0x1740, "Senao"},
{0x1743, "General Atomics"},
{0x174c, "ASMedia Technology Inc."},
{0x174f, "Syntek"},
{0x1753, "GERTEC Telecomunicacoes Ltda."},
{0x1756, "ENENSYS Technologies"},
{0x1759, "LucidPort Technology, Inc."},
{0x1761, "ASUSTek Computer, Inc. (wrong ID)"},
{0x1772, "System Level Solutions, Inc."},
{0x1776, "Arowana"},
{0x177f, "Sweex"},
{0x1781, "Multiple Vendors"},
{0x1782, "Spreadtrum Communications Inc."},
{0x1784, "TopSeed Technology Corp."},
{0x1787, "ATI AIB"},
{0x1788, "ShenZhen Litkconn Technology Co., Ltd."},
{0x1796, "Printrex, Inc."},
{0x1797, "JALCO CO., LTD."},
{0x1799, "Thales Norway A/S"},
{0x179d, "Ricavision International, Inc."},
{0x17a0, "Samson Technologies Corp."},
{0x17a4, "Concept2"},
{0x17a5, "Advanced Connection Technology Inc."},
{0x17a7, "MICOMSOFT CO., LTD."},
{0x17a8, "Kamstrup A/S"},
{0x17b3, "Grey Innovation"},
{0x17b5, "Lunatone"},
{0x17ba, "SAURIS GmbH"},
{0x17c3, "Singim International Corp."},
{0x17cc, "Native Instruments"},
{0x17cf, "Hip Hing Cable & Plug Mfy. Ltd."},
{0x17d0, "Sanford L.P."},
{0x17d3, "Korea Techtron Co., Ltd."},
{0x17e9, "DisplayLink"},
{0x17eb, "Cornice, Inc."},
{0x17ef, "Lenovo"},
{0x17f4, "WaveSense"},
{0x17f5, "K.K. Rocky"},
{0x17f6, "Unicomp, Inc"},
{0x1809, "Advantech"},
{0x1822, "Twinhan"},
{0x1831, "Gwo Jinn Industries Co., Ltd."},
{0x1832, "Huizhou Shenghua Industrial Co., Ltd."},
{0x183d, "VIVOphone"},
{0x1843, "Vaisala"},
{0x1849, "ASRock Incorporation"},
{0x1852, "GYROCOM C&C Co., LTD"},
{0x1854, "Memory Devices Ltd."},
{0x185b, "Compro"},
{0x1861, "Tech Technology Industrial Company"},
{0x1862, "Teridian Semiconductor Corp."},
{0x1870, "Nexio Co., Ltd"},
{0x1871, "Aveo Technology Corp."},
{0x1873, "Navilock"},
{0x187c, "Alienware Corporation"},
{0x187f, "Siano Mobile Silicon"},
{0x1892, "Vast Technologies, Inc."},
{0x1894, "Topseed"},
{0x1897, "Evertop Wire Cable Co."},
{0x189f, "3Shape A/S"},
{0x18a4, "CSSN"},
{0x18a5, "Verbatim, Ltd"},
{0x18b1, "Petalynx"},
{0x18b4, "e3C Technologies"},
{0x18b6, "Mikkon Technology Limited"},
{0x18b7, "Zotek Electronic Co., Ltd."},
{0x18c5, "AMIT Technology, Inc."},
{0x18cd, "Ecamm"},
{0x18d1, "Google Inc."},
{0x18d5, "Starline International Group Limited"},
{0x18d9, "Kaba"},
{0x18dc, "LKC Technologies, Inc."},
{0x18dd, "Planon System Solutions Inc."},
{0x18e3, "Fitipower Integrated Technology Inc"},
{0x18e8, "Qcom"},
{0x18ea, "Matrox Graphics, Inc."},
{0x18ec, "Arkmicro Technologies Inc."},
{0x18fd, "FineArch Inc."},
{0x1908, "GEMBIRD"},
{0x190d, "Motorola GSG"},
{0x1914, "Alco Digital Devices Limited"},
{0x1915, "Nordic Semiconductor ASA"},
{0x1923, "FitLinxx"},
{0x1926, "NextWindow"},
{0x192f, "Avago Technologies, Pte."},
{0x1930, "Shenzhen Xianhe Technology Co., Ltd."},
{0x1931, "Ningbo Broad Telecommunication Co., Ltd."},
{0x1934, "Feature Integration Technology Inc. (Fintek)"},
{0x1941, "Dream Link"},
{0x1943, "Sensoray Co., Inc."},
{0x1949, "Lab126, Inc."},
{0x194f, "PreSonus Audio Electronics, Inc."},
{0x1951, "Hyperstone AG"},
{0x1953, "Ironkey Inc."},
{0x1954, "Radiient Technologies"},
{0x195d, "Itron Technology iONE"},
{0x1965, "Uniden Corporation"},
{0x1967, "CASIO HITACHI Mobile Communications Co., Ltd."},
{0x196b, "Wispro Technology Inc."},
{0x1970, "Dane-Elec Corp. USA"},
{0x1975, "Dongguan Guneetal Wire & Cable Co., Ltd."},
{0x1976, "Chipsbrand Microelectronics (HK) Co., Ltd."},
{0x1977, "T-Logic"},
{0x197d, "Leuze electronic"},
{0x1989, "Nuconn Technology Corp."},
{0x198f, "Beceem Communications Inc."},
{0x1990, "Acron Precision Industrial Co., Ltd."},
{0x1995, "Trillium Technology Pty. Ltd."},
{0x199b, "MicroStrain, Inc."},
{0x199e, "The Imaging Source Europe GmbH"},
{0x199f, "Benica Corporation"},
{0x19a8, "Biforst Technology Inc."},
{0x19ab, "Bodelin"},
{0x19af, "S Life"},
{0x19b2, "Batronix"},
{0x19b4, "Celestron"},
{0x19b5, "B & W Group"},
{0x19b6, "Infotech Logistic, LLC"},
{0x19b9, "Data Robotics"},
{0x19c2, "Futuba"},
{0x19ca, "Mindtribe"},
{0x19cf, "Parrot SA"},
{0x19d2, "ZTE WCDMA Technologies MSM"},
{0x19db, "KFI Printers"},
{0x19e1, "WeiDuan Electronic Accessory (S.Z.) Co., Ltd."},
{0x19e8, "Industrial Technology Research Institute"},
{0x19ef, "Pak Heng Technology (Shenzhen) Co., Ltd."},
{0x19f7, "RODE Microphones"},
{0x19fa, "Gampaq Co.Ltd"},
{0x19ff, "Dynex"},
{0x1a08, "Bellwood International, Inc."},
{0x1a0a, "USB-IF non-workshop"},
{0x1a12, "KES Co., Ltd."},
{0x1a1d, "Veho"},
{0x1a25, "Amphenol East Asia Ltd."},
{0x1a2a, "Seagate Branded Solutions"},
{0x1a2c, "China Resource Semico Co., Ltd"},
{0x1a32, "Quanta Microsystems, Inc."},
{0x1a34, "ACRUX"},
{0x1a36, "Biwin Technology Ltd."},
{0x1a40, "Terminus Technology Inc."},
{0x1a41, "Action Electronics Co., Ltd."},
{0x1a44, "VASCO Data Security International"},
{0x1a4a, "Silicon Image"},
{0x1a4b, "SafeBoot International B.V."},
{0x1a5a, "Tandberg Data"},
{0x1a61, "Abbott Diabetes Care"},
{0x1a6a, "Spansion Inc."},
{0x1a6d, "SamYoung Electronics Co., Ltd"},
{0x1a6e, "Global Unichip Corp."},
{0x1a6f, "Sagem Orga GmbH"},
{0x1a72, "Physik Instrumente"},
{0x1a79, "Bayer Health Care LLC"},
{0x1a7b, "Lumberg Connect  GmbH & Co. KG"},
{0x1a7c, "Evoluent"},
{0x1a81, "Holtek Semiconductor, Inc."},
{0x1a86, "QinHeng Electronics"},
{0x1a89, "Dynalith Systems Co., Ltd."},
{0x1a8b, "SGS Taiwan Ltd."},
{0x1a8d, "BandRich, Inc."},
{0x1a98, "Leica Camera AG"},
{0x1aa4, "Data Drive Thru, Inc."},
{0x1aa5, "UBeacon Technologies, Inc."},
{0x1aa6, "eFortune Technology Corp."},
{0x1aad, "KeeTouch"},
{0x1ab1, "Rigol Technologies"},
{0x1acb, "Salcomp Plc"},
{0x1acc, "Midiplus Co, Ltd."},
{0x1ad1, "Desay Wire Co., Ltd."},
{0x1ad4, "APS"},
{0x1adb, "SEL C662 Serial Cable"},
{0x1ae4, "ic-design Reinhard Gottinger GmbH"},
{0x1ae7, "X-TENSIONS"},
{0x1aed, "High Top Precision Electronic Co., Ltd."},
{0x1aef, "Conntech Electronic (Suzhou) Corporation"},
{0x1af1, "Connect One Ltd."},
{0x1afe, "A. Eberle GmbH & Co. KG"},
{0x1b04, "Meilhaus Electronic GmbH"},
{0x1b0e, "BLUTRONICS S.r.l."},
{0x1b1c, "Corsair"},
{0x1b20, "MStar Semiconductor, Inc."},
{0x1b22, "WiLinx Corp."},
{0x1b26, "Cellex Power Products, Inc."},
{0x1b27, "Current Electronics Inc."},
{0x1b28, "NAVIsis Inc."},
{0x1b32, "Ugobe Life Forms, Inc."},
{0x1b36, "ViXS Systems, Inc."},
{0x1b3b, "iPassion Technology Inc."},
{0x1b3f, "Generalplus Technology Inc."},
{0x1b47, "Energizer Holdings, Inc."},
{0x1b48, "Plastron Precision Co., Ltd."},
{0x1b52, "ARH Inc."},
{0x1b59, "K.S. Terminals Inc."},
{0x1b5a, "Chao Zhou Kai Yuan Electric Co., Ltd."},
{0x1b65, "The Hong Kong Standards and Testing Centre Ltd."},
{0x1b71, "Fushicai"},
{0x1b72, "ATERGI TECHNOLOGY CO., LTD."},
{0x1b73, "Fresco Logic"},
{0x1b75, "Ovislink Corp."},
{0x1b76, "Legend Silicon Corp."},
{0x1b80, "Afatech"},
{0x1b86, "Dongguan Guanshang Electronics Co., Ltd."},
{0x1b88, "ShenMing Electron (Dong Guan) Co., Ltd."},
{0x1b8c, "Altium Limited"},
{0x1b8d, "e-MOVE Technology Co., Ltd."},
{0x1b8e, "Amlogic, Inc."},
{0x1b8f, "MA LABS, Inc."},
{0x1b96, "N-Trig"},
{0x1b98, "YMax Communications Corp."},
{0x1b99, "Shenzhen Yuanchuan Electronic"},
{0x1ba1, "JINQ CHERN ENTERPRISE CO., LTD."},
{0x1ba2, "Lite Metals & Plastic (Shenzhen) Co., Ltd."},
{0x1ba4, "Ember Corporation"},
{0x1ba6, "Abilis Systems"},
{0x1ba8, "China Telecommunication Technology Labs"},
{0x1bad, "Harmonix Music"},
{0x1bae, "Vuzix Corporation"},
{0x1bbb, "T & A Mobile Phones"},
{0x1bc4, "Ford Motor Co."},
{0x1bc5, "AVIXE Technology (China) Ltd."},
{0x1bc7, "Telit Wireless Solutions"},
{0x1bce, "Contac Cable Industrial Limited"},
{0x1bcf, "Sunplus Innovation Technology Inc."},
{0x1bd0, "Hangzhou Riyue Electronic Co., Ltd."},
{0x1bd5, "BG Systems, Inc."},
{0x1bde, "P-TWO INDUSTRIES, INC."},
{0x1bef, "Shenzhen Tongyuan Network-Communication Cables Co., Ltd"},
{0x1bf0, "RealVision Inc."},
{0x1bf5, "Extranet Systems Inc."},
{0x1bf6, "Orient Semiconductor Electronics, Ltd."},
{0x1bfd, "TouchPack"},
{0x1c02, "Kreton Corporation"},
{0x1c04, "QNAP System Inc."},
{0x1c0c, "Ionics EMS, Inc."},
{0x1c0d, "Relm Wireless"},
{0x1c10, "Lanterra Industrial Co., Ltd."},
{0x1c13, "ALECTRONIC LIMITED"},
{0x1c1a, "Datel Electronics Ltd."},
{0x1c1b, "Volkswagen of America, Inc."},
{0x1c1f, "Goldvish S.A."},
{0x1c20, "Fuji Electric Device Technology Co., Ltd."},
{0x1c21, "ADDMM LLC"},
{0x1c22, "ZHONGSHAN CHIANG YU ELECTRIC CO., LTD."},
{0x1c26, "Shanghai Haiying Electronics Co., Ltd."},
{0x1c27, "HuiYang D & S Cable Co., Ltd."},
{0x1c31, "LS Cable Ltd."},
{0x1c34, "SpringCard"},
{0x1c37, "Authorizer Technologies, Inc."},
{0x1c3d, "NONIN MEDICAL INC."},
{0x1c3e, "Wep Peripherals"},
{0x1c40, "EZPrototypes"},
{0x1c49, "Cherng Weei Technology Corp."},
{0x1c4f, "SiGma Micro"},
{0x1c6b, "Philips & Lite-ON Digital Solutions Corporation"},
{0x1c6c, "Skydigital Inc."},
{0x1c73, "AMT"},
{0x1c77, "Kaetat Industrial Co., Ltd."},
{0x1c78, "Datascope Corp."},
{0x1c79, "Unigen Corporation"},
{0x1c7a, "LighTuning Technology Inc."},
{0x1c7b, "LUXSHARE PRECISION INDUSTRY (SHENZHEN) CO., LTD."},
{0x1c83, "Schomcker GmbH"},
{0x1c87, "2N TELEKOMUNIKACE a.s."},
{0x1c88, "Somagic, Inc."},
{0x1c89, "HONGKONG WEIDIDA ELECTRON LIMITED"},
{0x1c8e, "ASTRON INTERNATIONAL CORP."},
{0x1c98, "ALPINE ELECTRONICS, INC."},
{0x1c9e, "OMEGA TECHNOLOGY"},
{0x1ca0, "ACCARIO Inc."},
{0x1ca1, "Symwave"},
{0x1cac, "Kinstone"},
{0x1cb3, "Aces Electronic Co., Ltd."},
{0x1cb4, "OPEX CORPORATION"},
{0x1cb6, "IdeaCom Technology Inc."},
{0x1cbe, "Luminary Micro Inc."},
{0x1cbf, "FORTAT SKYMARK INDUSTRIAL COMPANY"},
{0x1cc0, "PlantSense"},
{0x1cca, "NextWave Broadband Inc."},
{0x1ccd, "Bodatong Technology (Shenzhen) Co., Ltd."},
{0x1cd4, "adp corporation"},
{0x1cd5, "Firecomms Ltd."},
{0x1cd6, "Antonio Precise Products Manufactory Ltd."},
{0x1cde, "Telecommunications Technology Association (TTA)"},
{0x1cdf, "WonTen Technology Co., Ltd."},
{0x1ce0, "EDIMAX TECHNOLOGY CO., LTD."},
{0x1ce1, "Amphenol KAE"},
{0x1cf1, "Dresden Elektronik"},
{0x1cfc, "ANDES TECHNOLOGY CORPORATION"},
{0x1cfd, "Flextronics Digital Design Japan, LTD."},
{0x1d03, "iCON"},
{0x1d07, "Solid-Motion"},
{0x1d08, "NINGBO HENTEK DRAGON ELECTRONICS CO., LTD."},
{0x1d09, "TechFaith Wireless Technology Limited"},
{0x1d0a, "Johnson Controls, Inc. The Automotive Business Unit"},
{0x1d0b, "HAN HUA CABLE & WIRE TECHNOLOGY (J.X.) CO., LTD."},
{0x1d0f, "Sonix Technology Co., Ltd."},
{0x1d14, "ALPHA-SAT TECHNOLOGY LIMITED"},
{0x1d17, "C-Thru Music Ltd."},
{0x1d19, "Dexatek Technology Ltd."},
{0x1d1f, "Diostech Co., Ltd."},
{0x1d20, "SAMTACK INC."},
{0x1d27, "ASUS"},
{0x1d34, "Dream Cheeky"},
{0x1d45, "Touch"},
{0x1d4d, "PEGATRON CORPORATION"},
{0x1d50, "OpenMoko, Inc."},
{0x1d57, "Xenta"},
{0x1d5b, "Smartronix, Inc."},
{0x1d6b, "Linux Foundation"},
{0x1d90, "Citizen"},
{0x1de1, "Actions Microelectronics Co."},
{0x1e0e, "Qualcomm / Option"},
{0x1e10, "Point Grey Research, Inc."},
{0x1e17, "Mirion Technologies Dosimetry Services Division"},
{0x1e1d, "Lumension Security"},
{0x1e1f, "INVIA"},
{0x1e29, "Festo AG & Co. KG"},
{0x1e3d, "Chipsbank Microelectronics Co., Ltd"},
{0x1e41, "Cleverscope"},
{0x1e4e, "Cubeternet"},
{0x1e54, "TypeMatrix"},
{0x1e68, "TrekStor GmbH & Co. KG"},
{0x1e71, "NZXT"},
{0x1e74, "Coby Electronics Corporation"},
{0x1e7d, "ROCCAT"},
{0x1ebb, "NuCORE Technology, Inc."},
{0x1eda, "AirTies Wireless Networks"},
{0x1edb, "Blackmagic design"},
{0x1ee8, "ONDA COMMUNICATION S.p.a."},
{0x1ef6, "EADS Deutschland GmbH"},
{0x1f28, "Cal-Comp"},
{0x1f3a, "Onda (unverified)"},
{0x1f44, "The Neat Company"},
{0x1f48, "H-TRONIC GmbH"},
{0x1f4d, "G-Tek Electronics Group"},
{0x1f6f, "Aliph"},
{0x1f75, "Innostor Technology Corporation"},
{0x1f82, "TANDBERG"},
{0x1f84, "Alere, Inc."},
{0x1f87, "Stantum"},
{0x1f9b, "Ubiquiti Networks, Inc."},
{0x1fab, "Samsung Opto-Electroncs Co., Ltd."},
{0x1fbd, "Delphin Technology AG"},
{0x1fc9, "NXP Semiconductors"},
{0x1fde, "ILX Lightwave Corporation"},
{0x1fe7, "Vertex Wireless Co., Ltd."},
{0x1ff7, "CVT Electronics.Co.,Ltd"},
{0x1fff, "Ideofy Inc."},
{0x2001, "D-Link Corp."},
{0x2002, "DAP Technologies"},
{0x200c, "Reloop"},
{0x2013, "PCTV Systems"},
{0x2019, "PLANEX"},
{0x2040, "Hauppauge"},
{0x2047, "Texas Instruments"},
{0x2080, "Barnes & Noble"},
{0x2087, "Cando"},
{0x20a0, "Clay Logic"},
{0x20b1, "XMOS Ltd"},
{0x20b3, "Hanvon"},
{0x20b7, "Qi Hardware"},
{0x20df, "Simtec Electronics"},
{0x20f4, "TRENDnet"},
{0x2101, "ActionStar"},
{0x2149, "Advanced Silicon S.A."},
{0x2162, "Creative (?)"},
{0x2184, "GW Instek"},
{0x21a1, "Emotiv Systems Pty. Ltd."},
{0x21d6, "Agecodagis SARL"},
{0x2222, "MacAlly"},
{0x2227, "SAMWOO Enterprise"},
{0x2232, "Silicon Motion"},
{0x2233, "RadioShack Corporation"},
{0x2237, "Kobo Inc."},
{0x228d, "8D Technologies inc."},
{0x22a6, "Pie Digital, Inc."},
{0x22b8, "Motorola PCS"},
{0x22b9, "eTurboTouch Technology, Inc."},
{0x22ba, "Technology Innovation Holdings, Ltd"},
{0x2304, "Pinnacle Systems, Inc."},
{0x2318, "Shining Technologies, Inc. [hex]"},
{0x2341, "Arduino SA"},
{0x2373, "Pumatronix Ltda"},
{0x2375, "Digit@lway, Inc."},
{0x2406, "SANHO Digital Electronics Co., Ltd."},
{0x2443, "Aessent Technology Ltd"},
{0x2478, "Tripp-Lite"},
{0x2632, "TwinMOS"},
{0x2650, "Electronics For Imaging, Inc. [hex]"},
{0x2659, "Sundtek"},
{0x2730, "Citizen"},
{0x2735, "DigitalWay"},
{0x2770, "NHJ, Ltd"},
{0x2821, "ASUSTek Computer Inc."},
{0x2899, "Toptronic Industrial Co., Ltd"},
{0x2c02, "Planex Communications"},
{0x2c1a, "Dolphin Peripherals"},
{0x2fb2, "Fujitsu, Ltd"},
{0x3125, "Eagletron"},
{0x3176, "Whanam Electronics Co., Ltd"},
{0x3275, "VidzMedia Pte Ltd"},
{0x3334, "AEI"},
{0x3340, "Yakumo"},
{0x3504, "Micro Star"},
{0x3538, "Power Quotient International Co., Ltd"},
{0x3579, "DIVA"},
{0x3636, "InVibro"},
{0x3838, "WEM"},
{0x3923, "National Instruments Corp."},
{0x40bb, "I-O Data"},
{0x4101, "i-rocks"},
{0x4102, "iRiver, Ltd."},
{0x413c, "Dell Computer Corp."},
{0x4146, "USBest Technology"},
{0x4242, "USB Design by Example"},
{0x4317, "Broadcom Corp."},
{0x4348, "WinChipHead"},
{0x4572, "Shuttle, Inc."},
{0x4586, "Panram"},
{0x4670, "EMS Production"},
{0x4752, "Miditech"},
{0x4757, "GW Instek"},
{0x4766, "Aceeca"},
{0x4855, "Memorex"},
{0x4971, "SimpleTech"},
{0x4d46, "Musical Fidelity"},
{0x5032, "Grandtec"},
{0x5041, "Linksys (?)"},
{0x50c2, "Averatec (?)"},
{0x5173, "Sweex"},
{0x5219, "I-Tetra"},
{0x5345, "Owon"},
{0x544d, "Transmeta Corp."},
{0x5543, "UC-Logic Technology Corp."},
{0x5555, "Epiphan Systems Inc."},
{0x55aa, "OnSpec Electronic, Inc."},
{0x5656, "Uni-Trend Group Limited"},
{0x595a, "IRTOUCHSYSTEMS Co. Ltd."},
{0x5986, "Acer, Inc"},
{0x5a57, "Zinwell"},
{0x6000, "Beholder International Ltd."},
{0x601a, "Ingenic Semiconductor Ltd."},
{0x6189, "Sitecom"},
{0x6253, "TwinHan Technology Co., Ltd"},
{0x636c, "CoreLogic, Inc."},
{0x6472, "Unknown (Sony?)"},
{0x6547, "Arkmicro Technologies Inc."},
{0x6615, "IRTOUCHSYSTEMS Co. Ltd."},
{0x6666, "Prototype product Vendor ID"},
{0x6677, "WiseGroup, Ltd."},
{0x6891, "3Com"},
{0x695c, "Opera1"},
{0x6993, "Yealink Network Technology Co., Ltd."},
{0x6a75, "Shanghai Jujo Electronics Co., Ltd"},
{0x7104, "CME (Central Music Co.)"},
{0x726c, "StackFoundry LLC"},
{0x734c, "TBS Technologies China"},
{0x7392, "Edimax Technology Co., Ltd"},
{0x8086, "Intel Corp."},
{0x8087, "Intel Corp."},
{0x80ee, "VirtualBox"},
{0x8282, "Keio"},
{0x8341, "EGO Systems, Inc."},
{0x9016, "Sitecom"},
{0x9022, "TeVii Technology Ltd."},
{0x9148, "GeoLab, Ltd"},
{0x9710, "MosChip Semiconductor"},
{0x99fa, "Grandtec"},
{0x9ac4, "J. Westhues"},
{0xa128, "AnMo Electronics Corp. / Dino-Lite (?)"},
{0xa168, "AnMo Electronics Corporation"},
{0xa600, "Asix"},
{0xa727, "3Com"},
{0xc251, "Keil Software, Inc."},
{0xcace, "CACE Technologies Inc."},
{0xd209, "Ultimarc"},
{0xe4e4, "Xorcom Ltd."},
{0xeb03, "MakingThings"},
{0xeb1a, "eMPIA Technology, Inc."},
{0xeb2a, "KWorld"},
{0xf003, "Hewlett Packard"},
{0xf4ec, "Atten Electronics / Siglent Technologies"},

{0xffff, NULL},
};

const IdNamePair ClassCodes[] =
{
{0x00, "Deferred to Interface Descriptors"},
{0x01, "Audio"},
{0x02, "Communications and CDC Control"},
{0x03, "HID (Human Interface Device)"},
{0x05, "Physical"},
{0x06, "Image"},
{0x07, "Printer"},
{0x08, "Mass Storage"},
{0x09, "Hub"},
{0x0a, "CDC-Data"},
{0x0b, "Smart Card"},
{0x0d, "Content Security"},
{0x0e, "Video"},
{0x0f, "Personal Healthcare"},
{0x10, "Audio/Video Devices"},
{0xdc, "Diagnostic Device"},
{0xe0, "Wireless Controller"},
{0xef, "Miscellaneous"},
{0xfe, "Application Specific"},
{0xff, "Vendor Specific"},
};

struct UsageTableEntry
{
	U16			usagePage;
	U16			usageID;
	const char* usageName;

	bool operator < (const UsageTableEntry& rhs) const
	{
		if (usagePage == rhs.usagePage)
			return usageID < rhs.usageID;

		return usagePage < rhs.usagePage;
	}
};

const UsageTableEntry UsageTables[] =
{
// 1	Generic Desktop
{1,	0x00,	"Undefined"},
{1,	0x01,	"Pointer"},
{1,	0x02,	"Mouse"},
{1,	0x03,	"Reserved"},
{1,	0x04,	"Joystick"},
{1,	0x05,	"Game Pad"},
{1,	0x06,	"Keyboard"},
{1,	0x07,	"Keypad"},
{1,	0x08,	"Multi-axis Controller"},
{1,	0x09,	"Tablet PC System Controls"},
{1,	0x30,	"X"},
{1,	0x31,	"Y"},
{1,	0x32,	"Z"},
{1,	0x33,	"Rx"},
{1,	0x34,	"Ry"},
{1,	0x35,	"Rz"},
{1,	0x36,	"Slider"},
{1,	0x37,	"Dial"},
{1,	0x38,	"Wheel"},
{1,	0x39,	"Hat Switch"},
{1,	0x3A,	"Counted Buffer"},
{1,	0x3B,	"Byte Count"},
{1,	0x3C,	"Motion Wakeup"},
{1,	0x3D,	"Start"},
{1,	0x3E,	"Select"},
{1,	0x40,	"Vx"},
{1,	0x41,	"Vy"},
{1,	0x42,	"Vz"},
{1,	0x43,	"Vbrx"},
{1,	0x44,	"Vbry"},
{1,	0x45,	"Vbrx"},
{1,	0x46,	"Vno"},
{1,	0x47,	"Feature Notification"},
{1,	0x48,	"Resolution Multiplier"},
{1,	0x80,	"System Control"},
{1,	0x81,	"System Power Down"},
{1,	0x82,	"System Sleep"},
{1,	0x83,	"System Wake Up"},
{1,	0x84,	"System Context Menu"},
{1,	0x85,	"System Main Menu"},
{1,	0x86,	"System App Menu"},
{1,	0x87,	"System Menu Help"},
{1,	0x88,	"System Menu Exit"},
{1,	0x89,	"System Menu Select"},
{1,	0x8A,	"System Menu Right"},
{1,	0x8B,	"System Menu Left"},
{1,	0x8C,	"System Menu Up"},
{1,	0x8D,	"System Menu Down"},
{1,	0x8E,	"System Cold Restart"},
{1,	0x8F,	"System Warm Restart"},
{1,	0x90,	"D-pad Up"},
{1,	0x91,	"D-pad Down"},
{1,	0x92,	"D-pad Right"},
{1,	0x93,	"D-pad Left"},
{1,	0xA0,	"System Dock"},
{1,	0xA1,	"System Undock"},
{1,	0xA2,	"System Setup"},
{1,	0xA3,	"System Break"},
{1,	0xA4,	"System Debugger Break"},
{1,	0xA5,	"Application Break"},
{1,	0xA6,	"Application Debugger Break"},
{1,	0xA7,	"System Speaker Mute"},
{1,	0xA8,	"System Hibernate"},
{1,	0xB0,	"System Display Invert"},
{1,	0xB1,	"System Display Internal"},
{1,	0xB2,	"System Display External"},
{1,	0xB3,	"System Display Both"},
{1,	0xB4,	"System Display Dual"},
{1,	0xB5,	"System Display Toggle Int/Ext"},
{1,	0xB6,	"System Display Swap Primary/Secondary"},
{1,	0xB7,	"System Display LCD Autoscale"},

// 2	Simulation Controls
{2,	0x00,	"Undefined"},
{2,	0x01,	"Flight Simulation Device"},
{2,	0x02,	"Automobile Simulation Device"},
{2,	0x03,	"Tank Simulation Device"},
{2,	0x04,	"Spaceship Simulation Device"},
{2,	0x05,	"Submarine Simulation Device"},
{2,	0x06,	"Sailing Simulation Device"},
{2,	0x07,	"Motorcycle Simulation Device"},
{2,	0x08,	"Sports Simulation Device"},
{2,	0x09,	"Airplane Simulation Device"},
{2,	0x0A,	"Helicopter Simulation Device"},
{2,	0x0B,	"Magic Carpet Simulation Device"},
{2,	0x0C,	"Bicycle"},
{2,	0x20,	"Flight Control Stick"},
{2,	0x21,	"Flight Stick"},
{2,	0x22,	"Cyclic Control"},
{2,	0x23,	"Cyclic Trim"},
{2,	0x24,	"Flight Yoke"},
{2,	0x25,	"Track Control"},
{2,	0x26,	"Driving Control"},
{2,	0xB0,	"Aileron"},
{2,	0xB1,	"Aileron Trim"},
{2,	0xB2,	"Anti-Torque Control"},
{2,	0xB3,	"Auto-pilot Enable"},
{2,	0xB4,	"Chaff Release"},
{2,	0xB5,	"Collective Control"},
{2,	0xB6,	"Dive Brake"},
{2,	0xB7,	"Electronic Counter Measures"},
{2,	0xB8,	"Elevator"},
{2,	0xB9,	"Elevator Trim"},
{2,	0xBA,	"Rudder"},
{2,	0xBB,	"Throttle"},
{2,	0xBC,	"Flight Communication"},
{2,	0xBD,	"Flare Release"},
{2,	0xBE,	"Landing Gear"},
{2,	0xBF,	"Toe Brake"},
{2,	0xC0,	"Trigger"},
{2,	0xC1,	"Weapons Arm"},
{2,	0xC2,	"Weapons Select"},
{2,	0xC3,	"Wing Flaps"},
{2,	0xC4,	"Accelerator"},
{2,	0xC5,	"Brake"},
{2,	0xC6,	"Clutch"},
{2,	0xC7,	"Shifter"},
{2,	0xC8,	"Steering"},
{2,	0xC9,	"Turret Direction"},
{2,	0xCA,	"Barrel Elevation"},
{2,	0xCB,	"Dive Plane"},
{2,	0xCC,	"Ballast"},
{2,	0xCD,	"Bicycle Crank"},
{2,	0xCE,	"Handle Bars"},
{2,	0xCF,	"Front Brake"},
{2,	0xD0,	"Rear Brake"},

// 3	VR Controls
{3,	0x00,	"Unidentified"},
{3,	0x01,	"Belt"},
{3,	0x02,	"Body Suit"},
{3,	0x03,	"Flexor"},
{3,	0x04,	"Glove"},
{3,	0x05,	"Head Tracker"},
{3,	0x06,	"Head Mounted Display"},
{3,	0x07,	"Hand Tracker"},
{3,	0x08,	"Oculometer"},
{3,	0x09,	"Vest"},
{3,	0x0A,	"Animatronic Device"},
{3,	0x20,	"Stereo Enable"},
{3,	0x21,	"Display Enable"},

// 4	Sports Controls
{4,	0x00,	"Unidentified"},
{4,	0x01,	"Baseball Bat"},
{4,	0x02,	"Golf Club"},
{4,	0x03,	"Rowing Machine"},
{4,	0x04,	"Treadmill"},
{4,	0x30,	"Oar"},
{4,	0x31,	"Slope"},
{4,	0x32,	"Rate"},
{4,	0x33,	"Stick Speed"},
{4,	0x34,	"Stick Face Angle"},
{4,	0x35,	"Stick Heel/Toe"},
{4,	0x36,	"Stick Follow Through"},
{4,	0x37,	"Stick Tempo"},
{4,	0x38,	"Stick Type"},
{4,	0x39,	"Stick Height"},
{4,	0x50,	"Putter"},
{4,	0x51,	"1 Iron"},
{4,	0x52,	"2 Iron"},
{4,	0x53,	"3 Iron"},
{4,	0x54,	"4 Iron"},
{4,	0x55,	"5 Iron"},
{4,	0x56,	"6 Iron"},
{4,	0x57,	"7 Iron"},
{4,	0x58,	"8 Iron"},
{4,	0x59,	"9 Iron"},
{4,	0x5A,	"10 Iron"},
{4,	0x5B,	"11 Iron"},
{4,	0x5C,	"Sand Wedge"},
{4,	0x5D,	"Loft Wedge"},
{4,	0x5E,	"Power Wedge"},
{4,	0x5F,	"1 Wood"},
{4,	0x60,	"3 Wood"},
{4,	0x61,	"5 Wood"},
{4,	0x62,	"7 Wood"},
{4,	0x63,	"9 Wood"},

// 5	Game Controls
{5,	0x00,	"Undefined"},
{5,	0x01,	"3D Game Controller"},
{5,	0x02,	"Pinball Device"},
{5,	0x03,	"Gun Device"},
{5,	0x20,	"Point of View"},
{5,	0x21,	"Turn Right/Left"},
{5,	0x22,	"Pitch Right/Left"},
{5,	0x23,	"Roll Forward/Backward"},
{5,	0x24,	"Move Right/Left"},
{5,	0x25,	"Move Forward/Backward"},
{5,	0x26,	"Move Up/Down"},
{5,	0x27,	"Lean Right/Left"},
{5,	0x28,	"Lean Forward/Backward"},
{5,	0x29,	"Height of POV"},
{5,	0x2A,	"Flipper"},
{5,	0x2B,	"Secondary Flipper"},
{5,	0x2C,	"Bump"},
{5,	0x2D,	"New Game"},
{5,	0x2E,	"Shoot Ball"},
{5,	0x2F,	"Player"},
{5,	0x30,	"Gun Bolt"},
{5,	0x31,	"Gun Clip"},
{5,	0x32,	"Gun Selector"},
{5,	0x33,	"Gun Single Shot"},
{5,	0x34,	"Gun Burst"},
{5,	0x35,	"Gun Automatic"},
{5,	0x36,	"Gun Safety"},
{5,	0x37,	"Gamepad Fire/Jump"},
{5,	0x39,	"Gamepad Trigger"},

// 6	Device Controls
{6,	0x00,	"Undefined"},
{6,	0x20,	"Battery Strength"},
{6,	0x21,	"Wireless Channel"},
{6,	0x22,	"Wireless ID"},
{6,	0x23,	"Discover Wireless Control"},
{6,	0x24,	"Security Code Character Entered"},
{6,	0x25,	"Security Code Character Erased"},
{6,	0x26,	"Security Code Cleared"},

// 7	Keyboard
{7,	0x00,	"No Event"},
{7,	0x01,	"Keyboard ErrorRollOver"},
{7,	0x02,	"Keyboard POSTFail"},
{7,	0x03,	"Keyboard ErrorUndefined"},
{7,	0x04,	"Keyboard a and A"},
{7,	0x05,	"Keyboard b and B"},
{7,	0x06,	"Keyboard c and C"},
{7,	0x07,	"Keyboard d and D"},
{7,	0x08,	"Keyboard e and E"},
{7,	0x09,	"Keyboard f and F"},
{7,	0x0A,	"Keyboard g and G"},
{7,	0x0B,	"Keyboard h and H"},
{7,	0x0C,	"Keyboard i and I"},
{7,	0x0D,	"Keyboard j and J"},
{7,	0x0E,	"Keyboard k and K"},
{7,	0x0F,	"Keyboard l and L"},
{7,	0x10,	"Keyboard m and M"},
{7,	0x11,	"Keyboard n and N"},
{7,	0x12,	"Keyboard o and O"},
{7,	0x13,	"Keyboard p and P"},
{7,	0x14,	"Keyboard q and Q"},
{7,	0x15,	"Keyboard r and R"},
{7,	0x16,	"Keyboard s and S"},
{7,	0x17,	"Keyboard t and T"},
{7,	0x18,	"Keyboard u and U"},
{7,	0x19,	"Keyboard v and V"},
{7,	0x1A,	"Keyboard w and W"},
{7,	0x1B,	"Keyboard x and X"},
{7,	0x1C,	"Keyboard y and Y"},
{7,	0x1D,	"Keyboard z and Z"},
{7,	0x1E,	"Keyboard 1 and !"},
{7,	0x1F,	"Keyboard 2 and @"},
{7,	0x20,	"Keyboard 3 and #"},
{7,	0x21,	"Keyboard 4 and $"},
{7,	0x22,	"Keyboard 5 and %"},
{7,	0x23,	"Keyboard 6 and ^"},
{7,	0x24,	"Keyboard 7 and &"},
{7,	0x25,	"Keyboard 8 and *"},
{7,	0x26,	"Keyboard 9 and ("},
{7,	0x27,	"Keyboard 0 and )"},
{7,	0x28,	"Keyboard Return (ENTER)"},
{7,	0x29,	"Keyboard ESCAPE"},
{7,	0x2A,	"Keyboard DELETE (Backspace)"},
{7,	0x2B,	"Keyboard Tab"},
{7,	0x2C,	"Keyboard Spacebar"},
{7,	0x2D,	"Keyboard - and (underscore)"},
{7,	0x2E,	"Keyboard = and +"},
{7,	0x2F,	"Keyboard [ and {"},
{7,	0x30,	"Keyboard ] and }"},
{7,	0x31,	"Keyboard \\ and |"},
{7,	0x32,	"Keyboard Non-US # and ~"},
{7,	0x33,	"Keyboard ; and :"},
{7,	0x34,	"Keyboard ' and \""},
{7,	0x35,	"Keyboard Grave Accent and Tilde"},
{7,	0x36,	"Keyboard , and <"},
{7,	0x37,	"Keyboard . and >"},
{7,	0x38,	"Keyboard / and ?"},
{7,	0x39,	"Keyboard Caps Lock"},
{7,	0x3A,	"Keyboard F1"},
{7,	0x3B,	"Keyboard F2"},
{7,	0x3C,	"Keyboard F3"},
{7,	0x3D,	"Keyboard F4"},
{7,	0x3E,	"Keyboard F5"},
{7,	0x3F,	"Keyboard F6"},
{7,	0x40,	"Keyboard F7"},
{7,	0x41,	"Keyboard F8"},
{7,	0x42,	"Keyboard F9"},
{7,	0x43,	"Keyboard F10"},
{7,	0x44,	"Keyboard F11"},
{7,	0x45,	"Keyboard F12"},
{7,	0x46,	"Keyboard PrintScreen"},
{7,	0x47,	"Keyboard Scroll Lock"},
{7,	0x48,	"Keyboard Pause"},
{7,	0x49,	"Keyboard Insert"},
{7,	0x4A,	"Keyboard Home"},
{7,	0x4B,	"Keyboard PageUp"},
{7,	0x4C,	"Keyboard Delete Forward"},
{7,	0x4D,	"Keyboard End"},
{7,	0x4E,	"Keyboard PageDown"},
{7,	0x4F,	"Keyboard RightArrow"},
{7,	0x50,	"Keyboard LeftArrow"},
{7,	0x51,	"Keyboard DownArrow"},
{7,	0x52,	"Keyboard UpArrow"},
{7,	0x53,	"Keypad Num Lock and Clear"},
{7,	0x54,	"Keypad /"},
{7,	0x55,	"Keypad *"},
{7,	0x56,	"Keypad -"},
{7,	0x57,	"Keypad +"},
{7,	0x58,	"Keypad ENTER"},
{7,	0x59,	"Keypad 1 and End"},
{7,	0x5A,	"Keypad 2 and Down Arrow"},
{7,	0x5B,	"Keypad 3 and PageDn"},
{7,	0x5C,	"Keypad 4 and Left Arrow"},
{7,	0x5D,	"Keypad 5"},
{7,	0x5E,	"Keypad 6 and Right Arrow"},
{7,	0x5F,	"Keypad 7 and Home"},
{7,	0x60,	"Keypad 8 and Up Arrow"},
{7,	0x61,	"Keypad 9 and PageUp"},
{7,	0x62,	"Keypad 0 and Insert"},
{7,	0x63,	"Keypad . and Delete"},
{7,	0x64,	"Keyboard Non-US \\ and |"},
{7,	0x65,	"Keyboard Application"},
{7,	0x66,	"Keyboard Power"},
{7,	0x67,	"Keypad ="},
{7,	0x68,	"Keyboard F13"},
{7,	0x69,	"Keyboard F14"},
{7,	0x6A,	"Keyboard F15"},
{7,	0x6B,	"Keyboard F16"},
{7,	0x6C,	"Keyboard F17"},
{7,	0x6D,	"Keyboard F18"},
{7,	0x6E,	"Keyboard F19"},
{7,	0x6F,	"Keyboard F20"},
{7,	0x70,	"Keyboard F21"},
{7,	0x71,	"Keyboard F22"},
{7,	0x72,	"Keyboard F23"},
{7,	0x73,	"Keyboard F24"},
{7,	0x74,	"Keyboard Execute"},
{7,	0x75,	"Keyboard Help"},
{7,	0x76,	"Keyboard Menu"},
{7,	0x77,	"Keyboard Select"},
{7,	0x78,	"Keyboard Stop"},
{7,	0x79,	"Keyboard Again"},
{7,	0x7A,	"Keyboard Undo"},
{7,	0x7B,	"Keyboard Cut"},
{7,	0x7C,	"Keyboard Copy"},
{7,	0x7D,	"Keyboard Paste"},
{7,	0x7E,	"Keyboard Find"},
{7,	0x7F,	"Keyboard Mute"},
{7,	0x80,	"Keyboard Volume Up"},
{7,	0x81,	"Keyboard Volume Down"},
{7,	0x82,	"Keyboard Locking Caps Lock"},
{7,	0x83,	"Keyboard Locking Num Lock"},
{7,	0x84,	"Keyboard Locking Scroll Lock"},
{7,	0x85,	"Keypad Comma"},
{7,	0x86,	"Keypad Equal Sign"},
{7,	0x87,	"Keyboard International1"},
{7,	0x88,	"Keyboard International2"},
{7,	0x89,	"Keyboard International3"},
{7,	0x8A,	"Keyboard International4"},
{7,	0x8B,	"Keyboard International5"},
{7,	0x8C,	"Keyboard International6"},
{7,	0x8D,	"Keyboard International7"},
{7,	0x8E,	"Keyboard International8"},
{7,	0x8F,	"Keyboard International9"},
{7,	0x90,	"Keyboard LANG1"},
{7,	0x91,	"Keyboard LANG2"},
{7,	0x92,	"Keyboard LANG3"},
{7,	0x93,	"Keyboard LANG4"},
{7,	0x94,	"Keyboard LANG5"},
{7,	0x95,	"Keyboard LANG6"},
{7,	0x96,	"Keyboard LANG7"},
{7,	0x97,	"Keyboard LANG8"},
{7,	0x98,	"Keyboard LANG9"},
{7,	0x99,	"Keyboard Alternate Erase"},
{7,	0x9A,	"Keyboard SysReq/Attention"},
{7,	0x9B,	"Keyboard Cancel"},
{7,	0x9C,	"Keyboard Clear"},
{7,	0x9D,	"Keyboard Prior"},
{7,	0x9E,	"Keyboard Return"},
{7,	0x9F,	"Keyboard Separator"},
{7,	0xA0,	"Keyboard Out"},
{7,	0xA1,	"Keyboard Oper"},
{7,	0xA2,	"Keyboard Clear/Again"},
{7,	0xA3,	"Keyboard CrSel/Props"},
{7,	0xA4,	"Keyboard ExSel"},
{7,	0xB0,	"Keypad 00"},
{7,	0xB1,	"Keypad 000"},
{7,	0xB2,	"Thousands Separator"},
{7,	0xB3,	"Decimal Separator"},
{7,	0xB4,	"Currency Unit"},
{7,	0xB5,	"Currency Sub-unit"},
{7,	0xB6,	"Keypad ("},
{7,	0xB7,	"Keypad )"},
{7,	0xB8,	"Keypad {"},
{7,	0xB9,	"Keypad }"},
{7,	0xBA,	"Keypad Tab"},
{7,	0xBB,	"Keypad Backspace"},
{7,	0xBC,	"Keypad A"},
{7,	0xBD,	"Keypad B"},
{7,	0xBE,	"Keypad C"},
{7,	0xBF,	"Keypad D"},
{7,	0xC0,	"Keypad E"},
{7,	0xC1,	"Keypad F"},
{7,	0xC2,	"Keypad XOR"},
{7,	0xC3,	"Keypad ^"},
{7,	0xC4,	"Keypad %"},
{7,	0xC5,	"Keypad <"},
{7,	0xC6,	"Keypad >"},
{7,	0xC7,	"Keypad &"},
{7,	0xC8,	"Keypad &&"},
{7,	0xC9,	"Keypad |"},
{7,	0xCA,	"Keypad ||"},
{7,	0xCB,	"Keypad :"},
{7,	0xCC,	"Keypad #"},
{7,	0xCD,	"Keypad Space"},
{7,	0xCE,	"Keypad @"},
{7,	0xCF,	"Keypad !"},
{7,	0xD0,	"Keypad Memory Store"},
{7,	0xD1,	"Keypad Memory Recall"},
{7,	0xD2,	"Keypad Memory Clear"},
{7,	0xD3,	"Keypad Memory Add"},
{7,	0xD4,	"Keypad Memory Subtract"},
{7,	0xD5,	"Keypad Memory Multiply"},
{7,	0xD6,	"Keypad Memory Divide"},
{7,	0xD7,	"Keypad +/-"},
{7,	0xD8,	"Keypad Clear"},
{7,	0xD9,	"Keypad Clear Entry"},
{7,	0xDA,	"Keypad Binary"},
{7,	0xDB,	"Keypad Octal"},
{7,	0xDC,	"Keypad Decimal"},
{7,	0xDD,	"Keypad Hexadecimal"},
{7,	0xE0,	"Keyboard LeftControl"},
{7,	0xE1,	"Keyboard LeftShift"},
{7,	0xE2,	"Keyboard LeftAlt"},
{7,	0xE3,	"Keyboard Left GUI"},
{7,	0xE4,	"Keyboard RightControl"},
{7,	0xE5,	"Keyboard RightShift"},
{7,	0xE6,	"Keyboard RightAlt"},
{7,	0xE7,	"Keyboard Right GUI"},

// 8	LEDs
{8,	0x00,	"Undefined"},
{8,	0x01,	"Num Lock"},
{8,	0x02,	"Caps Lock"},
{8,	0x03,	"Scroll Lock"},
{8,	0x04,	"Compose"},
{8,	0x05,	"Kana"},
{8,	0x06,	"Power"},
{8,	0x07,	"Shift"},
{8,	0x08,	"Do Not Disturb"},
{8,	0x09,	"Mute"},
{8,	0x0A,	"Tone Enable"},
{8,	0x0B,	"High Cut Filter"},
{8,	0x0C,	"Low Cut Filter"},
{8,	0x0D,	"Equalizer Enable"},
{8,	0x0E,	"Sound Field On"},
{8,	0x0F,	"Surround Field On"},
{8,	0x10,	"Repeat"},
{8,	0x11,	"Stereo"},
{8,	0x12,	"Sampling Rate Detect"},
{8,	0x13,	"Spinning"},
{8,	0x14,	"CAV"},
{8,	0x15,	"CLV"},
{8,	0x16,	"Recording Format Detect"},
{8,	0x17,	"Off-Hook"},
{8,	0x18,	"Ring"},
{8,	0x19,	"Message Waiting"},
{8,	0x1A,	"Data Mode"},
{8,	0x1B,	"Battery Operation"},
{8,	0x1C,	"Battery OK"},
{8,	0x1D,	"Battery Low"},
{8,	0x1E,	"Speaker"},
{8,	0x1F,	"Head Set"},
{8,	0x20,	"Hold"},
{8,	0x21,	"Microphone"},
{8,	0x22,	"Coverage"},
{8,	0x23,	"Night Mode"},
{8,	0x24,	"Send Calls"},
{8,	0x25,	"Call Pickup"},
{8,	0x26,	"Conference"},
{8,	0x27,	"Stand-by"},
{8,	0x28,	"Camera On"},
{8,	0x29,	"Camera Off"},
{8,	0x2A,	"On-Line"},
{8,	0x2B,	"Off-Line"},
{8,	0x2C,	"Busy"},
{8,	0x2D,	"Ready"},
{8,	0x2E,	"Paper-Out"},
{8,	0x2F,	"Paper-Jam"},
{8,	0x30,	"Remote"},
{8,	0x31,	"Forward"},
{8,	0x32,	"Reverse"},
{8,	0x33,	"Stop"},
{8,	0x34,	"Rewind"},
{8,	0x35,	"Fast Forward"},
{8,	0x36,	"Play"},
{8,	0x37,	"Pause"},
{8,	0x38,	"Record"},
{8,	0x39,	"Error"},
{8,	0x3A,	"Usage Selected Indicator"},
{8,	0x3B,	"Usage In Use Indicator"},
{8,	0x3C,	"Usage Multi Mode Indicator"},
{8,	0x3D,	"Indicator On"},
{8,	0x3E,	"Indicator Flash"},
{8,	0x3F,	"Indicator Slow Blink"},
{8,	0x40,	"Indicator Fast Blink"},
{8,	0x41,	"Indicator Off"},
{8,	0x42,	"Flash On Time"},
{8,	0x43,	"Slow Blink On Time"},
{8,	0x44,	"Slow Blink Off Time"},
{8,	0x45,	"Fast Blink On Time"},
{8,	0x46,	"Fast Blink Off Time"},
{8,	0x47,	"Usage Indicator Color"},
{8,	0x48,	"Red"},
{8,	0x49,	"Green"},
{8,	0x4A,	"Amber"},
{8,	0x4B,	"Generic Indicator"},
{8,	0x4C,	"System Suspend"},
{8,	0x4D,	"External Power Connected"},

// 9	Button
{9,	0x00,	"No button pressed"},
{9,	0x01,	"Button 1 (primary/trigger)"},
{9,	0x02,	"Button 2 (secondary)"},
{9,	0x03,	"Button 3 (tertiary)"},

// 11	Telephony
{11,	0x00,	"Unassigned"},
{11,	0x01,	"Phone"},
{11,	0x02,	"Answering Machine"},
{11,	0x03,	"Message Controls"},
{11,	0x04,	"Handset"},
{11,	0x05,	"Headset"},
{11,	0x06,	"Telephony Key Pad"},
{11,	0x07,	"Programmable Button"},
{11,	0x20,	"Hook Switch"},
{11,	0x21,	"Flash"},
{11,	0x22,	"Feature"},
{11,	0x23,	"Hold"},
{11,	0x24,	"Redial"},
{11,	0x25,	"Transfer"},
{11,	0x26,	"Drop"},
{11,	0x27,	"Park"},
{11,	0x28,	"Forward Calls"},
{11,	0x29,	"Alternate Function"},
{11,	0x2A,	"Line"},
{11,	0x2B,	"Speaker Phone"},
{11,	0x2C,	"Conference"},
{11,	0x2D,	"Ring Enable"},
{11,	0x2E,	"Ring Select"},
{11,	0x2F,	"Phone Mute"},
{11,	0x30,	"Caller ID"},
{11,	0x31,	"Send"},
{11,	0x50,	"Speed Dial"},
{11,	0x51,	"Store Number"},
{11,	0x52,	"Recall Number"},
{11,	0x53,	"Phone Directory"},
{11,	0x70,	"Voice Mail"},
{11,	0x71,	"Screen Calls"},
{11,	0x72,	"Do Not Disturb"},
{11,	0x73,	"Message"},
{11,	0x74,	"Answer On/Off"},
{11,	0x90,	"Inside Dial Tone"},
{11,	0x91,	"Outside Dial Tone"},
{11,	0x92,	"Inside Ring Tone"},
{11,	0x93,	"Outside Ring Tone"},
{11,	0x94,	"Priority Ring Tone"},
{11,	0x95,	"Inside Ringback"},
{11,	0x96,	"Priority Ringback"},
{11,	0x97,	"Line Busy Tone"},
{11,	0x98,	"Reorder Tone"},
{11,	0x99,	"Call Waiting Tone"},
{11,	0x9A,	"Confirmation Tone 1"},
{11,	0x9B,	"Confirmation Tone 2"},
{11,	0x9C,	"Tones Off"},
{11,	0x9D,	"Outside Ringback"},
{11,	0x9E,	"Ringer"},
{11,	0xB0,	"Phone Key 0"},
{11,	0xB1,	"Phone Key 1"},
{11,	0xB2,	"Phone Key 2"},
{11,	0xB3,	"Phone Key 3"},
{11,	0xB4,	"Phone Key 4"},
{11,	0xB5,	"Phone Key 5"},
{11,	0xB6,	"Phone Key 6"},
{11,	0xB7,	"Phone Key 7"},
{11,	0xB8,	"Phone Key 8"},
{11,	0xB9,	"Phone Key 9"},
{11,	0xBA,	"Phone Key Star"},
{11,	0xBB,	"Phone Key Pound"},
{11,	0xBC,	"Phone Key A"},
{11,	0xBD,	"Phone Key B"},
{11,	0xBE,	"Phone Key C"},
{11,	0xBF,	"Phone Key D"},

// 12	Consumer
{12,	0x00,	"Unassigned"},
{12,	0x01,	"Consumer Control"},
{12,	0x02,	"Numeric Key Pad"},
{12,	0x03,	"Programmable Buttons"},
{12,	0x04,	"Microphone"},
{12,	0x05,	"Headphone"},
{12,	0x06,	"Graphic Equalizer"},
{12,	0x20,	"+10"},
{12,	0x21,	"+100"},
{12,	0x22,	"AM/PM"},
{12,	0x30,	"Power"},
{12,	0x31,	"Reset"},
{12,	0x32,	"Sleep"},
{12,	0x33,	"Sleep After"},
{12,	0x34,	"Sleep Mode"},
{12,	0x35,	"Illumination"},
{12,	0x36,	"Function Buttons"},
{12,	0x40,	"Menu"},
{12,	0x41,	"Menu  Pick"},
{12,	0x42,	"Menu Up"},
{12,	0x43,	"Menu Down"},
{12,	0x44,	"Menu Left"},
{12,	0x45,	"Menu Right"},
{12,	0x46,	"Menu Escape"},
{12,	0x47,	"Menu Value Increase"},
{12,	0x48,	"Menu Value Decrease"},
{12,	0x60,	"Data On Screen"},
{12,	0x61,	"Closed Caption"},
{12,	0x62,	"Closed Caption Select"},
{12,	0x63,	"VCR/TV"},
{12,	0x64,	"Broadcast Mode"},
{12,	0x65,	"Snapshot"},
{12,	0x66,	"Still"},
{12,	0x80,	"Selection"},
{12,	0x81,	"Assign Selection"},
{12,	0x82,	"Mode Step"},
{12,	0x83,	"Recall Last"},
{12,	0x84,	"Enter Channel"},
{12,	0x85,	"Order Movie"},
{12,	0x86,	"Channel"},
{12,	0x87,	"Media Selection"},
{12,	0x88,	"Media Select Computer"},
{12,	0x89,	"Media Select TV"},
{12,	0x8A,	"Media Select WWW"},
{12,	0x8B,	"Media Select DVD"},
{12,	0x8C,	"Media Select Telephone"},
{12,	0x8D,	"Media Select Program Guide"},
{12,	0x8E,	"Media Select Video Phone"},
{12,	0x8F,	"Media Select Games"},
{12,	0x90,	"Media Select Messages"},
{12,	0x91,	"Media Select CD"},
{12,	0x92,	"Media Select VCR"},
{12,	0x93,	"Media Select Tuner"},
{12,	0x94,	"Quit"},
{12,	0x95,	"Help"},
{12,	0x96,	"Media Select Tape"},
{12,	0x97,	"Media Select Cable"},
{12,	0x98,	"Media Select Satellite"},
{12,	0x99,	"Media Select Security"},
{12,	0x9A,	"Media Select Home"},
{12,	0x9B,	"Media Select Call"},
{12,	0x9C,	"Channel Increment"},
{12,	0x9D,	"Channel Decrement"},
{12,	0x9E,	"Media Select SAP"},
{12,	0xA0,	"VCR Plus"},
{12,	0xA1,	"Once"},
{12,	0xA2,	"Daily"},
{12,	0xA3,	"Weekly"},
{12,	0xA4,	"Monthly"},
{12,	0xB0,	"Play"},
{12,	0xB1,	"Pause"},
{12,	0xB2,	"Record"},
{12,	0xB3,	"Fast Forward"},
{12,	0xB4,	"Rewind"},
{12,	0xB5,	"Scan Next Track"},
{12,	0xB6,	"Scan Previous Track"},
{12,	0xB7,	"Stop"},
{12,	0xB8,	"Eject"},
{12,	0xB9,	"Random Play"},
{12,	0xBA,	"Select DisC"},
{12,	0xBB,	"Enter Disc"},
{12,	0xBC,	"Repeat"},
{12,	0xBD,	"Tracking"},
{12,	0xBE,	"Track Normal"},
{12,	0xBF,	"Slow Tracking"},
{12,	0xC0,	"Frame Forward"},
{12,	0xC1,	"Frame Back"},
{12,	0xC2,	"Mark"},
{12,	0xC3,	"Clear Mark"},
{12,	0xC4,	"Repeat From Mark"},
{12,	0xC5,	"Return To Mark"},
{12,	0xC6,	"Search Mark Forward"},
{12,	0xC7,	"Search Mark Backwards"},
{12,	0xC8,	"Counter Reset"},
{12,	0xC9,	"Show Counter"},
{12,	0xCA,	"Tracking Increment"},
{12,	0xCB,	"Tracking Decrement"},
{12,	0xCC,	"Stop/Eject"},
{12,	0xCD,	"Play/Pause"},
{12,	0xCE,	"Play/Skip"},
{12,	0xE0,	"Volume"},
{12,	0xE1,	"Balance"},
{12,	0xE2,	"Mute"},
{12,	0xE3,	"Bass"},
{12,	0xE4,	"Treble"},
{12,	0xE5,	"Bass Boost"},
{12,	0xE6,	"Surround Mode"},
{12,	0xE7,	"Loudness"},
{12,	0xE8,	"MPX"},
{12,	0xE9,	"Volume Up"},
{12,	0xEA,	"Volume Down"},
{12,	0xF0,	"Speed Select"},
{12,	0xF1,	"Playback Speed"},
{12,	0xF2,	"Standard Play"},
{12,	0xF3,	"Long Play"},
{12,	0xF4,	"Extended Play"},
{12,	0xF5,	"Slow"},
{12,	0x100,	"Fan Enable"},
{12,	0x101,	"Fan Speed"},
{12,	0x102,	"Light"},
{12,	0x103,	"Light Illumination Level"},
{12,	0x104,	"Climate Control Enable"},
{12,	0x105,	"Room Temperature"},
{12,	0x106,	"Security Enable"},
{12,	0x107,	"Fire Alarm"},
{12,	0x108,	"Police Alarm"},
{12,	0x109,	"Proximity"},
{12,	0x10A,	"Motion"},
{12,	0x10B,	"Duress Alarm"},
{12,	0x10C,	"Holdup Alarm"},
{12,	0x10D,	"Medical Alarm"},
{12,	0x150,	"Balance Right"},
{12,	0x151,	"Balance Left"},
{12,	0x152,	"Bass Increment"},
{12,	0x153,	"Bass Decrement"},
{12,	0x154,	"Treble Increment"},
{12,	0x155,	"Treble Decrement"},
{12,	0x160,	"Speaker System"},
{12,	0x161,	"Channel Left"},
{12,	0x162,	"Channel Right"},
{12,	0x163,	"Channel Center"},
{12,	0x164,	"Channel Front"},
{12,	0x165,	"Channel Center Front"},
{12,	0x166,	"Channel Side"},
{12,	0x167,	"Channel Surround"},
{12,	0x168,	"Channel Low Frequency Enhancement"},
{12,	0x169,	"Channel Top"},
{12,	0x16A,	"Channel Unknown"},
{12,	0x170,	"Sub-channel"},
{12,	0x171,	"Sub-channel Increment"},
{12,	0x172,	"Sub-channel Decrement"},
{12,	0x173,	"Alternate Audio Increment"},
{12,	0x174,	"Alternate Audio Decrement"},
{12,	0x180,	"Application Launch Buttons"},
{12,	0x181,	"AL Launch Button Configuration Tool"},
{12,	0x182,	"AL Programmable Button Configuration"},
{12,	0x183,	"AL Consumer Control Configuration"},
{12,	0x184,	"AL Word Processor"},
{12,	0x185,	"AL Text Editor"},
{12,	0x186,	"AL Spreadsheet"},
{12,	0x187,	"AL Graphics Editor"},
{12,	0x188,	"AL Presentation App"},
{12,	0x189,	"AL Database App"},
{12,	0x18A,	"AL Email Reader"},
{12,	0x18B,	"AL Newsreader"},
{12,	0x18C,	"AL Voicemail"},
{12,	0x18D,	"AL Contacts/Address Book"},
{12,	0x18E,	"AL Calendar/Schedule"},
{12,	0x18F,	"AL Task/Project Manager"},
{12,	0x190,	"AL Log/Journal/Timecard"},
{12,	0x191,	"AL Checkbook/Finance"},
{12,	0x192,	"AL Calculator"},
{12,	0x193,	"AL A/V Capture/Playback"},
{12,	0x194,	"AL Local Machine Browser"},
{12,	0x195,	"AL LAN/WAN Browser"},
{12,	0x196,	"AL Internet Browser"},
{12,	0x197,	"AL Remote Networking/ISP Connect"},
{12,	0x198,	"AL Network Conference"},
{12,	0x199,	"AL Network Chat"},
{12,	0x19A,	"AL Telephony/Dialer"},
{12,	0x19B,	"AL Logon"},
{12,	0x19C,	"AL Logoff"},
{12,	0x19D,	"AL Logon/Logoff"},
{12,	0x19E,	"AL Terminal Lock/Screensaver"},
{12,	0x19F,	"AL Control Panel"},
{12,	0x1A0,	"AL Command Line Processor/Run"},
{12,	0x1A1,	"AL Process/Task Manager"},
{12,	0x1A2,	"AL Select Tast/Application"},
{12,	0x1A3,	"AL Next Task/Application"},
{12,	0x1A4,	"AL Previous Task/Application"},
{12,	0x1A5,	"AL Preemptive Halt Task/Application"},
{12,	0x1A6,	"AL Integrated Help Center"},
{12,	0x1A7,	"AL My Documents"},
{12,	0x1A8,	"AL Thesaurus"},
{12,	0x1A9,	"AL Dictionary"},
{12,	0x1AA,	"AL Desktop"},
{12,	0x1AB,	"AC Spell"},
{12,	0x1AC,	"AL Grammar Check"},
{12,	0x1AD,	"AL Wireless Status"},
{12,	0x1AE,	"AL Keyboard Layout"},
{12,	0x1AF,	"AL Virus Protection"},
{12,	0x1B0,	"AL Encryption"},
{12,	0x1B1,	"AL Screen Saver"},
{12,	0x1B2,	"AL Alarms"},
{12,	0x1B3,	"AL Clock"},
{12,	0x1B4,	"AL File Browser"},
{12,	0x1B5,	"AL Power Status"},
{12,	0x1B6,	"AL My Pictures"},
{12,	0x1B7,	"AL My Music"},
{12,	0x1B8,	"AL Movie Browser"},
{12,	0x1B9,	"AL Digital Rights Manager"},
{12,	0x1BA,	"AL Digital Wallet"},
{12,	0x1BC,	"AL Instant Messaging"},
{12,	0x1BD,	"AL OEM Feature/Tips/Tutorial Browser"},
{12,	0x1BE,	"AL OEM Help"},
{12,	0x1BF,	"AL Online Community"},
{12,	0x1C0,	"AL Entertainment Content Browser"},
{12,	0x1C1,	"AL Online Shopping Browser"},
{12,	0x1C2,	"AL SmartCard Information/Help"},
{12,	0x1C3,	"AL Market Monitor/Finance Browser"},
{12,	0x1C4,	"AL Customized Corporate News Browser"},
{12,	0x1C5,	"AL Online Activity Browser"},
{12,	0x1C6,	"AL Research/Search Browser"},
{12,	0x1C7,	"AL Audio Player"},
{12,	0x200,	"Generic GUI Application Controls"},
{12,	0x201,	"AC New"},
{12,	0x202,	"AC Open"},
{12,	0x203,	"AC Close"},
{12,	0x204,	"AC Exit"},
{12,	0x205,	"AC Maximize"},
{12,	0x206,	"AC Minimize"},
{12,	0x207,	"AC Save"},
{12,	0x208,	"AC Print"},
{12,	0x209,	"AC Properties"},
{12,	0x21A,	"AC Undo"},
{12,	0x21B,	"AC Copy"},
{12,	0x21C,	"AC Cut"},
{12,	0x21D,	"AC Paste"},
{12,	0x21E,	"AC Select All"},
{12,	0x21F,	"AC Find"},
{12,	0x220,	"AC Find and Replace"},
{12,	0x221,	"AC Search"},
{12,	0x222,	"AC Go To"},
{12,	0x223,	"AC Home"},
{12,	0x224,	"AC Back"},
{12,	0x225,	"AC Forward"},
{12,	0x226,	"AC Stop"},
{12,	0x227,	"AC Refresh"},
{12,	0x228,	"AC Previous Link"},
{12,	0x229,	"AC Next Link"},
{12,	0x22A,	"AC Bookmarks"},
{12,	0x22B,	"AC History"},
{12,	0x22C,	"AC Subscriptions"},
{12,	0x22D,	"AC Zoom In"},
{12,	0x22E,	"AC Zoom Out"},
{12,	0x22F,	"AC Zoom"},
{12,	0x230,	"AC Full Screen View"},
{12,	0x231,	"AC Normal View"},
{12,	0x232,	"AC View Toggle"},
{12,	0x233,	"AC Scroll Up"},
{12,	0x234,	"AC Scroll Down"},
{12,	0x235,	"AC Scroll"},
{12,	0x236,	"AC Pan Left"},
{12,	0x237,	"AC Pan Right"},
{12,	0x238,	"AC Pan"},
{12,	0x239,	"AC New Window"},
{12,	0x23A,	"AC Tile Horizontally"},
{12,	0x23B,	"AC Tile Vertically"},
{12,	0x23C,	"AC Format"},
{12,	0x23D,	"AC Edit"},
{12,	0x23E,	"AC Bold"},
{12,	0x23F,	"AC Italics"},
{12,	0x240,	"AC Underline"},
{12,	0x241,	"AC Strikethrough"},
{12,	0x242,	"AC Subscript"},
{12,	0x243,	"AC Superscript"},
{12,	0x244,	"AC All Caps"},
{12,	0x245,	"AC Rotate"},
{12,	0x246,	"AC Resize"},
{12,	0x247,	"AC Flip Horizontal"},
{12,	0x248,	"AC Flip Vertical"},
{12,	0x249,	"AC Mirror Horizontal"},
{12,	0x24A,	"AC Mirror Vertical"},
{12,	0x24B,	"AC Font Select"},
{12,	0x24C,	"AC Font Color"},
{12,	0x24D,	"AC Font Size"},
{12,	0x24E,	"AC Justify Left"},
{12,	0x24F,	"AC Justify Center H"},
{12,	0x250,	"AC Justify Right"},
{12,	0x251,	"AC Justify Block H"},
{12,	0x252,	"AC Justify Top"},
{12,	0x253,	"AC Justify Center V"},
{12,	0x254,	"AC Justify Bottom"},
{12,	0x255,	"AC Justify Block V"},
{12,	0x256,	"AC Justify Decrease"},
{12,	0x257,	"AC Justify Increase"},
{12,	0x258,	"AC Numbered List"},
{12,	0x259,	"AC Restart Numbering"},
{12,	0x25A,	"AC Bulleted List"},
{12,	0x25B,	"AC Promote"},
{12,	0x25C,	"AC Demote"},
{12,	0x25D,	"AC Yes"},
{12,	0x25E,	"AC No"},
{12,	0x25F,	"AC Cancel"},
{12,	0x260,	"AC Catalog"},
{12,	0x261,	"AC Buy/Checkout"},
{12,	0x262,	"AC Add to Cart"},
{12,	0x263,	"AC Expand"},
{12,	0x264,	"AC Expand All"},
{12,	0x265,	"AC Collapse"},
{12,	0x266,	"AC Collapse All"},
{12,	0x267,	"AC Print Preview"},
{12,	0x268,	"AC Paste Special"},
{12,	0x269,	"AC Insert Mode"},
{12,	0x26A,	"AC Delete"},
{12,	0x26B,	"AC Lock"},
{12,	0x26C,	"AC Unlock"},
{12,	0x26D,	"AC Protect"},
{12,	0x26E,	"AC Unprotect"},
{12,	0x26F,	"AC Attach Comment"},
{12,	0x270,	"AC Delete Comment"},
{12,	0x271,	"AC View Comment"},
{12,	0x272,	"AC Select Word"},
{12,	0x273,	"AC Select Sentence"},
{12,	0x274,	"AC Select Paragraph"},
{12,	0x275,	"AC Select Column"},
{12,	0x276,	"AC Select Row"},
{12,	0x277,	"AC Select Table"},
{12,	0x278,	"AC Select Object"},
{12,	0x279,	"AC Redo"},
{12,	0x27A,	"AC Sort"},
{12,	0x27B,	"AC Sort Ascending"},
{12,	0x27C,	"AC Sort Descending"},
{12,	0x27D,	"AC Filter"},
{12,	0x27E,	"AC Set Clock"},
{12,	0x27F,	"AC View Clock"},
{12,	0x280,	"AC Select Time Zone"},
{12,	0x281,	"AC Edit Time Zones"},
{12,	0x282,	"AC Set Alarm"},
{12,	0x283,	"AC Clear Alarm"},
{12,	0x284,	"AC Snooze Alarm"},
{12,	0x285,	"AC Reset Alarm"},
{12,	0x286,	"AC Synchronize"},
{12,	0x287,	"AC Send/Receive"},
{12,	0x288,	"AC Send To"},
{12,	0x289,	"AC Reply"},
{12,	0x28A,	"AC Reply All"},
{12,	0x28B,	"AC Fwd"},
{12,	0x28C,	"AC Send"},
{12,	0x28D,	"AC Attach File"},
{12,	0x28E,	"AC Upload"},
{12,	0x28F,	"AC Download (Save Target As)"},
{12,	0x290,	"AC Set Borders"},
{12,	0x291,	"AC Insert Row"},
{12,	0x292,	"AC Insert Column"},
{12,	0x293,	"AC Insert File"},
{12,	0x294,	"AC Insert Picture"},
{12,	0x295,	"AC Insert Object"},
{12,	0x296,	"AC Insert Symbol"},
{12,	0x297,	"AC Save and Close"},
{12,	0x298,	"AC Rename"},
{12,	0x299,	"AC Merge"},
{12,	0x29A,	"AC Split"},
{12,	0x29B,	"AC Distribute Horizontally"},
{12,	0x29C,	"AC Distribute Vertically"},

// 13	Digitizer
{13,	0x00,	"Undefined"},
{13,	0x01,	"Digitizer"},
{13,	0x02,	"Pen"},
{13,	0x03,	"Light Pen"},
{13,	0x04,	"Touch Screen"},
{13,	0x05,	"Touch Pad"},
{13,	0x06,	"White Board"},
{13,	0x07,	"Coordinate Measuring Machine"},
{13,	0x08,	"3-D Digitizer"},
{13,	0x09,	"Stereo Plotter"},
{13,	0x0A,	"Articulated Arm"},
{13,	0x0B,	"Armature"},
{13,	0x0C,	"Multiple Point Digitizer"},
{13,	0x0D,	"Free Space Wand"},
{13,	0x0E,	"Device Configuration"},
{13,	0x20,	"Stylus"},
{13,	0x21,	"Puck"},
{13,	0x22,	"Finger"},
{13,	0x23,	"Device Settings"},
{13,	0x30,	"Tip Pressure"},
{13,	0x31,	"Barrel Pressure"},
{13,	0x32,	"In Range"},
{13,	0x33,	"Touch"},
{13,	0x34,	"Untouch"},
{13,	0x35,	"Tap"},
{13,	0x36,	"Quality"},
{13,	0x37,	"Data Valid"},
{13,	0x38,	"Transducer Index"},
{13,	0x39,	"Tablet Function Keys"},
{13,	0x3A,	"Program Change Keys"},
{13,	0x3B,	"Battery Strength"},
{13,	0x3C,	"Invert"},
{13,	0x3D,	"X Tilt"},
{13,	0x3E,	"Y Tilt"},
{13,	0x3F,	"Azimuth"},
{13,	0x40,	"Altitude"},
{13,	0x41,	"Twist"},
{13,	0x42,	"Tip Switch"},
{13,	0x43,	"Secondary Tip Switch"},
{13,	0x44,	"Barrel Switch"},
{13,	0x45,	"Eraser"},
{13,	0x46,	"Tablet Pick"},
{13,	0x47,	"Touch Valid"},
{13,	0x48,	"Width"},
{13,	0x49,	"Height"},
{13,	0x51,	"Contact Identifier"},
{13,	0x52,	"Device Mode"},
{13,	0x53,	"Device Identifier"},
{13,	0x54,	"Contact Count"},
{13,	0x55,	"Contact Count Maximum"},

// 15	Physical Interface Device
{15,	0x00,	"Undefined"},
{15,	0x01,	"Physical Interface Device"},
{15,	0x20,	"Normal"},
{15,	0x21,	"Set Effect Report"},
{15,	0x22,	"Effect Block Index"},
{15,	0x23,	"Parameter Block Offset"},
{15,	0x24,	"ROM Flag"},
{15,	0x25,	"Effect Type"},
{15,	0x26,	"ET Constant Force"},
{15,	0x27,	"ET Ramp"},
{15,	0x28,	"ET Custom Force Data"},
{15,	0x30,	"ET Square"},
{15,	0x31,	"ET Sine"},
{15,	0x32,	"ET Triangle"},
{15,	0x33,	"ET Sawtooth Up"},
{15,	0x34,	"ET Sawtooth Down"},
{15,	0x40,	"ET Spring"},
{15,	0x41,	"ET Damper"},
{15,	0x42,	"ET Inertia"},
{15,	0x43,	"ET Friction"},
{15,	0x50,	"Duration"},
{15,	0x51,	"Sample Period"},
{15,	0x52,	"Gain"},
{15,	0x53,	"Trigger Button"},
{15,	0x54,	"Trigger Repeat Interval"},
{15,	0x55,	"Axes Enable"},
{15,	0x56,	"Direction Enable"},
{15,	0x57,	"Direction"},
{15,	0x58,	"Type Specific Block Offset"},
{15,	0x59,	"Block Type"},
{15,	0x5A,	"Set Envelope Report"},
{15,	0x5B,	"Attack Level"},
{15,	0x5C,	"Attack Time"},
{15,	0x5D,	"Fade Level"},
{15,	0x5E,	"Fade Time"},
{15,	0x5F,	"Set Condition Report"},
{15,	0x60,	"CP Offset"},
{15,	0x61,	"Positive Coefficient"},
{15,	0x62,	"Negative Coefficient"},
{15,	0x63,	"Positive Saturation"},
{15,	0x64,	"Negative Saturation"},
{15,	0x65,	"Dead Band"},
{15,	0x66,	"Download Force Sample"},
{15,	0x67,	"Isoch Custom Force Enable"},
{15,	0x68,	"Custom Force Data Report"},
{15,	0x69,	"Custom Force Data"},
{15,	0x6A,	"Custom Force Vendor Defined Data"},
{15,	0x6B,	"Set Custom Force Report"},
{15,	0x6C,	"Custom Force Data Offset"},
{15,	0x6D,	"Sample Count"},
{15,	0x6E,	"Set Periodic Report"},
{15,	0x6F,	"Offset"},
{15,	0x70,	"Magnitude"},
{15,	0x71,	"Phase"},
{15,	0x72,	"Period"},
{15,	0x73,	"Set Constant Force Report"},
{15,	0x74,	"Set Ramp Force Report"},
{15,	0x75,	"Ramp Start"},
{15,	0x76,	"Ramp End"},
{15,	0x77,	"Effect Operation Report"},
{15,	0x78,	"Effect Operation"},
{15,	0x79,	"Op Effect Start"},
{15,	0x7A,	"Op Effect Start Solo"},
{15,	0x7B,	"Op Effect Stop"},
{15,	0x7C,	"Loop Count"},
{15,	0x7D,	"Device Gain Report"},
{15,	0x7E,	"Device Gain"},
{15,	0x7F,	"PID Pool Report"},
{15,	0x80,	"RAM Pool Size"},
{15,	0x81,	"ROM Pool Size"},
{15,	0x82,	"ROM Effect Block Count"},
{15,	0x83,	"Simultaneous Effects Max"},
{15,	0x84,	"Pool Alignment"},
{15,	0x85,	"PID Pool Move Report"},
{15,	0x86,	"Move Source"},
{15,	0x87,	"Move Destination"},
{15,	0x88,	"Move Length"},
{15,	0x89,	"PID Block Load Report"},
{15,	0x8B,	"Block Load Status"},
{15,	0x8C,	"Block Load Success"},
{15,	0x8D,	"Block Load Full"},
{15,	0x8E,	"Block Load Error"},
{15,	0x8F,	"Block Handle"},
{15,	0x90,	"PID Block Free Report"},
{15,	0x91,	"Type Specific Block Handle"},
{15,	0x92,	"PID State Report"},
{15,	0x94,	"Effect Playing"},
{15,	0x95,	"PID Device Control Report"},
{15,	0x96,	"PID Device Control"},
{15,	0x97,	"DC Enable Actuators"},
{15,	0x98,	"DC Disable Actuators"},
{15,	0x99,	"DC Stop All Effects"},
{15,	0x9A,	"DC Device Reset"},
{15,	0x9B,	"DC Device Pause"},
{15,	0x9C,	"DC Device Continue"},
{15,	0x9F,	"Device Paused"},
{15,	0xA0,	"Actuators Enable"},
{15,	0xA4,	"Safety Switch"},
{15,	0xA5,	"Actuator Override Switch"},
{15,	0xA6,	"Actuator Power"},
{15,	0xA7,	"Start Delay"},
{15,	0xA8,	"Parameter Block Size"},
{15,	0xA9,	"Device Managed Pool"},
{15,	0xAA,	"Shared Parameter Blocks"},
{15,	0xAB,	"Create New Effect Report"},
{15,	0xAC,	"RAM Pool Available"},

// 20	Alphnumeric Display
{20,	0x00,	"Undefined"},
{20,	0x01,	"Alphanumeric Display"},
{20,	0x02,	"Bitmapped Display"},
{20,	0x20,	"Display Attributes Report"},
{20,	0x21,	"ASCII Character Set"},
{20,	0x22,	"Data Read Back"},
{20,	0x23,	"Font Read Back"},
{20,	0x24,	"Display Control Report"},
{20,	0x25,	"Clear Display"},
{20,	0x26,	"Display Enable"},
{20,	0x27,	"Screen Saver Delay"},
{20,	0x28,	"Screen Saver Enable"},
{20,	0x29,	"Vertical Scroll"},
{20,	0x2A,	"Horizontal Scroll"},
{20,	0x2B,	"Character Report"},
{20,	0x2C,	"Display Data"},
{20,	0x2D,	"Display Status"},
{20,	0x2E,	"Stat Not Ready"},
{20,	0x2F,	"Stat Ready"},
{20,	0x30,	"Err Not a loadable character"},
{20,	0x31,	"Err Font data cannot be read"},
{20,	0x32,	"Cursor Position Report"},
{20,	0x33,	"Row"},
{20,	0x34,	"Column"},
{20,	0x35,	"Rows"},
{20,	0x36,	"Columns"},
{20,	0x37,	"Cursor Pixel Positioning"},
{20,	0x38,	"Cursor Mode"},
{20,	0x39,	"Cursor Enable"},
{20,	0x3A,	"Cursor Blink"},
{20,	0x3B,	"Font Report"},
{20,	0x3C,	"Font Data"},
{20,	0x3D,	"Character Width"},
{20,	0x3E,	"Character Height"},
{20,	0x3F,	"Character Spacing Horizontal"},
{20,	0x40,	"Character Spacing Vertical"},
{20,	0x41,	"Unicode Character Set"},
{20,	0x42,	"Font 7-Segment"},
{20,	0x43,	"7-Segment Direct Map"},
{20,	0x44,	"Font 14-Segment"},
{20,	0x45,	"14-Segment Direct Map"},
{20,	0x46,	"Display Brightness"},
{20,	0x47,	"Display Contrast"},
{20,	0x48,	"Character Attribute"},
{20,	0x49,	"Attribute Readback"},
{20,	0x4A,	"Attribute Data"},
{20,	0x4B,	"Char Attr Enhance"},
{20,	0x4C,	"Char Attr Underline"},
{20,	0x4D,	"Char Attr Blink"},
{20,	0x80,	"Bitmap Size X"},
{20,	0x81,	"Bitmap Size Y"},
{20,	0x83,	"Bit Depth Format"},
{20,	0x84,	"Display Orientation"},
{20,	0x85,	"Palette Report"},
{20,	0x86,	"Palette Data Size"},
{20,	0x87,	"Palette Data Offset"},
{20,	0x88,	"Palette Data"},
{20,	0x8A,	"Blit Report"},
{20,	0x8B,	"Blit Rectangle X1"},
{20,	0x8C,	"Blit Rectangle Y1"},
{20,	0x8D,	"Blit Rectangle X2"},
{20,	0x8E,	"Blit Rectangle Y2"},
{20,	0x8F,	"Blit Data"},
{20,	0x90,	"Soft Button"},
{20,	0x91,	"Soft Button ID"},
{20,	0x92,	"Soft Button Side"},
{20,	0x93,	"Soft Button Offset 1"},
{20,	0x94,	"Soft Button Offset 2"},
{20,	0x95,	"Soft Button Report"},

// 64	Medical Instrument
{64,	0x00,	"Undefined"},
{64,	0x01,	"Medical Ultrasound"},
{64,	0x20,	"VCR/Acquisition"},
{64,	0x21,	"Freeze/Thaw"},
{64,	0x22,	"Clip Store"},
{64,	0x23,	"Update"},
{64,	0x24,	"Next"},
{64,	0x25,	"Save"},
{64,	0x26,	"Print"},
{64,	0x27,	"Microphone Enable"},
{64,	0x40,	"Cine"},
{64,	0x41,	"Transmit Power"},
{64,	0x42,	"Volume"},
{64,	0x43,	"Focus"},
{64,	0x44,	"Depth"},
{64,	0x60,	"Soft Step - Primary"},
{64,	0x61,	"Soft Step - Secondary"},
{64,	0x70,	"Depth Gain Compensation"},
{64,	0x80,	"Zoom Select"},
{64,	0x81,	"Zoom Adjust"},
{64,	0x82,	"Spectral Doppler Mode Select"},
{64,	0x83,	"Spectral Doppler Adjust"},
{64,	0x84,	"Color Doppler Mode Select"},
{64,	0x85,	"Color Doppler Adjust"},
{64,	0x86,	"Motion Mode Select"},
{64,	0x87,	"Motion Mode Adjust"},
{64,	0x88,	"2-D Mode Select"},
{64,	0x89,	"2-D Mode Adjust"},
{64,	0xA0,	"Soft Control Select"},
{64,	0xA1,	"Soft Control Adjust"},

// 128	Monitor
{128,	0x00,	"Undefined"},
{128,	0x01,	"Monitor Control"},
{128,	0x02,	"EDID Information"},
{128,	0x03,	"VDIF Information"},
{128,	0x04,	"VESA Version"},
{128,	0x05,	"On Screen Display"},
{128,	0x06,	"Auto Size Center"},
{128,	0x07,	"Polarity Horz Synch"},
{128,	0x08,	"Polarity Vert Synch"},
{128,	0x09,	"Sync Type"},
{128,	0x0A,	"Screen Position"},
{128,	0x0B,	"Horizontal Frequency"},
{128,	0x0C,	"Vertical Frequency"},

// 130	VESA Virtual Controls
{130,	0x01,	"Degauss"},
{130,	0x10,	"Brightness"},
{130,	0x12,	"Contrast"},
{130,	0x16,	"Video Gain Red"},
{130,	0x18,	"Video Gain Green"},
{130,	0x1A,	"Video Gain Blue"},
{130,	0x1C,	"Focus"},
{130,	0x20,	"Horizontal Position"},
{130,	0x22,	"Horizontal Size"},
{130,	0x24,	"Horizontal Pincushion"},
{130,	0x26,	"Horizontal Pincushion Balance"},
{130,	0x28,	"Horizontal Misconvergence"},
{130,	0x2A,	"Horizontal Linearity"},
{130,	0x2C,	"Horizontal Linearity Balance"},
{130,	0x30,	"Vertical Position"},
{130,	0x32,	"Vertical Size"},
{130,	0x34,	"Vertical Pincushion"},
{130,	0x36,	"Vertical Pincushion Balance"},
{130,	0x38,	"Vertical Misconvergence"},
{130,	0x3A,	"Vertical Linearity"},
{130,	0x3C,	"Vertical Linearity Balance"},
{130,	0x40,	"Parallelogram Distortion"},
{130,	0x42,	"Trapezoidal Distortion"},
{130,	0x44,	"Tilt"},
{130,	0x46,	"Top Corner Distortion Control"},
{130,	0x48,	"Top Corner Distortion Balance"},
{130,	0x4A,	"Bottom Corner Distortion Control"},
{130,	0x4C,	"Bottom Corner Distortion Balance"},
{130,	0x56,	"Moire Horizontal"},
{130,	0x58,	"Moire Vertical"},
{130,	0x5E,	"Input Level Select"},
{130,	0x60,	"Input Source Select"},
{130,	0x62,	"Stereo Mode"},
{130,	0x6C,	"Video Black Level Red"},
{130,	0x6E,	"Video Black Level Green"},
{130,	0x70,	"Video Black Level Blue"},
{130,	0xA2,	"Auto Size Center"},
{130,	0xA4,	"Polarity Horizontal Synchronization"},
{130,	0xA6,	"Polarity Vertical Synchronization"},
{130,	0xA8,	"Synchronization Type"},
{130,	0xAA,	"Screen Orientation"},
{130,	0xAC,	"Horizontal Frequency"},
{130,	0xAE,	"Vertical Frequency"},
{130,	0xB0,	"Settings"},
{130,	0xCA,	"On Screen Display"},
{130,	0xD4,	"Stereo Mode"},

// 131	VESA Command
{131,	0x00,	"Undefined"},
{131,	0x01,	"Settings"},
{131,	0x02,	"Degauss"},

// 132	Power Device
{132,	0x00,	"Undefined"},
{132,	0x01,	"iName"},
{132,	0x02,	"PresentStatus"},
{132,	0x03,	"ChangedStatus"},
{132,	0x04,	"UPS"},
{132,	0x05,	"PowerSupply"},
{132,	0x10,	"BatterySystem"},
{132,	0x11,	"BatterySystemID"},
{132,	0x12,	"Battery"},
{132,	0x13,	"BatteryID"},
{132,	0x14,	"Charger"},
{132,	0x15,	"ChargerID"},
{132,	0x16,	"PowerConverter"},
{132,	0x17,	"PowerConverterID"},
{132,	0x18,	"OutletSystem"},
{132,	0x19,	"OutletSystemID"},
{132,	0x1A,	"Input"},
{132,	0x1B,	"InputID"},
{132,	0x1C,	"Output"},
{132,	0x1D,	"OutputID"},
{132,	0x1E,	"Flow"},
{132,	0x1F,	"FlowID"},
{132,	0x20,	"Outlet"},
{132,	0x21,	"OutletID"},
{132,	0x22,	"Gang"},
{132,	0x23,	"GangID"},
{132,	0x24,	"Sink"},
{132,	0x25,	"SinkID"},
{132,	0x30,	"Voltage"},
{132,	0x31,	"Current"},
{132,	0x32,	"Frequency"},
{132,	0x33,	"ApparentPower"},
{132,	0x34,	"ActivePower"},
{132,	0x35,	"PercentLoad"},
{132,	0x36,	"Temperature"},
{132,	0x37,	"Humidity"},
{132,	0x38,	"BadCount"},
{132,	0x40,	"ConfigVoltage"},
{132,	0x41,	"ConfigCurrent"},
{132,	0x42,	"ConfigFrequency"},
{132,	0x43,	"ConfigApparentPower"},
{132,	0x44,	"ConfigActivePower"},
{132,	0x45,	"ConfigPercentLoad"},
{132,	0x46,	"ConfigTemperature"},
{132,	0x47,	"ConfigHumidity"},
{132,	0x50,	"SwitchOnControl"},
{132,	0x51,	"SwitchOffControl"},
{132,	0x52,	"ToggleControl"},
{132,	0x53,	"LowVoltageTransfer"},
{132,	0x54,	"HighVoltageTransfer"},
{132,	0x55,	"DelayBeforeReboot"},
{132,	0x56,	"DelayBeforeStartup"},
{132,	0x57,	"DelayBeforeShutdown"},
{132,	0x58,	"Test"},
{132,	0x59,	"Vendorspecificcommand"},
{132,	0x5A,	"AudibleAlarmControl"},
{132,	0x60,	"Present"},
{132,	0x61,	"Good"},
{132,	0x62,	"InternalFailure"},
{132,	0x63,	"VoltageOutOfRange"},
{132,	0x64,	"FrequencyOutOfRange"},
{132,	0x65,	"Overload"},
{132,	0x66,	"OverCharged"},
{132,	0x67,	"OverTemperature"},
{132,	0x68,	"ShutdownRequested"},
{132,	0x69,	"ShutdownImminent"},
{132,	0x6A,	"VendorSpecificAnswerValid"},
{132,	0x6B,	"SwitchOn/Off"},
{132,	0x6C,	"Switchable"},
{132,	0x6D,	"Used"},
{132,	0x6E,	"Boost"},
{132,	0x6F,	"Buck"},
{132,	0x70,	"Initialized"},
{132,	0x71,	"Tested"},
{132,	0x72,	"Awaiting Power"},
{132,	0x73,	"CommunicationLost"},
{132,	0xFD,	"iManufacturer"},
{132,	0xFE,	"iProduct"},
{132,	0xFF,	"iSerialNumber"},

// 133	Battery System
{133,	0x00,	"Undefined"},
{133,	0x01,	"SMBBatteryMode"},
{133,	0x02,	"SMBBatteryStatus"},
{133,	0x03,	"SMBAlarmWarning"},
{133,	0x04,	"SMBChargerMode"},
{133,	0x05,	"SMBChargerStatus"},
{133,	0x06,	"SMBChargerSpecInfo"},
{133,	0x07,	"SMBSelectorState"},
{133,	0x08,	"SMBSelectorPreset"},
{133,	0x09,	"SMBSelectorInfo"},
{133,	0x10,	"OptionalMfgFunction1"},
{133,	0x11,	"OptionalMfgFunction2"},
{133,	0x12,	"OptionalMfgFunction3"},
{133,	0x13,	"OptionalMfgFunction4"},
{133,	0x14,	"OptionalMfgFunction5"},
{133,	0x15,	"ConnectionToSMBus"},
{133,	0x16,	"OutputConnection"},
{133,	0x17,	"ChargerConnection"},
{133,	0x18,	"BatteryInsertion"},
{133,	0x19,	"Usenext"},
{133,	0x1A,	"OKToUse"},
{133,	0x1B,	"BatterySupported"},
{133,	0x1C,	"SelectorRevision"},
{133,	0x1D,	"ChargingIndicator"},
{133,	0x28,	"ManufacturerAccess"},
{133,	0x29,	"RemainingCapacityLimit"},
{133,	0x2A,	"RemainingTimeLimit"},
{133,	0x2B,	"AtRate"},
{133,	0x2C,	"CapacityMode"},
{133,	0x2D,	"BroadcastToCharger"},
{133,	0x2E,	"PrimaryBattery"},
{133,	0x2F,	"ChargeController"},
{133,	0x40,	"TerminateCharge"},
{133,	0x41,	"TermminateDischarge"},
{133,	0x42,	"BelowRemainingCapacityLimit"},
{133,	0x43,	"RemainingTimeLimitExpired"},
{133,	0x44,	"Charging"},
{133,	0x45,	"Discharging"},
{133,	0x46,	"FullyCharged"},
{133,	0x47,	"FullyDischarged"},
{133,	0x48,	"ConditionningFlag"},
{133,	0x49,	"AtRateOK"},
{133,	0x4A,	"SMBErrorCode"},
{133,	0x4B,	"NeedReplacement"},
{133,	0x60,	"AtRateTimeToFull"},
{133,	0x61,	"AtRateTimeToEmpty"},
{133,	0x62,	"AverageCurrent"},
{133,	0x63,	"Maxerror"},
{133,	0x64,	"RelativeStateOfCharge"},
{133,	0x65,	"AbsoluteStateOfCharge"},
{133,	0x66,	"RemainingCapacity"},
{133,	0x67,	"FullChargeCapacity"},
{133,	0x68,	"RunTimeToEmpty"},
{133,	0x69,	"AverageTimeToEmpty"},
{133,	0x6A,	"AverageTimeToFull"},
{133,	0x6B,	"CycleCount"},
{133,	0x80,	"BattPackModelLevel"},
{133,	0x81,	"InternalChargeController"},
{133,	0x82,	"PrimaryBatterySupport"},
{133,	0x83,	"DesignCapacity"},
{133,	0x84,	"SpecificationInfo"},
{133,	0x85,	"ManufacturerDate"},
{133,	0x86,	"SerialNumber"},
{133,	0x87,	"iManufacturerName"},
{133,	0x88,	"iDevicename"},
{133,	0x89,	"iDeviceChemistery"},
{133,	0x8A,	"iManufacturerData"},
{133,	0x8B,	"Rechargeable"},
{133,	0x8C,	"WarningCapacityLimit"},
{133,	0x8D,	"CapacityGranularity1"},
{133,	0x8E,	"CapacityGranularity2"},
{133,	0x8F,	"iOEMInformation"},
{133,	0xC0,	"InhibitCharge"},
{133,	0xC1,	"EnablePolling"},
{133,	0xC2,	"ResetToZero"},
{133,	0xD0,	"ACPresent"},
{133,	0xD1,	"BatteryPresent"},
{133,	0xD2,	"PowerFail"},
{133,	0xD3,	"AlarmInhibited"},
{133,	0xD4,	"ThermistorUnderRange"},
{133,	0xD5,	"ThermistorHot"},
{133,	0xD6,	"ThermistorCold"},
{133,	0xD7,	"ThermistorOverRange"},
{133,	0xD8,	"VoltageOutOfRange"},
{133,	0xD9,	"CurrentOutOfRange"},
{133,	0xDA,	"CurrentNotRegulated"},
{133,	0xDB,	"VoltageNotRegulated"},
{133,	0xDC,	"MasterMode"},
{133,	0xDD,	"ChargerBattery/HostControlled"},
{133,	0xF0,	"ChargerSpecInfo"},
{133,	0xF1,	"ChargerSpecRef"},
{133,	0xF2,	"Level2"},
{133,	0xF3,	"Level3"},

// 140	Bar Code Scanner
{140,	0x00,	"Undefined"},
{140,	0x01,	"Bar Code Badge Reader"},
{140,	0x02,	"Bar Code Scanner"},
{140,	0x03,	"Dumb Bar Code Scanner"},
{140,	0x04,	"Cordless Scanner Base"},
{140,	0x05,	"Bar Code Scanner Cradle"},
{140,	0x10,	"Attribute Report"},
{140,	0x11,	"Settings Report"},
{140,	0x12,	"Scanned Data Report"},
{140,	0x13,	"Raw Scanned Data Report"},
{140,	0x14,	"Trigger Report"},
{140,	0x15,	"Status Report"},
{140,	0x16,	"UPC/EAN Control Report"},
{140,	0x17,	"EAN 2/3 Label Control Report"},
{140,	0x18,	"Code 39 Control Report"},
{140,	0x19,	"Interleaved 2 of 5 Control Report"},
{140,	0x1A,	"Standard 2 of 5 Control Report"},
{140,	0x1B,	"MSI Plessey Control Report"},
{140,	0x1C,	"Codabar Control Report"},
{140,	0x1D,	"Code 128 Control Report"},
{140,	0x1E,	"Misc 1D Control Report"},
{140,	0x1F,	"2D Control Report"},
{140,	0x30,	"Aiming/Pointer Mode"},
{140,	0x31,	"Bar Code Present Sensor"},
{140,	0x32,	"Class 1A Laser"},
{140,	0x33,	"Class 2 Laser"},
{140,	0x34,	"Heater Present"},
{140,	0x35,	"Contact Scanner"},
{140,	0x36,	"Electronic Article Surveillance Notification"},
{140,	0x37,	"Constant Electronic Article"},
{140,	0x38,	"Error Indication"},
{140,	0x39,	"Fixed Beeper"},
{140,	0x3A,	"Good Decode Indication"},
{140,	0x3B,	"Hands Free Scanning"},
{140,	0x3C,	"Intrinsically Safe"},
{140,	0x3D,	"Klasse Eins Laser"},
{140,	0x3E,	"Long Range Scanner"},
{140,	0x3F,	"Mirror Speed Control"},
{140,	0x40,	"Not On File Indication"},
{140,	0x41,	"Programmable Beeper"},
{140,	0x42,	"Triggerless"},
{140,	0x43,	"Wand"},
{140,	0x44,	"Water Resistant"},
{140,	0x45,	"Multi-Range Scanner"},
{140,	0x46,	"Proximity Sensor"},
{140,	0x4D,	"Fragment Decoding"},
{140,	0x4E,	"Scanner Read Confidence"},
{140,	0x4F,	"Data Prefix"},
{140,	0x50,	"Prefix AIMI"},
{140,	0x51,	"Prefix None"},
{140,	0x52,	"Prefix Proprietary"},
{140,	0x55,	"Active Time"},
{140,	0x56,	"Aiming Laser Pattern"},
{140,	0x57,	"Bar Code Present"},
{140,	0x58,	"Beeper State"},
{140,	0x59,	"Laser On Time"},
{140,	0x5A,	"Laser State"},
{140,	0x5B,	"Lockout Timeout"},
{140,	0x5C,	"Motor State"},
{140,	0x5D,	"Motor Timeout"},
{140,	0x5E,	"Power On Reset Scanner"},
{140,	0x5F,	"Prevent Read of Barcodes"},
{140,	0x60,	"Initiate Barcode Read"},
{140,	0x61,	"Trigger State"},
{140,	0x62,	"Trigger Mode"},
{140,	0x63,	"Trigger Mode Blinking Laser On"},
{140,	0x64,	"Trigger Mode Continuous Laser On"},
{140,	0x65,	"Trigger Mode Laser on while Pulled"},
{140,	0x66,	"Trigger Mode Laser stays on after Trigger release"},
{140,	0x6D,	"Commit Parameters to NVM"},
{140,	0x6E,	"Parameter Scanning"},
{140,	0x6F,	"Parameters Changed"},
{140,	0x70,	"Set parapeter default values"},
{140,	0x75,	"Scanner In Cradle"},
{140,	0x76,	"Scanner In Range"},
{140,	0x7A,	"Aim Duration"},
{140,	0x7B,	"Good Read Lamp Duration"},
{140,	0x7C,	"Good Read Lamp Intensity"},
{140,	0x7D,	"Good Read LED"},
{140,	0x7E,	"Good Read Tone Frequence"},
{140,	0x7F,	"Good Read Tone Length"},
{140,	0x80,	"Good Read Tone Volume"},
{140,	0x82,	"No Read Message"},
{140,	0x83,	"Not On File Volume"},
{140,	0x84,	"Powerup Beep"},
{140,	0x85,	"Sound Error Beep"},
{140,	0x86,	"Sound Good Read Beep"},
{140,	0x87,	"Sound Not On File Beep"},
{140,	0x88,	"Good Read When to Write"},
{140,	0x89,	"GRWTI After Decode"},
{140,	0x8A,	"GRWTI Beep/Lamp after transmit"},
{140,	0x8B,	"GRWTI No Beep/Lamp use at all"},
{140,	0x91,	"Bookland EAN"},
{140,	0x92,	"Convert EAN 8 to 13 Type"},
{140,	0x93,	"Convert UPC A to EAN-13"},
{140,	0x94,	"Convert UPC-E to A"},
{140,	0x95,	"EAN-13"},
{140,	0x96,	"EAN-8"},
{140,	0x97,	"EAN-99 128_Mandatory"},
{140,	0x98,	"EAN-99 P5/128_Optional"},
{140,	0x9A,	"UPC/EAN"},
{140,	0x9B,	"UPC/EAN Coupon Code"},
{140,	0x9C,	"UPC/EAN Periodicals"},
{140,	0x9D,	"UPC-A"},
{140,	0x9E,	"UPC-A with 128 Mandatory"},
{140,	0x9F,	"UPC-A with 128 Optional"},
{140,	0xA0,	"UPC-A with P5 Optional"},
{140,	0xA1,	"UPC-E"},
{140,	0xA2,	"UPC-E1"},
{140,	0xA9,	"Periodical"},
{140,	0xAA,	"Periodical Auto-Discriminate + 2"},
{140,	0xAB,	"Periodical Only Decode with + 2"},
{140,	0xAC,	"Periodical Ignore + 2"},
{140,	0xAD,	"Periodical Auto-Discriminate + 5"},
{140,	0xAE,	"Periodical Only Decode with + 5"},
{140,	0xAF,	"Periodical Ignore + 5"},
{140,	0xB0,	"Check"},
{140,	0xB1,	"Check Disable Price"},
{140,	0xB2,	"Check Enable 4 digit Price"},
{140,	0xB3,	"Check Enable 5 digit Price"},
{140,	0xB4,	"Check Enable European 4 digit Price"},
{140,	0xB5,	"Check European 5 digit Price"},
{140,	0xB7,	"EAN Two Label"},
{140,	0xB8,	"EAN Three Label"},
{140,	0xB9,	"EAN 8 Flag Digit 1"},
{140,	0xBA,	"EAN 8 Flag Digit 2"},
{140,	0xBB,	"EAN 8 Flag Digit 3"},
{140,	0xBC,	"EAN 13 Flag Digit 1"},
{140,	0xBD,	"EAN 13 Flag Digit 2"},
{140,	0xBE,	"EAN 13 Flag Digit 3"},
{140,	0xBF,	"Add EAN 2/3 Label Definition"},
{140,	0xC0,	"Clear all EAN 2/3 Label Definitions"},
{140,	0xC3,	"Codabar"},
{140,	0xC4,	"Code 128"},
{140,	0xC7,	"Code 39"},
{140,	0xC8,	"Code 93"},
{140,	0xC9,	"Full ASCII Conversion"},
{140,	0xCA,	"Interleaved 2 of 5"},
{140,	0xCB,	"Italian Pharmacy Code"},
{140,	0xCC,	"MSI/Plessey"},
{140,	0xCD,	"Standard 2 of 5 IATA"},
{140,	0xCE,	"Standard 2 of 5"},
{140,	0xD3,	"Transmit Start/Stop"},
{140,	0xD4,	"Tri-Optic"},
{140,	0xD5,	"UCC/EAN-128"},
{140,	0xD6,	"Check Digit"},
{140,	0xD7,	"Check Digit Disable"},
{140,	0xD8,	"Check Digit Enable Interleaved 2 of 5 OPCC"},
{140,	0xD9,	"Check Digit Enable Interleaved 2 of 5 USS"},
{140,	0xDA,	"Check Digit Enable Standard 2 of 5 OPCC"},
{140,	0xDB,	"Check Digit Enable Standard 2 of 5 USS"},
{140,	0xDC,	"Check Digit Enable One MSI Plessey"},
{140,	0xDD,	"Check Digit Enable Two MSI Plessey"},
{140,	0xDE,	"Check Digit Codabar Enable"},
{140,	0xDF,	"Check Digit Code 39 Enable"},
{140,	0xF0,	"Transmit Check Digit"},
{140,	0xF1,	"Disable Check Digit Transmit"},
{140,	0xF2,	"Enable Check Digit Transmit"},
{140,	0xFB,	"Symbology Identifier 1"},
{140,	0xFC,	"Symbology Identifier 2"},
{140,	0xFD,	"Symbology Identifier 3"},
{140,	0xFE,	"Decoded Data"},
{140,	0xFF,	"Decode Data Continued"},
{140,	0x100,	"Bar Space Data"},
{140,	0x101,	"Scanner Data Accuracy"},
{140,	0x102,	"Raw Data Polarity"},
{140,	0x103,	"Polarity Inverted Bar Code"},
{140,	0x104,	"Polarity Normal Bar Code"},
{140,	0x106,	"Minimum Length to Decode"},
{140,	0x107,	"Maximum Length to Decode"},
{140,	0x108,	"First Discrete Length to Decode"},
{140,	0x109,	"Second Discrete Length to Decode"},
{140,	0x10A,	"Data Length Method"},
{140,	0x10B,	"DL Method Read any"},
{140,	0x10C,	"DL Method Check in Range"},
{140,	0x10D,	"DL Method Check for Discrete"},
{140,	0x110,	"Aztec Code"},
{140,	0x111,	"BC412"},
{140,	0x112,	"Channel Code"},
{140,	0x113,	"Code 16"},
{140,	0x114,	"Code 32"},
{140,	0x115,	"Code 49"},
{140,	0x116,	"Code One"},
{140,	0x117,	"Colorcode"},
{140,	0x118,	"Data Matrix"},
{140,	0x119,	"MaxiCode"},
{140,	0x11A,	"MicroPDF"},
{140,	0x11B,	"PDF-417"},
{140,	0x11C,	"PosiCode"},
{140,	0x11D,	"QR Code"},
{140,	0x11E,	"SuperCode"},
{140,	0x11F,	"UltraCode"},
{140,	0x120,	"USD-5 (Slug Code)"},
{140,	0x121,	"VeriCode"},

// 141	Scale Device
{141,	0x00,	"Undefined"},
{141,	0x01,	"Weighing Device"},
{141,	0x20,	"Scale Device"},
{141,	0x21,	"Scale Class I Metric"},
{141,	0x22,	"Scale Class I Metric"},
{141,	0x23,	"Scale Class II Metric"},
{141,	0x24,	"Scale Class III Metric"},
{141,	0x25,	"Scale Class IIIL Metric"},
{141,	0x26,	"Scale Class IV Metric"},
{141,	0x27,	"Scale Class III English"},
{141,	0x28,	"Scale Class IIIL English"},
{141,	0x29,	"Scale Class IV English"},
{141,	0x2A,	"Scale Class Generic"},
{141,	0x30,	"Scale Attribute Report"},
{141,	0x31,	"Scale Control Report"},
{141,	0x32,	"Scale Data Report"},
{141,	0x33,	"Scale Status Report"},
{141,	0x34,	"Scale Weight Limit Report"},
{141,	0x35,	"Scale Statistics Report"},
{141,	0x40,	"Data Weight"},
{141,	0x41,	"Data Scaling"},
{141,	0x50,	"Weight Unit"},
{141,	0x51,	"Weight Unit Milligram"},
{141,	0x52,	"Weight Unit Gram"},
{141,	0x53,	"Weight Unit Kilogram"},
{141,	0x54,	"Weight Unit Carats"},
{141,	0x55,	"Weight Unit Taels"},
{141,	0x56,	"Weight Unit Grains"},
{141,	0x57,	"Weight Unit Pennyweights"},
{141,	0x58,	"Weight Unit Metric Ton"},
{141,	0x59,	"Weight Unit Avoir Ton"},
{141,	0x5A,	"Weight Unit Troy Ounce"},
{141,	0x5B,	"Weight Unit Ounce"},
{141,	0x5C,	"Weight Unit Pound"},
{141,	0x60,	"Calibration Count"},
{141,	0x61,	"Re-Zero Count"},
{141,	0x70,	"Scale Status"},
{141,	0x71,	"Scale Status Fault"},
{141,	0x72,	"Scale Status Stable at Center of Zero"},
{141,	0x73,	"Scale Status In Motion"},
{141,	0x74,	"Scale Status Weight Stable"},
{141,	0x75,	"Scale Status Under Zero"},
{141,	0x76,	"Scale Status Over Weight Limit"},
{141,	0x77,	"Scale Status Requires Rezeroing"},
{141,	0x80,	"Zero Scale"},
{141,	0x81,	"Enforced Zero Return"},

// 142	Magnetic Stripe Reader
{142,	0x00,	"Undefined"},
{142,	0x01,	"MSR Device Read-Only"},
{142,	0x11,	"Track 1 Length"},
{142,	0x12,	"Track 2 Length"},
{142,	0x13,	"Track 3 Length"},
{142,	0x14,	"Track JIS Length"},
{142,	0x20,	"Track Data"},
{142,	0x21,	"Track 1 Data"},
{142,	0x22,	"Track 2 Data"},
{142,	0x23,	"Track 3 Data"},
{142,	0x24,	"Track JIS Data"},

{0xffff,	0xffff,	"<unknown>"},
};

const char* GetLangName(U16 langID)
{
	const size_t NUM_LANGS = sizeof(USBLangIDs) / sizeof(IdNamePair);

	IdNamePair srch;
	srch.id = langID;

	const IdNamePair* res = std::lower_bound(USBLangIDs, USBLangIDs + NUM_LANGS, srch);

	if (res->id == langID  &&  res->name != NULL)
		return res->name;

	return "<unknown>";
}

const char* GetVendorName(U16 vendorID)
{
	const size_t NUM_VENDORS = sizeof(VendorIDs) / sizeof(IdNamePair);

	IdNamePair srch;
	srch.id = vendorID;

	const IdNamePair* res = std::lower_bound(VendorIDs, VendorIDs + NUM_VENDORS, srch);

	if (res->id == vendorID  &&  res->name != NULL)
		return res->name;

	return "<unknown>";
}

const char* GetUSBClassName(U8 classCode)
{
	const size_t NUM_ITEMS = sizeof(ClassCodes) / sizeof(IdNamePair);

	IdNamePair srch;
	srch.id = classCode;

	const IdNamePair* res = std::lower_bound(ClassCodes, ClassCodes + NUM_ITEMS, srch);

	if (res->id == classCode  &&  res->name != NULL)
		return res->name;

	return "<unknown>";
}

std::string GetHIDUsagePageName(U16 usagePage)
{
	switch (usagePage)
	{
	case 0x00:	return "Undefined";
	case 0x01:	return "Generic Desktop Controls";
	case 0x02:	return "Simulation Controls";
	case 0x03:	return "VR Controls";
	case 0x04:	return "Sport Controls";
	case 0x05:	return "Game Controls";
	case 0x06:	return "Generic Device Controls";
	case 0x07:	return "Keyboard/Keypad";
	case 0x08:	return "LEDs";
	case 0x09:	return "Button";
	case 0x0A:	return "Ordinal";
	case 0x0B:	return "Telephony";
	case 0x0C:	return "Consumer";
	case 0x0D:	return "Digitizer";
	case 0x0F:	return "PID Page";
	case 0x10:	return "Unicode";
	case 0x14:	return "Alphanumeric Display";
	case 0x40:	return "Medical Instruments";
	case 0x80:
	case 0x81:
	case 0x82:
	case 0x83:	return "Monitor page";
	case 0x84:
	case 0x85:
	case 0x86:
	case 0x87:	return "Power page";
	case 0x8C:	return "Bar Code Scanner page";
	case 0x8D:	return "Scale page";
	case 0x8E:	return "Magnetic Stripe Reading (MSR) Devices";
	case 0x8F:	return "Reserved Point of Sale pages";
	case 0x90:	return "Camera Control Page";
	case 0x91:	return "Arcade Page";
	}

	if (usagePage >= 0xff00)
		return "Vendor Usage " + int2str_sal(usagePage, Hexadecimal, 16);

	return "Reserved " + int2str_sal(usagePage, Hexadecimal, usagePage > 0xff ? 16 : 8);
}

std::string GetHIDUsageName(U16 usagePage, U16 usageID)
{
	const size_t NUM_ITEMS = sizeof(UsageTables) / sizeof(UsageTableEntry);

	UsageTableEntry srch;
	srch.usagePage = usagePage;
	srch.usageID = usageID;

	const UsageTableEntry* res = std::lower_bound(UsageTables, UsageTables + NUM_ITEMS, srch);

	if (res->usagePage == usagePage  &&  res->usageID == usageID)
		return res->usageName;

	// handle special cases
	if (usagePage == 0x09)		// button
		return "Button " + int2str(usageID);

	return "Usage Page=" + GetHIDUsagePageName(usagePage) + " ID=" + int2str_sal(usageID, Hexadecimal, usageID > 0x7f ? 16 : 8);
}
